﻿//#define USE_EXTRA_FAILSAFE

using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Data;
using System.Data.Common;
using System.Globalization;
using System.Reflection;
using System.Text;
using System.Text.RegularExpressions;
using System.Transactions;
using Produmex.Foundation;
using Produmex.Foundation.Barcode;
using Produmex.Foundation.Collections;
using Produmex.Foundation.Data;
using Produmex.Foundation.Data.DbClient;
using Produmex.Foundation.Data.Sbo;
using Produmex.Foundation.Data.Sbo.BusinessObjects;
using Produmex.Foundation.Data.Sbo.BusinessObjects.Convertors;
using Produmex.Foundation.Data.Sbo.BusinessObjects.Definitions;
using Produmex.Foundation.Data.Sbo.BusinessObjects.Definitions.Tables;
using Produmex.Foundation.Data.Sbo.DataObjects;
using Produmex.Foundation.Data.Sbo.Generators.BarcodeController;
using Produmex.Foundation.Data.Sbo.Providers;
using Produmex.Foundation.Data.Sbo.Utilities;
using Produmex.Foundation.Data.SqlClient;
using Produmex.Foundation.Diagnostics;
using Produmex.Foundation.GS1;
using Produmex.Foundation.SlimScreen;
using Produmex.Foundation.SlimScreen.Interfaces;
using Produmex.Foundation.SlimScreen.Interfaces.Definitions;
using Produmex.Foundation.SlimScreen.Interfaces.Definitions.KnownDataSets;
using Produmex.Foundation.SlimScreen.WinGui;
using Produmex.Foundation.Utilities;
using Produmex.Foundation.Workflows;
using Produmex.Foundation.Workflows.Parameters;
using Produmex.Foundation.Wwf.Sbo.LocalServices;
using Produmex.Sbo.Logex.Data.BusinessObjects;
using Produmex.Sbo.Logex.Data.BusinessObjects.Definitions;
using Produmex.Sbo.Logex.Data.BusinessObjects.Definitions.Tables;
using Produmex.Sbo.Logex.Data.DataObjects;
using Produmex.Sbo.Logex.Data.Functions.Definitions;
using Produmex.Sbo.Logex.Data.Providers;
using Produmex.Sbo.Logex.Data.Utilities;
using Produmex.Sbo.Logex.Data.ViewObjects.Definitions.Views;

// PmxWorkflowExecutionTypes.SUB_FLOW
namespace WorkflowScript_PickingScript
{
    /// <summary>
    /// Picking flow for Harvard engineering.
    /// It uses the 6.1 (early) version.
    /// The difference between 6.0 version and this one is the 'Force to scan SSCC for item picking' option
    /// It is always set to true in this flow
    /// </summary>
    public class WorkflowScript : WorkflowInstanceScriptBase
    {
        private static readonly ILog s_log = LogProvider.GetLogger(MethodInfo.GetCurrentMethod().DeclaringType);

        // Input parameters
        public ReadOnlyBinder<CultureInfo> DefaultCultureInfo;
        public ReadOnlyBinder<PmxOseCompany> PmxOseCompany;

        public ReadOnlyBinder<bool> MultiPicking;
        public ReadOnlyBinder<int> CrossDockingPickListDocEntry;
        public ReadOnlyBinder<int> CrossDockingWaveKey;
        public ReadOnlyBinder<bool> ZonePicking;
        public ReadOnlyBinder<bool> ProductionPicking;


        // Output parameters
        // <none>

        // Sub-flows
        WorkflowInstanceBase m_enterPickingQuantityHookFlow;
        WorkflowInstanceBase m_enterQuantitySubFlow;
        WorkflowInstanceBase m_cycleCountSubFlow;
        WorkflowInstanceBase m_countPickedColliSubFlow;
        WorkflowInstanceBase m_afterPickListPackedHookFlow;
        WorkflowInstanceBase m_printInvoiceDocumentsSubFlow;
        WorkflowInstanceBase m_beforePickingWaveHookFlow;
        WorkflowInstanceBase m_afterLogisticUnitFullHookScript;
        WorkflowInstanceBase m_enterSerialNumbersForFullPalletPickingHookFlow;

        WorkflowInstanceBase m_afterItemPickedHookScript;


        WorkflowInstanceBase m_printDeliveryDocumentsSubFlow;
        protected int? m_subFlowPickListEntry;

        protected bool m_subFlowBackRequested;
        protected Collection<string> m_subFlowListOfSerialNumbers;
        protected string m_subFlowBaseItemCode;
        protected Dictionary<string, object> m_subFlowDeliveryData;
        protected double m_subFlowQuantity;
        protected double? m_subFlowQuantityUom2;
        protected BarcodeContainer SubScannedBarcodeContainer;
        protected bool m_subFlowSuccess;
        protected IPreAndPostActionInTransaction m_preAndPostActionOnPickItem;
        protected bool m_subFlowAutomaticShipping;
        protected int m_subFlowWaveKey;
        protected bool m_subFlowSerialNumbersSelected;

        protected int m_luidAfterLogisticUnitFullHookFlow;


        //Enter LUID data
        private WorkflowInstanceBase m_enterLUIDDataFlow;
        protected int m_subLUID;


        //sales delivery created hook flow
        private WorkflowInstanceBase m_salesDeliveryCreatedHookFlow;
        private DocumentRef[] m_documentRefsSalesDeliveryCreatedHookFlow;

        private WorkflowInstanceBase m_salesInvoiceCreatedHookFlow;
        private Collection<DocumentRef> m_documentRefsSalesInvoiceCreatedHookFlow;

        //cycle count private  members
        private int m_otherOperationTypeCycleCountingFlow = (int)PmxOtherOperationForCycleCountType.Picking;

        //cycle count private  members
        private int m_otherOperationTypeAlternateCycleCountingFlow = (int)PmxOtherOperationForCycleCountType.AlternatePicking;

        // Private fields
        ISboProviderService m_sboProviderService;
        List<string> m_availablePickListTypes = null;
        List<string> m_availableItemPickType1 = null;
        List<string> m_availableItemPickType2 = null;

        //enter sales delivery data sub flow
        private WorkflowInstanceBase m_enterSalesDeliveryDataSubFlow;
        private bool m_backRequestedEnterSalesDeliveryDataSubFlow = false;
        private int? m_waveKeyEnterSalesDeliveryDataSubFlow = null;
        private int? m_routeKeyEnterSalesDeliveryDataSubFlow = null;
        private Collection<int> m_picklistEntriesEnterSalesDeliveryDataSubFlow = null;
        private Dictionary<string, object> m_deliveryHeaderDataEnterSalesDeliveryDataSubFlow;


        private static class Definitions
        {
            /// <summary>
            /// Minimum number of characters when entering reason
            /// </summary>
            public const int MIMIMUM_NUMBER_OF_CHARACTERS_FOR_REASON = 10;
        }


        private class MultiPickingPickListInfo
        {
            public int DocEntry;
            public string MovableLocation;
            public bool Scanned;
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="WorkflowScript"/> class.
        /// </summary>
        /// <param name="parent">The parent.</param>
        /// <param name="factory">The factory.</param>
        public WorkflowScript(WorkflowInstanceBase parent, WorkflowInstanceFactory factory)
            : base(parent, factory)
        {
            m_enterPickingQuantityHookFlow = this.Create("EnterPickingQuantityHookFlow");
            m_enterQuantitySubFlow = this.Create("EnterQuantityScript");
            m_cycleCountSubFlow = this.Create("CycleCountingScript");
            m_countPickedColliSubFlow = Create("CountPickedColliScript");
            m_afterPickListPackedHookFlow = this.Create("AfterPickListPackedHookScript");
            m_beforePickingWaveHookFlow = this.Create("BeforePickingWaveHookScript");
            m_afterLogisticUnitFullHookScript = this.Create("AfterLogisticUnitFullHookScript");
            m_enterSalesDeliveryDataSubFlow = Create("EnterSalesDeliveryDataScript");
            m_afterItemPickedHookScript = this.Create("AfterItemPickedHookScript");
            m_enterSerialNumbersForFullPalletPickingHookFlow = this.Create( "EnterSerialNumbersForFullPalletPickingHookScript" );
        }


        #region WorkflowInstanceScriptBase Members

        /// <summary>
        /// Executes this instance.
        /// </summary>
        protected override void Execute()
        {
            // Parameters in scope
            Session session = (Session)GetScopeParameter("Session");
            m_sboProviderService = (ISboProviderService)GetScopeParameter("<WwfService>ISboService");

            m_salesDeliveryCreatedHookFlow = Create("SalesDeliveryCreatedHookScript");
            m_salesDeliveryCreatedHookFlow.BindParameters(
                this.DefaultCultureInfo,
                this.PmxOseCompany,
                new ReadOnlyBinder<DocumentRef[]>("DocumentRefs", this, "m_documentRefsSalesDeliveryCreatedHookFlow")
            );


            m_salesInvoiceCreatedHookFlow = Create("SalesInvoiceCreatedHookScript");
            m_salesInvoiceCreatedHookFlow.BindParameters(
                this.DefaultCultureInfo,
                this.PmxOseCompany,
                new ReadOnlyBinder<Collection<DocumentRef>>("DocRefs", this, "m_documentRefsSalesInvoiceCreatedHookFlow")
                );

            m_enterLUIDDataFlow = Create("EnterLUIDDataScript");
            m_enterLUIDDataFlow.BindParameters(
                this.DefaultCultureInfo,
                this.PmxOseCompany,
                new ReadOnlyBinder<int>("LUID", this, "m_subLUID")
            );

            string pmxWarehouseCode = null;
            string pmxWarehouseCodeOfDevice = null;
            string[] customDescrFields = null;
            string customDescrJoin = null;
            bool autoFillQuantityOnPackagingType = false;
            bool autoSelectWave = false;
            string waveOrderBy = null;
            bool autoSelectItri = false;
            bool allowToIdentifySSCCToPickOn = false;
            bool allowToSelectItemToPick = false;
            bool consolidateItems = false;
            bool disableSelectionOfItems = false;
            bool adHocAlternatePicking = false;
            bool copyBatchNumberOnAlternate = false;
            bool forceUserToPickFullPallet = false;
            bool allowCycleCountOfOriginalPickLocationOnAlternatePicking = false;
            bool lockWaveByWaveZone = false;
            bool onlyShowCompletelyUnlockedWavesOnZonePicking = false;
            bool keepPickingFromTheSameLocation = false;
			bool shownPickLocationScreen = false;
            bool automaticSelectionOfMoveableLocation = false;
            BarcodeContainer bcScanned = null;
            bool allowToCreateMasterSSCC = false;
            bool forceToScanSSCCOnItemPick = false;


            ReadOnlyCollection<string> listOfQualityStatusCanBeShipped = null;
            IBarcodeController barcodeController = null;

            List<string> pmxWarehouseCodes = null;

            string initialErrorKey = null;
            Produmex.Foundation.Messages.Message msg = null;
            ProdumexError pmxError = null;
            string query;
            string[] buttonKeys;

			List<string> buttonList = new List<string>();

            int maxTimeShowMsgItemHasBeenPicked = 0;

            m_sboProviderService.InvokeMethodWithDbConnection<object>(false, false, null, null,
                delegate(PmxDbConnection conn, object[] parameters)
                {
                    PmxOseWarehouseProvider prov = new PmxOseWarehouseProvider(conn);
                    pmxWarehouseCodeOfDevice = prov.GetWarehouseOfElement(session.DeviceInstanceId);

                    //Get the assigned warehouses to the scanner
                    pmxWarehouseCodes = new List<string>();

                    query = BuildQuery.GetAssignedWarehousesOnDeviceID(session.DeviceInstanceId);
                    using (ISboRecordset rs = SboRecordsetHelper.RunQuery(s_log, query, conn))
                    {
                        while (!rs.EoF)
                        {
                            pmxWarehouseCodes.Add(rs[0].ToString());
                            rs.MoveNext();
                        }
                    }

                    if (!pmxWarehouseCodes.Contains(pmxWarehouseCode))
                    {
                        pmxWarehouseCodes.Add(pmxWarehouseCode);
                    }





                    PmxPickListProvider plProv = new PmxPickListProvider(conn);
                    Dictionary<string, PmxParameterValue> plcParams = plProv.GetParametersPickListController();
                    PmxParameterValue paramValue;
                    if (plcParams.TryGetValue(PmxExtensionParameterCodes.PICKLIST_CONTROLLER_CUSTOM_WAVE_DESCR_FIELDS, out paramValue))
                    {
                        if (!string.IsNullOrEmpty(paramValue.ParameterValue))
                        {
                            customDescrFields = paramValue.ParameterValue.Split(',', ';', '|');
                        }
                    }
                    if (plcParams.TryGetValue(PmxExtensionParameterCodes.PICKLIST_CONTROLLER_CUSTOM_WAVE_DESCR_JOIN_SQL, out paramValue))
                    {
                        customDescrJoin = paramValue.ParameterValue;
                    }

                    autoFillQuantityOnPackagingType = PmxParameterValueProvider.GetParameterValue<bool>(plcParams, PmxExtensionParameterCodes.PICKLIST_CONTROLLER_AUTO_FILL_QTY_PACK_TYPES, false);
                    autoSelectWave = PmxParameterValueProvider.GetParameterValue<bool>(plcParams, PmxExtensionParameterCodes.PICKLIST_CONTROLLER_AUTO_SELECT_WAVE, false);
                    waveOrderBy = PmxParameterValueProvider.GetParameterValue<string>(plcParams, PmxExtensionParameterCodes.PICKLIST_CONTROLLER_WAVE_ORDER_BY, PmxWaveOrderByCodes.PRIORITY_DUE_KEY);
                    autoSelectItri = PmxParameterValueProvider.GetParameterValue<bool>(plcParams, PmxExtensionParameterCodes.PICKLIST_CONTROLLER_AUTO_SELECT_ITRI, false);
                    allowToIdentifySSCCToPickOn = PmxParameterValueProvider.GetParameterValue<bool>(plcParams, PmxExtensionParameterCodes.PICKLIST_CONTROLLER_ALLOW_TO_IDENTIFY_SSCC_TO_PICK_ON, false);
                    allowToSelectItemToPick = PmxParameterValueProvider.GetParameterValue<bool>(plcParams, PmxExtensionParameterCodes.PICKLIST_CONTROLLER_ALLOW_TO_SELECT_ITEM_TO_PICK, false);
                    consolidateItems = PmxParameterValueProvider.GetParameterValue<bool>(plcParams, PmxExtensionParameterCodes.PICKLIST_CONTROLLER_CONSOLIDATE_ITEMS_TO_PICK, false);
                    disableSelectionOfItems = PmxParameterValueProvider.GetParameterValue<bool>(plcParams, PmxExtensionParameterCodes.PICKLIST_CONTROLLER_DISABLE_SELECTION_OF_ITEMS, false);
                    adHocAlternatePicking = PmxParameterValueProvider.GetParameterValue<bool>(plcParams, PmxExtensionParameterCodes.PICKLIST_CONTROLLER_AD_HOC_ALTERNATE_ITEM, false);
                    copyBatchNumberOnAlternate = PmxParameterValueProvider.GetParameterValue<bool>(plcParams, PmxExtensionParameterCodes.PICKLIST_CONTROLLER_COPY_BATCH_FOR_ALTERNATE_ITEMS, false);
                    forceUserToPickFullPallet = PmxParameterValueProvider.GetParameterValue<bool>(plcParams, PmxExtensionParameterCodes.PICKLIST_CONTROLLER_FORCE_PICK_FULL_LUID, false);
                    allowCycleCountOfOriginalPickLocationOnAlternatePicking = PmxParameterValueProvider.GetParameterValue<bool>(plcParams, PmxExtensionParameterCodes.PICKLIST_CONTROLLER_ALLOW_CYCLE_COUNT_OF_ORIGINAL_PICK_LOCATION_ON_ALTERNATE_PICKING, false);
                    lockWaveByWaveZone = PmxParameterValueProvider.GetParameterValue<bool>(plcParams, PmxExtensionParameterCodes.PICKLIST_CONTROLLER_LOCK_WAVE_BY_ZONE_USER_FOR_ZONE_PICKING, false);
                    onlyShowCompletelyUnlockedWavesOnZonePicking = PmxParameterValueProvider.GetParameterValue<bool>(plcParams, PmxExtensionParameterCodes.PICKLIST_CONTROLLER_ONLY_SHOW_COMPLETELY_UNLOCKED_WAVES_FOR_ZONE_PICKING, false);
                    maxTimeShowMsgItemHasBeenPicked = PmxParameterValueProvider.GetParameterValue<int>(plcParams, PmxExtensionParameterCodes.PICKLIST_CONTROLLER_MSG_ITEM_IS_PICKED_TIME_INTERVAL, 0);
                    keepPickingFromTheSameLocation = PmxParameterValueProvider.GetParameterValue<bool>(plcParams, PmxExtensionParameterCodes.PICKLIST_CONTROLLER_KEEP_PICKING_SAME_LOCATION_PICKING, false);
                    automaticSelectionOfMoveableLocation = PmxParameterValueProvider.GetParameterValue<bool>( plcParams, PmxExtensionParameterCodes.PICKLIST_CONTROLLER_AUTOMATIC_SELECTION_OF_MOVEABLE_LOCATION, false );
                    allowToCreateMasterSSCC = PmxParameterValueProvider.GetParameterValue<bool>(plcParams, PmxExtensionParameterCodes.PICKLIST_CONTROLLER_ALLOW_TO_CREATE_MASTER_SSCC, false);
                    //CODE BELOW IS FOR THE 6.1 VERSION
                    //ALWAYS SET TO TRUE FOR NOW
                    //forceToScanSSCCOnItemPick = PmxParameterValueProvider.GetParameterValue<bool>(plcParams, PmxExtensionParameterCodes.PICKLIST_CONTROLLER_PICKING_FORCE_SCAN_SSCC_ON_ITEM_PICKING, false);
                    forceToScanSSCCOnItemPick = true;
                    PmxQualityStatusProvider qsProv = new PmxQualityStatusProvider(conn);
                    listOfQualityStatusCanBeShipped = qsProv.GetQualityStatusesCanBeShippedAsCollection();


                    PmxBarcodeProvider bcProv = new PmxBarcodeProvider(conn);
                    barcodeController = bcProv.GetBarcodeController();

                    return null;
                });



            List<string> buttonKeyList = null;

            bool multiPicking = (this.MultiPicking != null) && this.MultiPicking.Get();
            SortedList<int, MultiPickingPickListInfo> multiPickingPickListInfos = null;

            bool zonePicking = (this.ZonePicking != null) && this.ZonePicking.Get();

            bool isProductionPicking = (this.ProductionPicking != null) && this.ProductionPicking.Get();


            //bool isAutomaticInvoicing = false;
            int nrOfDocsPrinted = 0;

            Collection<string> listOfSerialNumbers = null;

            List<string> tempStringList = new List<string>();

            bool allowMultipleCustomersOnSSCC = m_sboProviderService.InvokeMethodWithDbConnection<bool>(false, false, null, null,
                   delegate(PmxDbConnection dbConn, object[] parameters)
                   {
                       PmxPickListProvider prov = new PmxPickListProvider(dbConn);

                       return prov.AllowMultipleCustomersOnSSCC();

                   });
            bool askReprint = false;

            bool crossDockingCompleted = false;
            bool goBack = false;
            bool goBackZonePicking = false;
            bool userEnteredSSCCToPickOn = false;
            bool ssccToPickOnHasStock = true;
            //bool multiPickingFirstPickListHasBeenScanned = false;

            int selectedItemToPickIndex = 0;
            bool needToSelectItems = false;
            DataSet dsZones = null;

            bool itemLabelsPrinted = false;


            string alternateSSCC = null;
            //string alternatePickLocation = null;
            DataSet dsItemsOnSSCC = null;
            bool alternateCopyBatchNumbers = false;
            int? alternateLUID = null;
            //string alternateLUIDLocationCode = null;
            bool alternateIsMonoLotPallet = false;
            //int? alternateItri = null;
            string enteredLocationCode = null;
            string enteredBatchNumber = null;
            string enteredQualityStatusCode = null;
            DateTime? enteredBestBeforeDate = null;
            int? enteredItriKey = null;

            Collection<int> listOfPickListsToBeAutoShipped = null;

            string reasonCode = null;
            string reasonName = null;
            string reasonFreeText = null;
            Dictionary<string, bool> reasonsCodesRequiresFreeText = null;


            DataSet dsSupplierPalletNumber = null;
            string supplierPalletNumber = null;
            int waveKey = 0;

            bool alwaysSetToStatusPicked = false;
            DataSet ds = null;

            bool createMasterSSCC = false;
            bool masterLUIDFull = false;
            PickingItemInfo pickingItemInfo = null;

            goto Step_SelectWave;

        Step_LogStopWave:

            #region LOG STOP WAVE

            ExecuteMethodWithDbConnection<bool>(out pmxError, false, false,
               delegate(PmxDbConnection conn, object[] parameters)
               {
                   PmxWaveProvider waveProv = new PmxWaveProvider(conn);

                   //Log stop time of wave/picklist
                   waveProv.LogStop(waveKey, zonePicking ? PmxDocStartStopOperationTypeColumnValues.PickList.ZonePicking.VALUE : PmxDocStartStopOperationTypeColumnValues.PickList.Picking.VALUE);

                   return true;
               });
            if (pmxError != null)
            {

                session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
                    this.DefaultCultureInfo, BuildParamCollection(
                    "PmxError", pmxError,
                    "ShowButton", true
                    ));

                msg = WaitForMessage();

            }
            goto Step_SelectWave;


            #endregion

        Step_SelectWave:

            #region SELECT WAVE

            alwaysSetToStatusPicked = false;


            waveKey = 0;
            pmxWarehouseCode = pmxWarehouseCodeOfDevice;


            if (multiPicking)
            {
                #region Multi-picking wave selection/assembly and picklist/movable location assignment


                Collection<string> listOfUsedMoveableLocation = new Collection<string>();


                if (consolidateItems)
                {
                    session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
                     DefaultCultureInfo.Get(), BuildParamCollection(
                         "MessageKey", "MSG_ERROR_CANNOT_MULTIPICK_AND_CONSOLIDATE_PICKING",
                         "IsError", true,
                         "ShowButton", true));

                    msg = WaitForMessage();
                    return;
                }

                session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IScanStringValueScreen),
                    this.DefaultCultureInfo, BuildParamCollection(
                    "InitialErrorKey", initialErrorKey,
                    "TitleKey", "MSG_TITLE_SCAN_PICKLIST_NUMBER",
                    "ForceDataEntry", true,
                    "AllowToGoBack", true));
                msg = WaitForMessage();
                initialErrorKey = null;

                if (msg.Name.EndsWith(".IScanStringValueScreenReply.BackRequested"))
                {
                    return;
                }
                if (!msg.Name.EndsWith(".IScanStringValueScreenReply.StringScanned"))
                {
                    initialErrorKey = KnownMessageKeys.MSG_ERROR_OPERATION_NOT_SUPPORTED;
                    goto Step_SelectWave;
                }


                // Get info on scanned picklist
                string pickListBarcode = ExtractParameter<string>(msg.Parameters, "stringValue");
                DataSet dsPickLists = null;
                bool pickingNeeded = false;
                multiPickingPickListInfos = m_sboProviderService.InvokeMethodWithDbConnection<SortedList<int, MultiPickingPickListInfo>>(false, false, null, null,
                    delegate(PmxDbConnection dbConn, object[] parameters)
                    {
                        return GetMultiPickingPickListInfo(pickListBarcode, isProductionPicking, pmxWarehouseCodes, out dsPickLists, out pickingNeeded, dbConn);
                    });

                pmxWarehouseCode = Convert.ToString(dsPickLists.Tables[0].Rows[0][PmxOseGeneralTable.Columns.PmxWarehouseCode.NAME]);

                // Picking needed?
                if (!pickingNeeded)
                {
                    session.ShowScreen(typeof(IShowMessageScreen),
                        this.DefaultCultureInfo, BuildParamCollection(
                        "MessageKey", "MSG_ERROR_NO_PICKLIST_FOUND",
                        "MessageAddendum", pickListBarcode,
                        "ShowButton", true));
                    msg = WaitForMessage();

                    goto Step_SelectWave;
                }

                // Locked by another user?
                string lockedBy = Convert.ToString(dsPickLists.Tables[0].Rows[0][SboUserTable.Columns.UserCode.NAME]);
                if (!String.IsNullOrEmpty(lockedBy) && String.Compare(lockedBy, m_sboProviderService.GetCurrentUser().Username, true) != 0)
                {
                    initialErrorKey = "MSG_ERROR_WAVE_ALREADY_LOCKED";
                    goto Step_SelectWave;
                }

                int scannedPickListDocEntry = Convert.ToInt32(dsPickLists.Tables[0].Rows[0]["spl_DocEntry"]);

                if (GetNrOfRows(dsPickLists) == 1)
                {
                    // Assemble new wave
                    // Scan all picklists & boxes and update info in db

                    int nrOfPickListsScanned = 0;

                    while (true)
                    {
                        if (scannedPickListDocEntry == 0)
                        {
                            session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IScanStringValueScreen),
                                this.DefaultCultureInfo, BuildParamCollection(
                                "InitialErrorKey", initialErrorKey,
                                "TitleKey", "MSG_TITLE_SCAN_PICKLIST_NUMBER",
                                "ForceDataEntry", true,
                                "Information", String.Format("#{0}", nrOfPickListsScanned + 1),
                                "AllowedButtonKeys", new string[] { "MSG_BUTTON_DONE" },
                                "AllowToGoBack", true));
                            msg = WaitForMessage();
                            initialErrorKey = null;

                            if (msg.Name.EndsWith(".IScanStringValueScreenReply.BackRequested"))
                            {
                                goto Step_SelectWave;
                            }
                            if (msg.Name.EndsWith(".IScanStringValueScreenReply.ButtonClicked"))
                            {
                                // Done with adding picklists
                                break;
                            }
                            if (!msg.Name.EndsWith(".IScanStringValueScreenReply.StringScanned"))
                            {
                                initialErrorKey = KnownMessageKeys.MSG_ERROR_OPERATION_NOT_SUPPORTED;
                                continue;
                            }
                            string pickListToAddBarcode = ExtractParameter<string>(msg.Parameters, "stringValue");


                            int dummy;
                            if (!int.TryParse(pickListToAddBarcode, out dummy))
                            {
                                session.ShowScreen(typeof(IShowMessageScreen),
                                    this.DefaultCultureInfo, BuildParamCollection(
                                    "MessageKey", "MSG_ERROR_NO_PICKLIST_FOUND",
                                    "MessageAddendum", pickListToAddBarcode,
                                    "ShowButton", true));
                                msg = WaitForMessage();
                                continue;
                            }

                            DataSet dsPickListToAdd = null;
                            m_sboProviderService.InvokeMethodWithDbConnection<object>(false, false, null, null,
                                delegate(PmxDbConnection dbConn, object[] parameters)
                                {
                                    GetMultiPickingPickListInfo(pickListToAddBarcode, isProductionPicking, pmxWarehouseCode, out dsPickListToAdd, out pickingNeeded, dbConn);
                                    return null;
                                });
                            if (GetNrOfRows(dsPickListToAdd) > 1)
                            {
                                initialErrorKey = "MSG_ERROR_PICKLIST_FROM_OTHER_WAVE";
                                continue;
                            }
                            else if (GetNrOfRows(dsPickListToAdd) < 1 || !pickingNeeded)
                            {
                                session.ShowScreen(typeof(IShowMessageScreen),
                                    this.DefaultCultureInfo, BuildParamCollection(
                                    "MessageKey", "MSG_ERROR_NO_PICKLIST_FOUND",
                                    "MessageAddendum", pickListToAddBarcode,
                                    "ShowButton", true));
                                msg = WaitForMessage();
                                continue;
                            }

                            // Locked by another user?
                            lockedBy = Convert.ToString(dsPickListToAdd.Tables[0].Rows[0][SboUserTable.Columns.UserCode.NAME]);
                            if (!String.IsNullOrEmpty(lockedBy) && String.Compare(lockedBy, m_sboProviderService.GetCurrentUser().Username, true) != 0)
                            {
                                initialErrorKey = "MSG_ERROR_PICKLIST_FROM_OTHER_WAVE";
                                continue;
                            }

                            scannedPickListDocEntry = Convert.ToInt32(dsPickListToAdd.Tables[0].Rows[0]["spl_DocEntry"]);
                        }
                        MultiPickingPickListInfo info;
                        if (!multiPickingPickListInfos.TryGetValue(scannedPickListDocEntry, out info))
                        {
                            info = new MultiPickingPickListInfo();
                            info.DocEntry = scannedPickListDocEntry;
                            multiPickingPickListInfos.Add(scannedPickListDocEntry, info);
                        }
                        else if (info.Scanned)
                        {
                            initialErrorKey = "MSG_ERROR_PICKLIST_ALREADY_SCANNED";
                            scannedPickListDocEntry = 0;
                            continue;
                        }

                        session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IScanStringValueScreen),
                            this.DefaultCultureInfo, BuildParamCollection(
                            "InitialErrorKey", initialErrorKey,
                            "TitleKey", "MSG_TITLE_IDENTIFY_MOVEABLE_LOCATION",
                            "ForceDataEntry", true,
                            "Information", "Pick list",
                            "MessageAddendum", scannedPickListDocEntry.ToString(),
                            "AllowToGoBack", true));
                        msg = WaitForMessage();
                        initialErrorKey = null;

                        if (msg.Name.EndsWith(".IScanStringValueScreenReply.BackRequested"))
                        {
                            scannedPickListDocEntry = 0;
                            goto Step_SelectWave;
                        }
                        if (!msg.Name.EndsWith(".IScanStringValueScreenReply.StringScanned"))
                        {
                            initialErrorKey = KnownMessageKeys.MSG_ERROR_OPERATION_NOT_SUPPORTED;
                            continue;
                        }

                        string scannedMovLoc = ExtractParameter<string>(msg.Parameters, "stringValue");
                        m_sboProviderService.InvokeMethodWithDbConnection<object>(false, false, null, null,
                            delegate(PmxDbConnection dbConn, object[] parameters)
                            {
                                query = BuildQuery.GetMovableLocations(scannedMovLoc, pmxWarehouseCode);
                                using (ISboRecordset rs = SboRecordsetHelper.RunQuery(s_log, query, dbConn))
                                {
                                    if (rs.EoF)
                                    {
                                        initialErrorKey = "MSG_ERROR_INVALID_MOVEABLE_LOCATION";
                                        return null;
                                    }
                                    scannedMovLoc = rs.GetTypedValue<string>(LocationDataSet.Columns.CODE.Name);
                                }

                                return null;
                            });
                        if (!String.IsNullOrEmpty(initialErrorKey))
                        {
                            continue;
                        }

                        if (listOfUsedMoveableLocation.Contains(scannedMovLoc))
                        {
                            initialErrorKey = "MSG_ERROR_MOVEABLE_LOCATION_USED_FOR_WAVE";
                            continue;
                        }
                        else
                        {
                            listOfUsedMoveableLocation.Add(scannedMovLoc);
                        }

                        info.MovableLocation = scannedMovLoc;
                        info.Scanned = true;
                        nrOfPickListsScanned++;

                        scannedPickListDocEntry = 0;
                    }

                    int? newWaveKey = m_sboProviderService.InvokeMethodWithDbConnection<int?>(false, false, null, null,
                        delegate(PmxDbConnection dbConn, object[] parameters)
                        {
                            List<int> pickListDocEntries = new List<int>();
                            foreach (MultiPickingPickListInfo infoToVerify in multiPickingPickListInfos.Values)
                            {
                                DataSet dsVerify;
                                GetMultiPickingPickListInfo(pickListBarcode, isProductionPicking, pmxWarehouseCode, out dsVerify, out pickingNeeded, dbConn);

                                if (GetNrOfRows(dsVerify) != 1 || !pickingNeeded)
                                {
                                    initialErrorKey = "MSG_ERROR_PICKLISTS_CHANGED";
                                    return null;
                                }

                                // Locked by another user?
                                lockedBy = Convert.ToString(dsVerify.Tables[0].Rows[0][SboUserTable.Columns.UserCode.NAME]);
                                if (!String.IsNullOrEmpty(lockedBy) && String.Compare(lockedBy, m_sboProviderService.GetCurrentUser().Username, true) != 0)
                                {
                                    initialErrorKey = "MSG_ERROR_PICKLISTS_CHANGED";
                                    return null;
                                }

                                pickListDocEntries.Add(infoToVerify.DocEntry);
                            }

                            PmxWaveProvider waveProv = new PmxWaveProvider(dbConn);
                            int? newWaveKey2 = waveProv.AddPickListsToWave(pickListDocEntries, null, true);

                            initialErrorKey = UpdateMultiPickingMovLoc(multiPickingPickListInfos, dbConn);

                            return newWaveKey2;
                        });
                    if (!newWaveKey.HasValue)  // Some error occurred
                    {
                        goto Step_SelectWave;
                    }

                    waveKey = newWaveKey.Value;
                }
                else
                {
                    // Already assembled wave
                    // Scan all picklists & boxes and match with or update info in db

                    int nrOfPickListsToScan = multiPickingPickListInfos.Count;

                    //get all current moveable locations

                    foreach (MultiPickingPickListInfo pl in multiPickingPickListInfos.Values)
                    {
                        if (!string.IsNullOrEmpty(pl.MovableLocation))
                        {
                            listOfUsedMoveableLocation.Add(pl.MovableLocation);
                        }
                    }


                    while (nrOfPickListsToScan > 0)
                    {
                        if (scannedPickListDocEntry == 0)
                        {
                            session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IScanStringValueScreen),
                                this.DefaultCultureInfo, BuildParamCollection(
                                "InitialErrorKey", initialErrorKey,
                                "TitleKey", "MSG_TITLE_SCAN_PICKLIST_NUMBER",
                                "ForceDataEntry", true,
                                "Information", String.Format("{0}/{1}", multiPickingPickListInfos.Count - nrOfPickListsToScan + 1, multiPickingPickListInfos.Count),
                                "AllowToGoBack", true));
                            msg = WaitForMessage();
                            initialErrorKey = null;

                            if (msg.Name.EndsWith(".IScanStringValueScreenReply.BackRequested"))
                            {
                                goto Step_SelectWave;
                            }
                            if (!msg.Name.EndsWith(".IScanStringValueScreenReply.StringScanned"))
                            {
                                initialErrorKey = KnownMessageKeys.MSG_ERROR_OPERATION_NOT_SUPPORTED;
                                continue;
                            }
                            scannedPickListDocEntry = Convert.ToInt32(ExtractParameter<string>(msg.Parameters, "stringValue"));
                            if (!multiPickingPickListInfos.ContainsKey(scannedPickListDocEntry))
                            {
                                initialErrorKey = "MSG_ERROR_PICKLIST_NOT_IN_WAVE";
                                scannedPickListDocEntry = 0;
                                continue;
                            }
                        }
                        MultiPickingPickListInfo info = multiPickingPickListInfos[scannedPickListDocEntry];
                        if (info.Scanned)
                        {
                            initialErrorKey = "MSG_ERROR_PICKLIST_ALREADY_SCANNED";
                            scannedPickListDocEntry = 0;
                            continue;
                        }

                        session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IScanStringValueScreen),
                            this.DefaultCultureInfo, BuildParamCollection(
                            "InitialErrorKey", initialErrorKey,
                            "TitleKey", "MSG_TITLE_IDENTIFY_MOVEABLE_LOCATION",
                            "ForceDataEntry", true,
                            "Information", "Pick list: location",
                            "MessageAddendum", String.Format("{0}: {1}", scannedPickListDocEntry, String.IsNullOrEmpty(info.MovableLocation) ? "..." : info.MovableLocation),
                            "AllowToGoBack", true));
                        msg = WaitForMessage();
                        initialErrorKey = null;

                        if (msg.Name.EndsWith(".IScanStringValueScreenReply.BackRequested"))
                        {
                            scannedPickListDocEntry = 0;
                            continue;
                        }
                        if (!msg.Name.EndsWith(".IScanStringValueScreenReply.StringScanned"))
                        {
                            initialErrorKey = KnownMessageKeys.MSG_ERROR_OPERATION_NOT_SUPPORTED;
                            continue;
                        }
                        string scannedMovLoc = ExtractParameter<string>(msg.Parameters, "stringValue");
                        if (String.IsNullOrEmpty(info.MovableLocation))
                        {
                            m_sboProviderService.InvokeMethodWithDbConnection<object>(false, false, null, null,
                                delegate(PmxDbConnection dbConn, object[] parameters)
                                {
                                    query = BuildQuery.GetMovableLocations(scannedMovLoc, pmxWarehouseCode);
                                    using (ISboRecordset rs = SboRecordsetHelper.RunQuery(s_log, query, dbConn))
                                    {
                                        if (rs.EoF)
                                        {
                                            initialErrorKey = "MSG_ERROR_INVALID_MOVEABLE_LOCATION";
                                            return null;
                                        }
                                        scannedMovLoc = rs.GetTypedValue<string>(LocationDataSet.Columns.CODE.Name);
                                    }

                                    return null;
                                });
                            if (!String.IsNullOrEmpty(initialErrorKey))
                            {
                                continue;
                            }

                            if (listOfUsedMoveableLocation.Contains(scannedMovLoc))
                            {
                                initialErrorKey = "MSG_ERROR_MOVEABLE_LOCATION_USED_FOR_WAVE";
                                continue;
                            }
                            else
                            {
                                listOfUsedMoveableLocation.Add(scannedMovLoc);
                            }


                            info.MovableLocation = scannedMovLoc;
                        }
                        else if (String.Compare(info.MovableLocation, scannedMovLoc, true) != 0)
                        {
                            initialErrorKey = "MSG_ERROR_INCORRECT_MOVABLE_LOCATION";
                            continue;
                        }
                        info.Scanned = true;
                        nrOfPickListsToScan--;

                        scannedPickListDocEntry = 0;
                    }

                    m_sboProviderService.InvokeMethodWithDbConnection<object>(false, false, null, null,
                        delegate(PmxDbConnection dbConn, object[] parameters)
                        {
                            initialErrorKey = UpdateMultiPickingMovLoc(multiPickingPickListInfos, dbConn);

                            return null;
                        });
                    if (!String.IsNullOrEmpty(initialErrorKey))
                    {
                        goto Step_SelectWave;
                    }

                    waveKey = Convert.ToInt32(dsPickLists.Tables[0].Rows[0][PmxPickListTable.Columns.WaveKey.NAME]);
                }
                #endregion
            }
            else if (this.CrossDockingWaveKey != null)
            {
                if (goBack || crossDockingCompleted)
                {
                    this.CrossDockingWaveKey = null;
                    this.CrossDockingPickListDocEntry = null;
                    return;
                }

                waveKey = this.CrossDockingWaveKey;
            }

            if (autoSelectWave && goBack)
            {
                return;
            }

            goBack = false;

            if (m_availablePickListTypes == null)  // Get available PickList types for current user
            {
                m_availablePickListTypes = new List<string>();
                m_availableItemPickType1 = new List<string>();
                m_availableItemPickType2 = new List<string>();
                m_sboProviderService.InvokeMethodWithDbConnection<object>(false, false, null, null, delegate(PmxDbConnection dbConn, object[] parameters)
                {
                    DbCommand dbCmd = dbConn.CreateCommand(BuildQuery.GetPickListTypesForUser(m_sboProviderService.GetCurrentUser().Username, dbConn.DbTool));
                    using (DbDataReader rdr = dbCmd.ExecuteReader())
                    {
                        while (rdr.Read())
                        {
                            m_availablePickListTypes.Add((string)rdr[PmxUserPickListTypeTable.Columns.PickListType.NAME]);
                        }
                    }

                    if (zonePicking) 
					{
                        //only zone picking will be split up by item pick type.
                        //Normally picking is locked for 1 user, so if user only can see part of picking items, picking cannot be completed
                        dbCmd = dbConn.CreateCommand(BuildQuery.GetItemPickTypesForUser(m_sboProviderService.GetCurrentUser().Username, dbConn.DbCultureInfo));
                        using (DbDataReader rdr = dbCmd.ExecuteReader())
                        {
                            while (rdr.Read())
                            {
                                if ((string)rdr[PmxUserItemPickTypeTable.Columns.ItemPickTypeProperty.NAME] == PmxItemPickTypePropertyColumnValues.ItemPickType1.VALUE)
                                {
                                    m_availableItemPickType1.Add((string)rdr[PmxUserItemPickTypeTable.Columns.ItemPickType.NAME]);
                                }
                                else if ((string)rdr[PmxUserItemPickTypeTable.Columns.ItemPickTypeProperty.NAME] == PmxItemPickTypePropertyColumnValues.ItemPickType2.VALUE)
                                {
                                    m_availableItemPickType2.Add((string)rdr[PmxUserItemPickTypeTable.Columns.ItemPickType.NAME]);
                                }
                            }
                        }
                    }

                    return null;
                });
            }

            itemLabelsPrinted = false;

            query = BuildQuery.GetPickListForPickingByWave(m_sboProviderService.GetCurrentUser().Username, pmxWarehouseCode, session.DeviceInstanceId, customDescrFields, customDescrJoin, m_availablePickListTypes, m_availableItemPickType1, m_availableItemPickType2, waveOrderBy, zonePicking, zonePicking && onlyShowCompletelyUnlockedWavesOnZonePicking, isProductionPicking, m_sboProviderService.GetDbTool());
            DataSet dsWave = m_sboProviderService.RunView(false, null, null, query);

            if (!multiPicking && this.CrossDockingWaveKey == null)
            {
                if (autoSelectWave && GetNrOfRows(dsWave) > 0)
                {
                    bool foundWave = false;
                    for (int iLoop = 0; iLoop < GetNrOfRows(dsWave); iLoop++)
                    {
                        //autoselect wave
                        waveKey = Convert.ToInt32(dsWave.Tables[0].Rows[iLoop][WaveDataSet.Columns.INTERNAL_KEY.Name]);

                        //autoselect whscode
                        pmxWarehouseCode = dsWave.Tables[0].Rows[iLoop][PmxOseGeneralTable.Columns.PmxWarehouseCode.NAME].ToString();

                        int user = -1;

                        //check if there are items to pick for user
                        if (ExecuteMethodWithDbConnection<bool>(out pmxError, false, false,
                            delegate(PmxDbConnection conn, object[] parameters)
                            {
                                user = conn.InternalKeySboUser;

                                PmxPickListProvider plProv = new PmxPickListProvider(conn);
                                plProv.MakePicklistReadyFromWave(waveKey);

                                Collection<PickingItemInfo> pickingItemInfoListTemp = plProv.GetNextItems(waveKey, null, consolidateItems, m_availableItemPickType1, m_availableItemPickType2);

#if USE_EXTRA_FAILSAFE

                                //Extra failsafe
                                if (pickingItemInfoListTemp != null && pickingItemInfoListTemp.Count > 0 && zonePicking)
                                {
                                    //zone picking, and we found items to pick, so check if actual zones are found
                                    query = BuildQuery.GetAllRecursiveZonesForWave(waveKey, m_sboProviderService.GetCurrentUser().Username, pmxWarehouseCode);

                                    using (ISboRecordset rs = SboRecordsetHelper.RunQuery(s_log, query, conn))
                                    {
                                        if (rs.EoF)
                                        {
                                            //no zones found, so check if there are items to pick
                                            pickingItemInfoListTemp = plProv.GetNextItems(waveKey, "NO_ZONE_ALTERNATE_PICKING", consolidateItems, m_availableItemPickType1, m_availableItemPickType2);
                                        }
                                    }
                                }

#endif



                                return pickingItemInfoListTemp != null && pickingItemInfoListTemp.Count > 0;
                            }))
                        {
                            //So we only show 1 line in the screen
                            DataSet tempDS = dsWave.Clone();
                            tempDS.Tables[0].Rows.Add(dsWave.Tables[0].Rows[iLoop].ItemArray);

                            dsWave = tempDS;

                            foundWave = true;

                            break;
                        }
                        else
                        {
                            s_log.WarnFormat("The wave {0} is skipped for user {1} on auto selection of wave because there are no items to pick.", waveKey, user);
                        }
                    }

                    if (!foundWave)
                    {
                        //no wave found, so clear initial list.
                        DataSet tempDS = dsWave.Clone();
                        dsWave = tempDS;
                    }

                }

                session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.Picking.ISelectWaveScreen),
                    this.DefaultCultureInfo, BuildParamCollection(
                    "InitialErrorKey", initialErrorKey,
                    "TitleKey", "MSG_TITLE_SELECT_PICKLIST",
                    "WaveDS", dsWave));
                msg = WaitForMessage();
                initialErrorKey = null;

                if (msg.Name == "Produmex.Foundation.SlimScreen.Interfaces.Picking.ISelectWaveScreenReply.BackRequested")
                {
                    return;
                }
                if (msg.Name == "Produmex.Foundation.SlimScreen.Interfaces.Picking.ISelectWaveScreenReply.RefreshRequested")
                {
                    goto Step_SelectWave;
                }
                if (msg.Name != "Produmex.Foundation.SlimScreen.Interfaces.Picking.ISelectWaveScreenReply.WaveSelected")
                {
                    initialErrorKey = KnownMessageKeys.MSG_ERROR_OPERATION_NOT_SUPPORTED;
                    goto Step_SelectWave;
                }

                waveKey = ExtractParameter<int>(msg.Parameters, "waveKey");


            }

            DateTime dueDate;
            bool waveSameDelivery = false;
            string waveDescription = null;

            query = BuildQuery.GetAlwaysSetToStatusPickedAfterPicking(waveKey);
            ds = m_sboProviderService.RunView(false, null, null, query);

            alwaysSetToStatusPicked = this.GetNrOfRows(ds) > 0;

            if (GetNrOfRows(dsWave) > 0)
            {
                DataRow[] rows = dsWave.Tables[0].Select(WaveDataSet.Columns.INTERNAL_KEY.Name + "=" + waveKey.ToString());
                if (rows.Length > 0)
                {
                    //if (rows[0][PmxPickListTable.Columns.RouteKey.NAME] is int)
                    //{
                    //    this.RouteKey = Convert.ToInt32(rows[0][PmxPickListTable.Columns.RouteKey.NAME]);
                    //}
                    //else
                    //{
                    //    this.RouteKey = null;
                    //}

                    dueDate = Convert.ToDateTime(rows[0][WaveDataSet.Columns.DUE_DATE.Name]);

                    waveSameDelivery = Convert.ToString(rows[0][PmxWaveTable.Columns.SameDelivery.NAME]) == SboBooleanColumnValues.Y.VALUE;
                    if (waveSameDelivery)
                    {
                        waveDescription = Convert.ToString(rows[0][WaveDataSet.Columns.DESCRIPTION.Name]);
                    }

                    itemLabelsPrinted = Convert.ToString(rows[0][PmxWaveTable.Columns.ItemLabelsPrinted.NAME]) == SboBooleanColumnValues.Y.VALUE;

                    pmxWarehouseCode = Convert.ToString(rows[0][PmxOseGeneralTable.Columns.PmxWarehouseCode.NAME]);

                }
            }

            string pickPackRemarks = null;

            ExecuteMethodWithDbConnection<bool>(out pmxError, false, false,
                delegate(PmxDbConnection conn, object[] parameters)
                {
                    PmxWaveProvider waveProv = new PmxWaveProvider(conn);

                    if (!zonePicking)
                    {
                        //only lock complete wave if no zone picking is done
                        waveProv.LockWave(waveKey);

                    }

                    //Log start time of wave/picklist
                    waveProv.LogStart(waveKey, !zonePicking, zonePicking ? PmxDocStartStopOperationTypeColumnValues.PickList.ZonePicking.VALUE : PmxDocStartStopOperationTypeColumnValues.PickList.Picking.VALUE);

                    pickPackRemarks = waveProv.GetPickPackRemarks(waveKey);

                    return true;
                });
            if (pmxError != null)
            {
                if (pmxError.ErrorCode == Produmex.Sbo.Logex.Data.ProdumexErrorCodes.WaveAlreadyLocked)
                {
                    session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
                        this.DefaultCultureInfo, BuildParamCollection(
                        "MessageKey", "MSG_ERROR_WAVE_ALREADY_LOCKED",
                        "ShowButton", true
                        ));
                }
                else
                {
                    session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
                        this.DefaultCultureInfo, BuildParamCollection(
                        "PmxError", pmxError,
                        "ShowButton", true
                        ));
                }
                msg = WaitForMessage();

                goto Step_SelectWave;
            }

		Step_MakePickListReady:
			
			

			//Split up making pick list ready
			//This is to avoid the transaction to take too long
			query = BuildQuery.GetPickListsForWave(waveKey, m_sboProviderService.GetDbTool().CultureInfo);
			ds = m_sboProviderService.RunView(false, null, null, query);


			if (GetNrOfRows(ds) > 0)
			{
				foreach (DataRow row in ds.Tables[0].Rows)
				{
					int pickListDocEntryForMakePickListReady = Convert.ToInt32(row[PmxPickListTable.Columns.DocEntry.NAME]);
					ExecuteMethodWithDbConnection<bool>(out pmxError, false, false,
					   delegate(PmxDbConnection conn, object[] parameters)
					   {
						   PmxPickListProvider plProv = new PmxPickListProvider(conn);
						   plProv.MakePicklistReadyFromDocEntry(pickListDocEntryForMakePickListReady);


						   return true;
					   });

					if (pmxError != null)
					{
						//error occured, so break;
						break;
					}
				}
			}

			if (pmxError != null)
			{
				session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
					this.DefaultCultureInfo, BuildParamCollection(
					"PmxError", pmxError,
					"ShowButton", true));
				msg = WaitForMessage();

				//retry?
				session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IDecisionScreen),
					DefaultCultureInfo.Get(), BuildParamCollection(
						"TitleKey", "MSG_TITLE_RETRY_UPDATE",
						"MessageKey", "MSG_QUESTION_AN_ERROR_OCCURED_WANT_RETRY"));
				msg = WaitForMessage();

				if (msg.Name.ToLower().EndsWith(".yes"))
				{
					goto Step_MakePickListReady;
				}

				goto Step_SelectWave;
			}

			




            //HOOK
            m_beforePickingWaveHookFlow.BindParameters(
                  this.DefaultCultureInfo,
                  this.PmxOseCompany,
                  new ReadOnlyBinder<int>("WaveKey", waveKey),
                  new ReadWriteBinder<bool>("BackRequested", this, "m_subFlowBackRequested")
              );
            m_subFlowBackRequested = false;
            this.StartSubFlowAndWait(m_beforePickingWaveHookFlow, true);

            if (m_subFlowBackRequested)
            {
                goBack = true;

                goto Step_LogStopWave;
                //goto Step_SelectWave;
            }

            if (!itemLabelsPrinted)
            {

                session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
                        this.DefaultCultureInfo, BuildParamCollection(
                        "MessageKey", "MSG_PRINTING_ITEM_LABELS",
                        "ShowButton", false
                        ));

                //print item labels
                PmxOsePrinter printer = ExecuteMethodWithDbConnection<PmxOsePrinter>(out pmxError, false, false,
                   delegate(PmxDbConnection conn, object[] parameters)
                   {
                       PmxWaveProvider waveProv = new PmxWaveProvider(conn);

                       return waveProv.PrintItemLabels(waveKey, session.DeviceInstanceId);
                   });
                if (pmxError != null)
                {
                    if (pmxError.ErrorCode == Produmex.Sbo.Logex.Data.ProdumexErrorCodes.WaveAlreadyLocked)
                    {
                        session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
                            this.DefaultCultureInfo, BuildParamCollection(
                            "MessageKey", "MSG_ERROR_WAVE_ALREADY_LOCKED",
                            "ShowButton", true
                            ));
                    }
                    else
                    {
                        session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
                            this.DefaultCultureInfo, BuildParamCollection(
                            "PmxError", pmxError,
                            "ShowButton", true
                            ));
                    }
                    msg = WaitForMessage();
                    //on error proceed, because failed printing cannot affect process
                }

                if (printer != null)
                {
                    session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
                            this.DefaultCultureInfo, BuildParamCollection(
                            "MessageKey", "MSG_LABELS_ITEM_PRINTED",
                            "ShowButton", true
                            ));
                    msg = WaitForMessage();
                }
            }


            if (!String.IsNullOrEmpty(pickPackRemarks))
            {
                session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
                    this.DefaultCultureInfo, BuildParamCollection(
                    "MessageKey", pickPackRemarks,
                    "NoTranslationOfMessageKey", true,
                    "ShowButton", true
                    ));
                msg = WaitForMessage();
            }



            needToSelectItems = false;


            // *** From here we really start on the picklist ***
            int pickListDocEntry = 0;

            string targetMoveableLocation = null;
            int? targetLuid = null;
            int? targetMasterLuid = null;
            string targetLogCarrierCode = null;
            string targetLogCarrierStorLocCode = null;
            bool targetLogCarrierCanBeFull = false;
            string dstLocationCode = null;

            string zoneCode = null;

            bool alternateZonePickingForNoZones = false;

            #endregion

        Step_ShowNextItems:

            #region SHOW NEXT ITEMS

            alternateZonePickingForNoZones = false;
			goBack = false;

            if (zonePicking && (string.IsNullOrEmpty(zoneCode) || goBackZonePicking))
            {
                zoneCode = null;

                //zone picking, and no zone is set yet
                dsZones = ExecuteMethodWithDbConnection<DataSet>( out pmxError, false, false,
                    delegate( PmxDbConnection conn, object[] parameters )
                    {
                        string preQueryForHana;
                        query = BuildQuery.GetAllRecursiveZonesForWave( waveKey, m_sboProviderService.GetCurrentUser().Username, pmxWarehouseCode, out preQueryForHana, conn.DbTool );
                        if ( conn.DbTool.DbServerType == Produmex.Foundation.Data.DbClient.DbServerType.HANA )
                        {
                            SboRecordsetHelper.RunQueryNoReturn( s_log, preQueryForHana, conn );
                        }
                        ISboRecordset rs = SboRecordsetHelper.RunQuery( s_log, query, conn );
                        return SboRecordsetHelper.ConvertRecordsetToDataSet( s_log, rs );
                    } );

                if ( pmxError != null )
                {
                    session.ShowScreen( typeof( Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen ),
                        this.DefaultCultureInfo, BuildParamCollection(
                        "PmxError", pmxError,
                        "ShowButton", true
                        ) );
                    msg = WaitForMessage();

                    //select another one
                    goto Step_SelectWave;
                }

                //the wave was selectable, but no zones where available.
                //this means that the remaining items on pick list do not have a location set.
                if (this.GetNrOfRows(dsZones) == 0)
                {
                    if (goBackZonePicking)
                    {
                        goBackZonePicking = false;
                        goBack = true;
                        goto Step_LogStopWave;
                    }
                    ////Get all zones
                    //query = BuildQuery.GetAllRecursiveZonesForWaveAndNoLocationsSet(waveKey);
                    //dsZones = m_sboProviderService.RunView(false, null, null, query);

                    alternateZonePickingForNoZones = true;

                    goto Step_SelectItemToPick;
                }

                goBackZonePicking = false;


                session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.ISelectLocationScreen),
                    DefaultCultureInfo.Get(), BuildParamCollection(
                    "InitialErrorKey", initialErrorKey,
                    "TitleKey", "MSG_TITLE_SELECT_ZONE",
                    "LocationDS", dsZones));

                msg = WaitForMessage();
                // Reset latest error
                initialErrorKey = null;
                zoneCode = null;

                if (msg.Name.EndsWith(".LocationSelected"))
                {
                    zoneCode = ExtractParameter<string>(msg.Parameters, "locationCode");
                }
                else if (msg.Name.EndsWith(".BackRequested"))
                {
                    goBack = true;
                    goto Step_LogStopWave;
                    //goto Step_SelectWave;
                }
            }

        Step_SelectItemToPick:

            Collection<PickingItemInfo> pickingItemInfoList = null;

            ExecuteMethodWithDbConnection<bool>(out pmxError, false, false,
                delegate(PmxDbConnection conn, object[] parameters)
                {
                    PmxPickListProvider plProv = new PmxPickListProvider(conn);
                    pickingItemInfoList = plProv.GetNextItems(waveKey, alternateZonePickingForNoZones ? "NO_ZONE_ALTERNATE_PICKING" : zoneCode, consolidateItems, m_availableItemPickType1, m_availableItemPickType2);

                    return true;
                });

            if (pmxError != null)
            {
                session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
                    this.DefaultCultureInfo, BuildParamCollection(
                    "PmxError", pmxError,
                    "ShowButton", true
                    ));
                msg = WaitForMessage();

                //select another one
                goto Step_SelectWave;
            }



            //we get the zone, so now we lock all items in the zone
            if (lockWaveByWaveZone)
            {
                //lock the wave item
                ExecuteMethodWithDbConnection<bool>(out pmxError, false, false,
                    delegate(PmxDbConnection conn, object[] parameters)
                    {
                        PmxWaveLockProvider waveLockProv = new PmxWaveLockProvider(conn);
                        waveLockProv.LockWaveItems(waveKey, pickingItemInfoList, true);

                        return true;
                    });


                if (pmxError != null)
                {
                    if (pmxError.ErrorCode == Produmex.Sbo.Logex.Data.ProdumexErrorCodes.WaveItemAlreadyLocked)
                    {
                        session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
                            this.DefaultCultureInfo, BuildParamCollection(
                            "MessageKey", "MSG_ERROR_WAVE_ITEM_ALREADY_LOCKED",
                            "ShowButton", true
                            ));
                    }
                    else
                    {
                        session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
                            this.DefaultCultureInfo, BuildParamCollection(
                            "PmxError", pmxError,
                            "ShowButton", true
                            ));
                    }
                    msg = WaitForMessage();

                    //select another one
                    goto Step_SelectWave;
                }
            }

            DataSet dsNextItems = BuildAlternateList(pickingItemInfoList);

            //DataSet dsNextItems = new DataSet();
            //DataTable tblNextItems = dsNextItems.Tables.Add("NextItem");
            //tblNextItems.Columns.Add(AlternateItemDataSet.Columns.INTERNAL_KEY.Name, typeof(int));
            //tblNextItems.Columns.Add(AlternateItemDataSet.Columns.STORAGE_LOCATION.Name);
            //tblNextItems.Columns.Add(AlternateItemDataSet.Columns.QUANTITY.Name, typeof(double));
            //tblNextItems.Columns.Add(AlternateItemDataSet.Columns.ITEMCODE.Name);
            //tblNextItems.Columns.Add(AlternateItemDataSet.Columns.BESTBEFOREDATE.Name, typeof(DateTime));
            //tblNextItems.Columns.Add(AlternateItemDataSet.Columns.BATCHNUMBER.Name);
            //tblNextItems.Columns.Add(AlternateItemDataSet.Columns.UOM.Name);
            //tblNextItems.Columns.Add(AlternateItemDataSet.Columns.ITEM_DESCRIPTION.Name);
            //tblNextItems.Columns.Add(AlternateItemDataSet.Columns.ITEM_BARCODE.Name);
            //tblNextItems.Columns.Add(AlternateItemDataSet.Columns.SSCC.Name);
            //tblNextItems.Columns.Add(AlternateItemDataSet.Columns.UOM2.Name);
            //tblNextItems.Columns.Add(AlternateItemDataSet.Columns.QUANTITY_UOM2.Name, typeof(double));
            //tblNextItems.Columns.Add(AlternateItemDataSet.Columns.BATCHNUMBER2.Name);

            if (pickingItemInfoList != null && pickingItemInfoList.Count > 0)
            {
                //for (int i = 0; i < pickingItemInfoList.Count; i++)
                //{
                //    DataRow row = tblNextItems.Rows.Add(i,
                //        pickingItemInfoList[i].LocationCode,
                //        pickingItemInfoList[i].Quantity,
                //        pickingItemInfoList[i].ItemCode,
                //        pickingItemInfoList[i].BestBeforeDate,
                //        pickingItemInfoList[i].BatchNumber,
                //        pickingItemInfoList[i].Uom,
                //        pickingItemInfoList[i].ItemDescription,
                //        pickingItemInfoList[i].ItemBarcode,
                //        pickingItemInfoList[i].IsFullMasterSSCC ? pickingItemInfoList[i].MasterSSCC : pickingItemInfoList[i].IsFullPallet ? pickingItemInfoList[i].SSCC : "",
                //        pickingItemInfoList[i].Uom2,
                //        pickingItemInfoList[i].QuantityUom2,
                //        pickingItemInfoList[i].BatchNumber2
                //    );
                //}

                session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.Picking.ISelectAlternateItemScreen),
                    this.DefaultCultureInfo, BuildParamCollection(
                        "InitialErrorKey", initialErrorKey,
                        "TitleKey", allowToSelectItemToPick ? "MSG_TITLE_SELECT_ITEM_TO_PICK" : "MSG_TITLE_ITEMS_STILL_TO_PICK",
                        "DSAlternateItem", dsNextItems));
                msg = WaitForMessage();
                initialErrorKey = null;

                if (msg.Name == "Produmex.Foundation.SlimScreen.Interfaces.Picking.ISelectAlternateItemScreenReply.BackRequested")
                {
                    goBack = true;
                    if (zonePicking)
                    {
                        goBackZonePicking = true;
                        goto Step_ShowNextItems;
                    }
                    goto Step_LogStopWave;
                    //goto Step_SelectWave;
                }
                if (msg.Name != "Produmex.Foundation.SlimScreen.Interfaces.Picking.ISelectAlternateItemScreenReply.AlternateItemSelected")
                {
                    initialErrorKey = KnownMessageKeys.MSG_ERROR_OPERATION_NOT_SUPPORTED;
                    goto Step_SelectWave;
                }

                //get selected item
                if (allowToSelectItemToPick)
                {
					needToSelectItems = false;
                    selectedItemToPickIndex = ExtractParameter<int>(msg.Parameters, "internalKey");
                }
                else
                {
                    //always take first one
                    selectedItemToPickIndex = 0;
                }

            }

            #endregion

        Step_GetPickingItemInfo:  // This step was originally in front of Step_GetItemInfo

            #region GET PICKING ITEM INFO


            if (allowToSelectItemToPick &&
                needToSelectItems)
            {
                needToSelectItems = false;
                goto Step_ShowNextItems;
            }

            if (allowToSelectItemToPick)
            {
                //set option to true, so user can select item next time.
                needToSelectItems = true;
            }

            PickingItemInfo proposedPickingItemInfo = null;

            m_subFlowListOfSerialNumbers = null;
            listOfSerialNumbers = null;

            if (allowToSelectItemToPick)
            {
                if (pickingItemInfoList != null &&
                    pickingItemInfoList.Count > selectedItemToPickIndex)
                {
                    proposedPickingItemInfo = pickingItemInfoList[selectedItemToPickIndex];
                }
            }
            else
            {
                proposedPickingItemInfo = ExecuteMethodWithDbConnection<PickingItemInfo>(out pmxError, false, false,
                     delegate(PmxDbConnection conn, object[] parameters)
                     {
                         PmxPickListProvider plProv = new PmxPickListProvider(conn);
                         return plProv.GetNextItem(waveKey, alternateZonePickingForNoZones ? "NO_ZONE_ALTERNATE_PICKING" : zoneCode, consolidateItems, m_availableItemPickType1, m_availableItemPickType2);
                     });
            }
            if (pmxError != null)
            {
                session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
                    this.DefaultCultureInfo, BuildParamCollection(
                    "PmxError", pmxError,
                    "ShowButton", true));
                msg = WaitForMessage();

                goto Step_SelectWave;
            }
            if (proposedPickingItemInfo == null)
            {
                if (pickListDocEntry == 0 && this.CrossDockingPickListDocEntry != null)
                {
                    // Assign pickListDocEntry in case of no items needed to be picked and this has never been assigned
                    pickListDocEntry = this.CrossDockingPickListDocEntry;
                }
                goto Step_IsPickListPicked;
            }

            pickListDocEntry = proposedPickingItemInfo.PickListDocEntry;  // Keep Docentry for Step_IsPickListPicked
            pickingItemInfo = RuntimeHelper.DeepCopy<PickingItemInfo>(proposedPickingItemInfo);  // Keep a copy to start from when user selects alternative item

            //Get supplier pallet number
            query = BuildQuery.GetSupplierPalletNumberForPickList(pickingItemInfo.PickListDocEntry, pickingItemInfo.PickListLineID, m_sboProviderService.GetDbTool());
            dsSupplierPalletNumber = m_sboProviderService.RunView(true, null, null, query);

            supplierPalletNumber = null;

            if (this.GetNrOfRows(dsSupplierPalletNumber) > 0)
            {
                supplierPalletNumber = dsSupplierPalletNumber.Tables[0].Rows[0][Produmex.Sbo.Logex.Data.BusinessObjects.Definitions.Tables.PmxDocumentLineColumns.SupplierPalletNumber.NAME] as string;
            }

            if (zonePicking &&
                !lockWaveByWaveZone)
            {
                //lock the wave item
                ExecuteMethodWithDbConnection<bool>(out pmxError, false, false,
                    delegate(PmxDbConnection conn, object[] parameters)
                    {
                        PmxWaveLockProvider waveLockProv = new PmxWaveLockProvider(conn);
                        waveLockProv.LockWaveItem(waveKey, pickingItemInfo);

                        return true;
                    });


                if (pmxError != null)
                {
                    if (pmxError.ErrorCode == Produmex.Sbo.Logex.Data.ProdumexErrorCodes.WaveItemAlreadyLocked)
                    {
                        session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
                            this.DefaultCultureInfo, BuildParamCollection(
                            "MessageKey", "MSG_ERROR_WAVE_ITEM_ALREADY_LOCKED",
                            "ShowButton", true
                            ));
                    }
                    else
                    {
                        session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
                            this.DefaultCultureInfo, BuildParamCollection(
                            "PmxError", pmxError,
                            "ShowButton", true
                            ));
                    }
                    msg = WaitForMessage();

                    //select another one
                    goto Step_ShowNextItems;
                }
            }

            //get settings

            alternateCopyBatchNumbers = m_sboProviderService.InvokeMethodWithDbConnection<bool>(false, false, null, null,
                delegate(PmxDbConnection conn, object[] parameters)
                {
                    bool copy = copyBatchNumberOnAlternate;

                    //The setting is not set, so we now chek if batches where set on the the order
                    query = Produmex.Sbo.Logex.Data.SboCommands.GetQueryGetBatchNumbersForPickList(pickingItemInfo.PickListDocEntry, pickingItemInfo.PickListLineID, conn.DbCultureInfo);

                    using (ISboRecordset rs = SboRecordsetHelper.RunQuery(s_log, query, conn))
                    {
                        copy = !rs.EoF;
                    }

                    if (pickingItemInfo.ListOfSerialNumbers != null &&
                            pickingItemInfo.ListOfSerialNumbers.Count > 0)
                    {
                        //linked serial numbers, so we need to keep batchnumbers!
                        copy = true;
                    }

                    return copy;
                });


            bool alternateItemMode = false;

            #endregion

        Step_WorkWithPickingItemInfo:  // Re-entry point for picking of alternate item

            List<KeyValuePair<double, double>> weightDetails = null;

        Step_IdentifyMoveableLocation:

            #region IDENTIFY MOVEABLE LOCATION

            userEnteredSSCCToPickOn = false;
            ssccToPickOnHasStock = true;

            if ( automaticSelectionOfMoveableLocation && pickingItemInfo.IsFullPallet )
            {
                targetMoveableLocation = "";
                goto Step_IdentifyLogisticCarrier;
            }
            
            if (multiPicking)
            {
                targetMoveableLocation = multiPickingPickListInfos[pickingItemInfo.PickListDocEntry].MovableLocation;
            }

            if (targetMoveableLocation != null ||
                targetLuid.HasValue || 
                targetMasterLuid.HasValue)
            {
                // MoveableLocation already chosen or multi-picking, skip this step
                goto Step_IdentifyLogisticCarrier;
            }


            query = BuildQuery.GetMovableLocations(null, pmxWarehouseCode);
            DataSet dsMoveableLocations = m_sboProviderService.RunView(false, null, null, query);

            if ( GetNrOfRows( dsMoveableLocations ) == 1 && automaticSelectionOfMoveableLocation && !pickingItemInfo.IsFullPallet && !isProductionPicking )
            {
                targetMoveableLocation = dsMoveableLocations.Tables[ 0 ].Rows[ 0 ][ LocationDataSet.Columns.CODE.Name ].ToString();
                goto Step_IdentifyLogisticCarrier;
            }

            if (GetNrOfRows(dsMoveableLocations) < 1 && (waveSameDelivery || allowMultipleCustomersOnSSCC))
            {
				if (allowToIdentifySSCCToPickOn)
				{
					tempStringList = new List<string>();
					tempStringList.Add("MSG_BUTTON_ENTER_SSCC");
					tempStringList.Add("MSG_BUTTON_NEW_SSCC");

					 session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.ISelectButtonScreen),
                            DefaultCultureInfo.Get(), BuildParamCollection(
                            "InitialErrorKey", initialErrorKey,
							"TitleKey", "MSG_TITLE_SELECT_TASK",
							"AllowedButtons", tempStringList.ToArray()
                            ));
                    msg = WaitForMessage();
                    initialErrorKey = null;
					if (msg.Name.EndsWith(".ButtonClicked"))
					{
						string filterButtonKey = ExtractParameter<string>(msg.Parameters, "buttonKey");
						if (filterButtonKey == "MSG_BUTTON_ENTER_SSCC")
						{
							targetMoveableLocation = null;
							goto Step_EnterSSCCToPickOn;
						}
                        targetLuid = null;
					}
					else
					{
						//go back
						goto Step_ShowNextItems;
					}
				}

                // No moveable locations exist and it is not required, skip this step
                goto Step_IdentifyLogisticCarrier;
            }

            tempStringList = new List<string>();
            tempStringList.Add("MSG_BUTTON_SELECT_MOVEABLE_LOCATION");

            string[] allowedButtons;

            if (waveSameDelivery || alwaysSetToStatusPicked || allowMultipleCustomersOnSSCC)
            {
                tempStringList.Add("MSG_BUTTON_NO_MOVEABLE_LOCATION");

                if (allowToIdentifySSCCToPickOn)
                {
                    tempStringList.Add("MSG_BUTTON_ENTER_SSCC");
                }
            }

            allowedButtons = tempStringList.ToArray();



            session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.MoveOrder.IIdentifyLocationScreen),
                this.DefaultCultureInfo, BuildParamCollection(
                "InitialErrorKey", initialErrorKey,
                "TitleKey", "MSG_TITLE_IDENTIFY_MOVEABLE_LOCATION",
                "HideDefaultLocation", true,
                "AllowedButtons", allowedButtons));
            msg = WaitForMessage();
            initialErrorKey = null;

            if (msg.Name == "Produmex.Foundation.SlimScreen.Interfaces.MoveOrder.IIdentifyLocationScreenReply.LocationEntered")
            {
                targetMoveableLocation = ExtractParameter<string>(msg.Parameters, "locationCode");
                if (GetNrOfRows(dsMoveableLocations) > 0)
                {
                    DataRow[] rows = dsMoveableLocations.Tables[0].Select(LocationDataSet.Columns.CODE.Name + " = '" + targetMoveableLocation + "'");

                    if (rows.Length > 0)
                    {
                        //get code from database: 
                        targetMoveableLocation = rows[0][LocationDataSet.Columns.CODE.Name] as string;
                        goto Step_IdentifyLogisticCarrier;
                    }
                }
                // initialErrorKey = "";  // Should be some error -> invalid location?
                targetMoveableLocation = null;  // Failure on this step
                goto Step_IdentifyMoveableLocation;
            }
            if (msg.Name == "Produmex.Foundation.SlimScreen.Interfaces.MoveOrder.IIdentifyLocationScreenReply.ButtonClicked")
            {
                string buttonKey = ExtractParameter<string>(msg.Parameters, "buttonKey");
                if (buttonKey == "MSG_BUTTON_SELECT_MOVEABLE_LOCATION")
                {
                    goto Step_SelectMoveableLocation;
                }
                if (buttonKey == "MSG_BUTTON_NO_MOVEABLE_LOCATION")
                {
                    targetMoveableLocation = "";
                    goto Step_IdentifyLogisticCarrier;
                }
                if (buttonKey == "MSG_BUTTON_ENTER_SSCC")
                {
                    targetMoveableLocation = null;
                    goto Step_EnterSSCCToPickOn;
                }
            }

            // Back/Stop/Non-supported operation
            goto Step_ShowNextItems;

            #endregion

        Step_SelectMoveableLocation:

            #region SELECT MOVEABLE LOCATION

            createMasterSSCC = false;

            session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.ISelectLocationScreen),
                this.DefaultCultureInfo, BuildParamCollection(
                "InitialErrorKey", initialErrorKey,
                "TitleKey", "MSG_TITLE_SELECT_MOVEABLE_LOCATION",
                "LocationDS", dsMoveableLocations));
            msg = WaitForMessage();
            initialErrorKey = null;

            if (msg.Name == "Produmex.Foundation.SlimScreen.Interfaces.ISelectLocationScreenReply.LocationSelected")
            {
                targetMoveableLocation = ExtractParameter<string>(msg.Parameters, "locationCode");
                goto Step_IdentifyLogisticCarrier;
            }
            if (msg.Name != "Produmex.Foundation.SlimScreen.Interfaces.ISelectLocationScreenReply.ButtonClicked")
            {
                // Back/Stop/Non-supported operation
                goto Step_IdentifyMoveableLocation;
            }

            #endregion

        Step_EnterSSCCToPickOn:

            #region ENTER SSCC TO PICK ON

            ssccToPickOnHasStock = true;


            session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.Reception.IIdentifySSCCScreen),
                    DefaultCultureInfo.Get(), BuildParamCollection(
                    "InitialErrorKey", initialErrorKey,
                    "TitleKey", "MSG_TITLE_IDENTIFY_SSCC_TO_PICK_ON",
					"BarcodeController", barcodeController,
					"ButtonNoSSCC", false));
            msg = WaitForMessage();

            //clear data
            initialErrorKey = null;

            if (msg.Name.EndsWith(".SSCCEntered"))
            {
                string inputSSCC = ExtractParameter<string>(msg.Parameters, "sscc");

                targetLogCarrierStorLocCode = null;
                targetLogCarrierCanBeFull = false;

                //SSCC is valid, so we check if we can find the LUID
                if (!m_sboProviderService.InvokeMethodWithDbConnection<bool>(false, false, null, null,
                   delegate(PmxDbConnection conn, object[] parameters)
                   {
                       PmxSSCCProvider ssccProv = new PmxSSCCProvider(conn);

                       return ssccProv.IsValidChecksumSSCC(inputSSCC);
                   }))
                {
                    initialErrorKey = "MSG_ERROR_INVALID_SSCC";
                    goto Step_EnterSSCCToPickOn;
                }
                else
                {
                    //SSCC is valid, so we check if we can find the LUID
                    if (!m_sboProviderService.InvokeMethodWithDbConnection<bool>(false, false, null, null,
                    delegate(PmxDbConnection conn, object[] parameters)
                    {
                        PmxLogisticUnitIDProvider luidProv = new PmxLogisticUnitIDProvider(conn);

                        int tempLuid = luidProv.GetLUIDBySSCC(inputSSCC, true);

                        if (tempLuid == 0)
                        {
                            //No Luid found in stock, so check if we find it in LUID table
                            tempLuid = luidProv.GetLUIDBySSCC(inputSSCC, false);
                            if (tempLuid == 0)
                            {
                                //no luid found, so create a new one
                                targetLuid = luidProv.GetCreateLUID(inputSSCC, null);
                            }
                            else
                            {
                                //Check if the LUID exists in the master LUID table
                                PmxLinkMasterLUIDProvider lmluProv = new PmxLinkMasterLUIDProvider(conn);

                                CollectionWithDispose<PmxLinkMasterLUID>  links = lmluProv.GetLink(tempLuid);

                                if (links.Count > 0)
                                {
                                    targetMasterLuid = tempLuid;
                                }
                                else
                                {
                                    //take luid from LUID table
                                    targetLuid = tempLuid;
                                }
                            }
                            //no stock, so allow to select logistic carrier
                            ssccToPickOnHasStock = false;
                        }
                        else
                        {
                            //SSCC found in stock, so check if SSCC is allowed
                            //Only SSCC with no stock or SSCC with stock, but only picked items for the current wave (taking in account the multi customer sscc setting)

                            //Check if we can find it in the logistic units for customer
                            query = GetQueryForLogisticUnitsForCustomer(pickingItemInfoList, waveSameDelivery, waveKey, allowMultipleCustomersOnSSCC, tempLuid, pmxWarehouseCode);
                            using (ISboRecordset rs = SboRecordsetHelper.RunQuery(s_log, query, conn))
                            {
                                if (rs.EoF)
                                {
                                    //SSCC not allowed
                                    return false;
                                }

                                targetLuid = tempLuid;
                                targetLogCarrierCanBeFull = true;
                                targetLogCarrierStorLocCode = rs.GetTypedValue<string>(LogisticCarrierDataSet.Columns.STORAGE_LOCATION_CODE.Name);

                                DbCommand dbCmd = conn.CreateCommand(BuildQuery.SetLogUnitReadyForShipping(tempLuid, false, conn.DbTool));
                                dbCmd.ExecuteNonQuery();
                            }
                        }

                        return true;

                    }))
                    {
                        initialErrorKey = "MSG_ERROR_INVALID_SSCC";
                        goto Step_EnterSSCCToPickOn;
                    }

                    //SSCC is good to use
                    userEnteredSSCCToPickOn = true;
                    goto Step_IdentifyLogisticCarrier;
                }

            }
            else if (msg.Name.EndsWith(".NoSSCC"))
            {
                //not allowed
                goto Step_EnterSSCCToPickOn;
            }
            else if (msg.Name.EndsWith(".BackRequested"))
            {
                goto Step_IdentifyMoveableLocation;
            }

            goto Step_IdentifyMoveableLocation;




        #endregion

        Step_IdentifyLogisticCarrier:

            #region IDENTIFY LOGISTIC CARRIER

            bool searchedForLUIDOfCustomer = false;
            DataSet dsLogisticUnitsForCustomer = null;
            //targetMasterLuid = null;

            if (String.IsNullOrEmpty(targetMoveableLocation) &&
                allowToCreateMasterSSCC &&
                !pickingItemInfo.IsFullMasterSSCC &&
                !targetMasterLuid.HasValue)
            {

                searchedForLUIDOfCustomer = true;

                query = GetQueryForLogisticUnitsForCustomer(pickingItemInfoList, waveSameDelivery, waveKey, allowMultipleCustomersOnSSCC, null, pmxWarehouseCode);

                dsLogisticUnitsForCustomer = m_sboProviderService.RunView(false, null, null, query);

                int? tempLuid = null;
                int? tempMasterLuid = null;

                // string ssccSelected = null;  // Not used
                if (GetNrOfRows(dsLogisticUnitsForCustomer) > 0)
                {
                    session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.Picking.ISelectLogisticCarrierScreen),
                        this.DefaultCultureInfo, BuildParamCollection(
                        "InitialErrorKey", initialErrorKey,
                        "TitleKey", "MSG_TITLE_SELECT_CUSTOMER_PALLET",
                        "BarcodeController", barcodeController,
                        "LogisticCarrierDS", dsLogisticUnitsForCustomer,
                        "ButtonKey", "MSG_BUTTON_NEW_LOGISTIC_UNIT"));
                    msg = WaitForMessage();
                    initialErrorKey = null;

                    if (msg.Name == "Produmex.Foundation.SlimScreen.Interfaces.Picking.ISelectLogisticCarrierScreenReply.LogisticCarrierSelected")
                    {
                        tempLuid = ExtractParameter<int>(msg.Parameters, "luid");
                        tempMasterLuid = ExtractParameter<int?>(msg.Parameters, "masterLuid");

                        DataRow[] rows = dsLogisticUnitsForCustomer.Tables[0].Select(LogisticCarrierDataSet.Columns.LUID.Name + " = " + tempLuid.ToString());
                        if (rows.Length > 0)
                        {
                            targetLogCarrierStorLocCode = rows[0][LogisticCarrierDataSet.Columns.STORAGE_LOCATION_CODE.Name] as string;
                        }

                        if (allowToCreateMasterSSCC && tempMasterLuid.HasValue)
                        {
                            if (pickingItemInfo.IsFullPallet)
                            {
                                //Next item is a full pallet
                                //This means we cannot use a sub SSCC, because we keep the SSCC
                                targetMasterLuid = tempMasterLuid;
                                tempLuid = null;
                            }
                            else
                            {
                                //We allow to create master SSCC and selected SSCC is a master, so ask user what to do:
                                //Use master, use LUID, or use both
                                buttonKeys = new string[3] { "MSG_BUTTON_USE_MASTER_LUID", "MSG_BUTTON_USE_SUB_SSCC", "MSG_BUTTON_USE_MASTER_AND_SUB_SSCC" };

                                session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.ISelectButtonScreen),
                                        DefaultCultureInfo.Get(), BuildParamCollection(
                                        "InitialErrorKey", initialErrorKey,
                                        "TitleKey", "MSG_TITLE_SELECT_TASK",
                                        "AllowedButtons", buttonKeys,
                                        "EnabledToolBarButtons", ToolBarButtons.Back
                                        ));
                                msg = WaitForMessage();
                                initialErrorKey = null;
                                if (msg.Name.EndsWith(".ButtonClicked"))
                                {
                                    string filterButtonKey = ExtractParameter<string>(msg.Parameters, "buttonKey");

                                    if (filterButtonKey == "MSG_BUTTON_USE_MASTER_LUID" ||
                                       filterButtonKey == "MSG_BUTTON_USE_MASTER_AND_SUB_SSCC")
                                    {
                                        targetMasterLuid = tempMasterLuid;
                                        if (filterButtonKey == "MSG_BUTTON_USE_MASTER_LUID")
                                        {
                                            tempLuid = null;
                                        }
                                    }
                                }
                                else
                                {
                                    //go back
                                    goto Step_SelectWave;
                                }
                            }
                        }



                        targetLuid = tempLuid;
                        targetLogCarrierCanBeFull = true;

                        m_sboProviderService.InvokeMethodWithDbConnection<object>(false, false, null, null, delegate (PmxDbConnection dbConn, object[] parameters)
                        {
                            DbCommand dbCmd;
                            if (targetMasterLuid.HasValue)
                            {
                                dbCmd = dbConn.CreateCommand(BuildQuery.SetMasterLogUnitReadyForShipping(targetMasterLuid.Value, false, dbConn.DbTool));
                            }
                            else
                            {
                                dbCmd = dbConn.CreateCommand(BuildQuery.SetLogUnitReadyForShipping(targetLuid.Value, false, dbConn.DbTool));
                            }
                            dbCmd.ExecuteNonQuery();

                            return null;
                        });

                        goto Step_IdentifyLogisticCarrierAfterMasterSSCC;
                    }
                    else if (msg.Name == "Produmex.Foundation.SlimScreen.Interfaces.Picking.ISelectLogisticCarrierScreenReply.ButtonClicked")
                    {
                        // Continue with code below if-else's
                        targetLuid = null;
                    }
                    else if (msg.Name == "Produmex.Foundation.SlimScreen.Interfaces.ISelectLogisticCarrierScreenReply.BackRequested")
                    {
                        goto Step_SelectWave;
                    }
                    else if (msg.Name == "Produmex.Foundation.SlimScreen.Interfaces.ISelectLogisticCarrierScreenReply.StopRequested")
                    {
                        return;
                    }
                    else
                    {
                        //go back
                        goto Step_SelectWave;
                    }
                }


                //Allowing to create a master SSCC, so we need to ask what the user wants to do
                buttonKeys = new string[2] { "MSG_BUTTON_CREATE_MASTER_LUID", "MSG_BUTTON_CREATE_NORMAL_SSCC" };

                ToolBarButtons toolBarButtons = ToolBarButtons.Back;

                session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.ISelectButtonScreen),
                        DefaultCultureInfo.Get(), BuildParamCollection(
                        "InitialErrorKey", initialErrorKey,
                        "TitleKey", "MSG_TITLE_SELECT_TASK",
                        "AllowedButtons", buttonKeys,
                        "EnabledToolBarButtons", toolBarButtons
                        ));
                msg = WaitForMessage();
                initialErrorKey = null;
                if (msg.Name.EndsWith(".ButtonClicked"))
                {
                    string filterButtonKey = ExtractParameter<string>(msg.Parameters, "buttonKey");

                    if (filterButtonKey == "MSG_BUTTON_CREATE_MASTER_LUID")
                    {
                        if (userEnteredSSCCToPickOn &&
                            !ssccToPickOnHasStock &&
                            !targetMasterLuid.HasValue)
                        {
                            //Scanned SSCC is a master SSCC
                            targetMasterLuid = targetLuid;
                            targetLuid = null;
                        }
                        createMasterSSCC = true;
                    }

                }
                else
                {
                    //go back
                    goto Step_SelectWave;
                }
            }

        Step_IdentifyLogisticCarrierAfterMasterSSCC:

            if (String.IsNullOrEmpty(targetMoveableLocation) &&
                !pickingItemInfo.IsFullPallet &&
                (!targetLuid.HasValue || (targetLuid.HasValue && !ssccToPickOnHasStock))
                && targetLogCarrierCode == null &&
                !alternateItemMode &&
                waveSameDelivery)
            {
                
                if (!userEnteredSSCCToPickOn &&
                    !searchedForLUIDOfCustomer)
                {
                    query = GetQueryForLogisticUnitsForCustomer(pickingItemInfoList, waveSameDelivery, waveKey, allowMultipleCustomersOnSSCC, null, pmxWarehouseCode);

                    dsLogisticUnitsForCustomer = m_sboProviderService.RunView(false, null, null, query);

                    // string ssccSelected = null;  // Not used
                    if (GetNrOfRows(dsLogisticUnitsForCustomer) > 0)
                    {
                        session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.Picking.ISelectLogisticCarrierScreen),
                            this.DefaultCultureInfo, BuildParamCollection(
                            "InitialErrorKey", initialErrorKey,
                            "TitleKey", "MSG_TITLE_SELECT_CUSTOMER_PALLET",
							"BarcodeController", barcodeController,
							"LogisticCarrierDS", dsLogisticUnitsForCustomer,
                            "ButtonKey", "MSG_BUTTON_NEW_LOGISTIC_UNIT"));
                        msg = WaitForMessage();
                        initialErrorKey = null;

                        if (msg.Name == "Produmex.Foundation.SlimScreen.Interfaces.Picking.ISelectLogisticCarrierScreenReply.LogisticCarrierSelected")
                        {
                            targetLuid = ExtractParameter<int>(msg.Parameters, "luid");
                            // ssccSelected = ExtractParameter<string>( msg.Parameters, "sscc" );
                            targetLogCarrierCanBeFull = true;

                            DataRow[] rows = dsLogisticUnitsForCustomer.Tables[0].Select(LogisticCarrierDataSet.Columns.LUID.Name + " = " + targetLuid.ToString());
                            if (rows.Length > 0)
                            {
                                targetLogCarrierStorLocCode = rows[0][LogisticCarrierDataSet.Columns.STORAGE_LOCATION_CODE.Name] as string;
                            }

                            m_sboProviderService.InvokeMethodWithDbConnection<object>(false, false, null, null, delegate(PmxDbConnection dbConn, object[] parameters)
                            {
                                DbCommand dbCmd = dbConn.CreateCommand(BuildQuery.SetLogUnitReadyForShipping(targetLuid.Value, false, dbConn.DbTool));
                                dbCmd.ExecuteNonQuery();

                                return null;
                            });
                        }
                        else if (msg.Name == "Produmex.Foundation.SlimScreen.Interfaces.Picking.ISelectLogisticCarrierScreenReply.ButtonClicked")
                        {
                            // Continue with code below if-else's
                            targetLuid = null;
                        }
                        else if (msg.Name == "Produmex.Foundation.SlimScreen.Interfaces.ISelectLogisticCarrierScreenReply.BackRequested")
                        {
                            goto Step_SelectWave;
                        }
                        else if (msg.Name == "Produmex.Foundation.SlimScreen.Interfaces.ISelectLogisticCarrierScreenReply.StopRequested")
                        {
                            return;
                        }
                        else
                        {
                            //go back
                            goto Step_SelectWave;
                            //initialErrorKey = KnownMessageKeys.MSG_ERROR_OPERATION_NOT_SUPPORTED;
                            //goto Step_IdentifyLogisticCarrier;
                        }
                    }
                }

                if (!targetLuid.HasValue ||
                    (userEnteredSSCCToPickOn && !ssccToPickOnHasStock))
                {
                    query = BuildQuery.GetLogisticCarriersOrderByName(true, true, m_sboProviderService.GetDbTool());
                    dsLogisticUnitsForCustomer = m_sboProviderService.RunView(false, null, null, query);
                    if (GetNrOfRows(dsLogisticUnitsForCustomer) > 0)
                    {
                        session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.ISelectProductScreen),
                            this.DefaultCultureInfo, BuildParamCollection(
                            "InitialErrorKey", initialErrorKey,
                            "TitleKey", "MSG_TITLE_SELECT_LOGISTIC_CARRIER",
                            "ProductDS", dsLogisticUnitsForCustomer,
                            "AllowNoSelection", false,
                            "ButtonKey", "MSG_BUTTON_NO_LOGISTIC_CARRIER"));
                        msg = WaitForMessage();
                        initialErrorKey = null;

                        if (msg.Name == "Produmex.Foundation.SlimScreen.Interfaces.ISelectProductScreenReply.ProductSelected")
                        {
                            targetLogCarrierCode = ExtractParameter<string>(msg.Parameters, "itemCode");
                        }
                        else if (msg.Name == "Produmex.Foundation.SlimScreen.Interfaces.ISelectProductScreenReply.ButtonClicked")
                        {
                            targetLogCarrierCode = "";  // Remember we chose MSG_BUTTON_NO_LOGISTIC_CARRIER
                        }
                        else if (msg.Name == "Produmex.Foundation.SlimScreen.Interfaces.ISelectProductScreenReply.BackRequested")
                        {
                            goto Step_SelectWave;
                        }
                        else if (msg.Name == "Produmex.Foundation.SlimScreen.Interfaces.ISelectProductScreenReply.StopRequested")
                        {
                            return;
                        }
                        else
                        {
                            initialErrorKey = KnownMessageKeys.MSG_ERROR_OPERATION_NOT_SUPPORTED;
                            goto Step_IdentifyLogisticCarrier;
                        }
                    }
                }
            }

            #endregion

        Step_GetItemInfo:

            #region GET ITEM INFO

            masterLUIDFull = false;

            PmxItemInfo itemInfo = ExecuteMethodWithDbConnection<PmxItemInfo>(out pmxError, false, false,
                delegate(PmxDbConnection conn, object[] parameters)
                {
                    PmxItemAllConnectionsProvider itemProv = new PmxItemAllConnectionsProvider(conn);
                    return itemProv.GetItemInfoWithBaseComponent(pickingItemInfo.ItemCode, 1);
                });

            #endregion

        Step_PickItem:

            #region PICK ITEM

            if (adHocAlternatePicking && alternateItemMode)
            {
                //we are in alternate mode, and use ad hoc alternate picking, so we let the user scan the data he wants.
                goto Step_Alternate_IdentifySSCCOrLocation;
            }

            if (pickingItemInfo.IsFullMasterSSCC && pickingItemInfo.IsConsolidated && !userEnteredSSCCToPickOn)
            {
                goto Step_PickFullMasterSSCC;
            }

            if (pickingItemInfo.IsFullPallet && !userEnteredSSCCToPickOn)
            {
                goto Step_PickFullPallet;
            }

            Step_PickNotFullPallet:

            shownPickLocationScreen = false;


            if (forceToScanSSCCOnItemPick && pickingItemInfo.LogisticUnitIdentKey.HasValue)
            {

                buttonList = new List<string>();
             
                buttonList.Add("MSG_BUTTON_OTHER_TASK");

                session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.Picking.IIdentifyFullPalletPickingScreen),
                   this.DefaultCultureInfo, BuildParamCollection(
                   "InitialErrorKey", initialErrorKey,
                   "TitleKey", "MSG_TITLE_IDENTIFY_SSCC",
                   "SSCC", pickingItemInfo.SSCC,
                   "LocationCode", pickingItemInfo.LocationCode,
                   "ItemCode", pickingItemInfo.ItemCode,
                   "ItemDescription", pickingItemInfo.CustomItemDescription,
                   "ItemBarcode", pickingItemInfo.ItemBarcode,
                   "ImagePath", itemInfo.PictureName,
                   "StillToPick", pickingItemInfo.Quantity, 
                   "Uom", pickingItemInfo.Uom,
                   "BatchNumber", pickingItemInfo.BatchNumber,
                   "BestBeforeDate", pickingItemInfo.BestBeforeDate,
                   "ShowFullNextItemData", false,
                   "AllowedButtons", buttonList.ToArray()));
                msg = WaitForMessage();
                initialErrorKey = null;

                if (msg.Name == "Produmex.Foundation.SlimScreen.Interfaces.Picking.IIdentifyFullPalletPickingScreenReply.ButtonClicked")
                {
                    goto Step_OtherTask;
                }
                if (msg.Name != "Produmex.Foundation.SlimScreen.Interfaces.Picking.IIdentifyFullPalletPickingScreenReply.SSCCEntered")
                {
                    //Go back
                    goto Step_ShowNextItems;
                }

                //Check if SSCC is on valid location
                if (string.IsNullOrEmpty(pickingItemInfo.LocationCode))
                {
                    initialErrorKey = "MSG_ERROR_PICK_LOCATION_UNKNOWN_FOR_FULL_PALLET";
                    goto Step_PickNotFullPallet;
                }

                string enteredSSCCItemPick = ExtractParameter<string>(msg.Parameters, "sscc");
                if (enteredSSCCItemPick != pickingItemInfo.SSCC)
                {
                    initialErrorKey = "MSG_ERROR_INCORRECT_SSCC_TO_PICK";
                    goto Step_PickNotFullPallet;
                }

                enteredLocationCode = pickingItemInfo.LocationCode;

                goto Step_SelectLocation;

            }

            if (keepPickingFromTheSameLocation &&
                !string.IsNullOrEmpty(enteredLocationCode) &&
                enteredLocationCode == pickingItemInfo.LocationCode)
            {
				if (goBack)
				{
					goto Step_ShowNextItems;
				}
                //Same location as previous location, so do not ask again
                goto Step_SelectLocation;
            }

			shownPickLocationScreen = true;
            enteredLocationCode = null;

            if (!itemInfo.IsInventoryItem)
            {
                //Non inventory, so proceed to go to ener quantity
                goto Step_GetFreeQuantity;
            }

            buttonList = new List<string>();

            if (!string.IsNullOrEmpty(pickingItemInfo.LocationCode))
            {
                buttonList.Add("MSG_BUTTON_SELECT_LOCATION");
            }

            buttonList.Add("MSG_BUTTON_OTHER_TASK");

            session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.Picking.IIdentifyLocationScreen),
                this.DefaultCultureInfo, BuildParamCollection(
                "InitialErrorKey", initialErrorKey,
                "TitleKey", "MSG_TITLE_IDENTIFY_PICK_LOCATION",
                "ItemCode", pickingItemInfo.ItemCode,
                "ItemDescription", pickingItemInfo.CustomItemDescription,
                "ItemBarcode", pickingItemInfo.ItemBarcode,
                "ImagePath", itemInfo.PictureName,
                "StillToPick", pickingItemInfo.Quantity,  // take from proposedPickingItemInfo ?
                "TotalQuantityToPick", proposedPickingItemInfo.TotalQuantityToPick,
                "Uom", pickingItemInfo.Uom,
                "BatchNumber", pickingItemInfo.BatchNumber,
                "BestBeforeDate", pickingItemInfo.BestBeforeDate,
                "LocationCode", pickingItemInfo.LocationCode,
                "SSCC", pickingItemInfo.SSCC,
                "IsAlternate", (adHocAlternatePicking && alternateItemMode),
                "ButtonKeys", buttonList.ToArray()));
            msg = WaitForMessage();
            initialErrorKey = null;

            if (msg.Name == "Produmex.Foundation.SlimScreen.Interfaces.Picking.IIdentifyLocationScreenReply.LocationEntered")
            {
                enteredLocationCode = ExtractParameter<string>(msg.Parameters, "locationCode");
            }
            else if (msg.Name == "Produmex.Foundation.SlimScreen.Interfaces.Picking.IIdentifyLocationScreenReply.ButtonClicked")
            {
                string buttonKey = ExtractParameter<string>(msg.Parameters, "buttonKey");
                if (buttonKey == "MSG_BUTTON_OTHER_TASK")
                {
                    goto Step_OtherTask;
                }
            }
            else if (msg.Name == "Produmex.Foundation.SlimScreen.Interfaces.Picking.IIdentifyLocationScreenReply.BackRequested")
            {
                goto Step_ShowNextItems;
            }
            else if (msg.Name == "Produmex.Foundation.SlimScreen.Interfaces.Picking.IdentifyLocationScreenReply.StopRequested")
            {
                return;
            }
            else
            {
                initialErrorKey = KnownMessageKeys.MSG_ERROR_OPERATION_NOT_SUPPORTED;
                goto Step_PickNotFullPallet;
            }

            #endregion

        Step_SelectLocation:

            #region SELECT LOCATION

            if (String.IsNullOrEmpty(enteredLocationCode))
            {
                query = BuildQuery.GetPickLocation(pmxWarehouseCode);
                DataSet dsPickLocations = m_sboProviderService.RunView(false, null, null, query);
                session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.ISelectLocationScreen),
                    this.DefaultCultureInfo, BuildParamCollection(
                    "InitialErrorKey", initialErrorKey,
                    "TitleKey", "MSG_TITLE_SELECT_LOCATION",
                    "LocationDS", dsPickLocations));
                msg = WaitForMessage();
                initialErrorKey = null;

                if (msg.Name == "Produmex.Foundation.SlimScreen.Interfaces.ISelectLocationScreenReply.LocationSelected")
                {
                    enteredLocationCode = ExtractParameter<string>(msg.Parameters, "locationCode");
                }
                else if (msg.Name == "Produmex.Foundation.SlimScreen.Interfaces.ISelectLocationScreenReply.BackRequested")
                {
                    if (adHocAlternatePicking && alternateItemMode)
                    {
                        goto Step_Alternate_IdentifySSCCOrLocation;
                    }
                    else
                    {
                        goto Step_PickNotFullPallet;
                    }
                }
                else if (msg.Name == "Produmex.Foundation.SlimScreen.Interfaces.ISelectLocationScreenReply.StopRequested")
                {
                    return;
                }
                else if (msg.Name == "Produmex.Foundation.SlimScreen.Interfaces.ISelectLocationScreenReply.RefreshRequested")
                {
                    goto Step_SelectLocation;
                }
                else
                {
                    initialErrorKey = KnownMessageKeys.MSG_ERROR_OPERATION_NOT_SUPPORTED;
                    goto Step_SelectLocation;
                }
            }

            #endregion

        Step_CheckLocation:

            #region CHECK LOCATION

            if (adHocAlternatePicking && alternateItemMode)
            {
                //user can pick from any location
            }
            else
            {

                if (enteredLocationCode.ToUpper() != pickingItemInfo.LocationCode.ToUpper())
                {
                    initialErrorKey = "MSG_ERROR_INCORRECT_PICK_LOCATION";
                    goto Step_PickNotFullPallet;
                }

                //make sure uppercase/lowercase is not a problem
                enteredLocationCode = pickingItemInfo.LocationCode;
            }

            #endregion

        Step_IdentifyBaseComponent:

            #region IDENTIFY BASE COMPONENT

            string enteredItemCode = null;
            string enteredItemDescr = null;
            GS1Data gs1DataScanned = null;
            bcScanned = null;

            if (itemInfo.ScanBaseComponent)
            {
                session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
                    this.DefaultCultureInfo, BuildParamCollection(
                    "InitialErrorKey", initialErrorKey,
                    "MessageKey", "MSG_IDENTIFY_BASE_COMPONENT",
                    "ShowButton", true));
                msg = WaitForMessage();
                initialErrorKey = null;


                BarcodeContainer baseComponent = null;
                string gtinBaseComponent = null;
                string scannedItemCodeBaseComponent = null;

            Step_IdentifyBaseComponent_ReEntry:

                //session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IIdentifyGS1DataScreen),
                //    DefaultCultureInfo.Get(), BuildParamCollection(
                //    "InitialErrorKey", initialErrorKey,
                //    "TitleKey", "MSG_TITLE_IDENTIFY_BASE_PRODUCT",
                //    "BarcodeController", barcodeController,
                //    "AllowedButtons", null,
                //    "ForceGS1_128", false
                //    ));

                session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.Picking.IIdentifyBarcodePickingScreen),
                    this.DefaultCultureInfo, BuildParamCollection(
                    "InitialErrorKey", initialErrorKey,
                    "TitleKey", "MSG_TITLE_IDENTIFY_BASE_PRODUCT",
					"BarcodeController", barcodeController,
                    "AllowedButtons", disableSelectionOfItems || PmxOseCompany.Get().DisableItemSelectionInFlows ? null : new string[] { "MSG_BUTTON_SELECT_PRODUCT" },
                    "ItemCode", itemInfo.BaseItemInfo.ItemCode,
                    "ItemDescription", itemInfo.BaseItemInfo.Description,
                    "ItemBarcode", itemInfo.BaseItemInfo.CodeBars,
                    "ImagePath", itemInfo.PictureName,
                    "StillToPick", pickingItemInfo.Quantity,
                    "TotalQuantityToPick", pickingItemInfo.TotalQuantityToPick,
                    "Uom", pickingItemInfo.Uom,
                    "BatchNumber", (adHocAlternatePicking && alternateItemMode) ? "" : pickingItemInfo.BatchNumber,
                    "BestBeforeDate", (adHocAlternatePicking && alternateItemMode) ? null : pickingItemInfo.BestBeforeDate,
                    "LocationCode", (adHocAlternatePicking && alternateItemMode) ? enteredLocationCode : pickingItemInfo.LocationCode,
                    "SSCC", (adHocAlternatePicking && alternateItemMode) ? alternateSSCC : pickingItemInfo.SSCC,
                    "IsAlternate", (adHocAlternatePicking && alternateItemMode)
                    ));
                msg = WaitForMessage();
                initialErrorKey = null;

                if (msg.Name == "Produmex.Foundation.SlimScreen.Interfaces.Picking.IIdentifyBarcodePickingScreenReply.DataEntered")
                {
                    try
                    {
                        string scannedValue = ExtractParameter<string>(msg.Parameters, "data");

                        baseComponent = barcodeController.GetScannedData(scannedValue, false, baseComponent);

                        if (baseComponent == null ||
                            (string.IsNullOrEmpty(BarcodeHelper.GetProductIdentification(baseComponent)) && string.IsNullOrEmpty(baseComponent.ItemCode)))
                        {
                            initialErrorKey = "MSG_ERROR_NO_PRODUCT_ON_BARCODE";
                            goto Step_IdentifyBaseComponent_ReEntry;
                        }

                        gtinBaseComponent = BarcodeHelper.GetProductIdentification(baseComponent);
                        scannedItemCodeBaseComponent = baseComponent.ItemCode;
                    }
                    catch (Exception ex)
                    {
                        s_log.Error("", ex);
                        if (!(ex is ProdumexException) || ((ProdumexException)ex).Error == null)
                        {
                            // Inform the user with the error
                            session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
                                DefaultCultureInfo.Get(), BuildParamCollection(
                                    "MessageKey", ExceptionHelper.Format(ExceptionHelper.Content.InnerExceptions, ex, DefaultCultureInfo.Get()),
                                    "NoTranslationOfMessageKey", true,
                                    "IsError", true,
                                    "ShowButton", true));
                            WaitForMessage();

                        }
                        else
                        {
                            // Inform the user with the error
                            session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
                                DefaultCultureInfo.Get(), BuildParamCollection(
                                    "PmxError", ((ProdumexException)ex).Error,
                                    "ShowButton", true));
                            WaitForMessage();
                        }

                        goto Step_IdentifyBaseComponent_ReEntry;
                    }
                }
                else if (msg.Name.EndsWith("BackRequested"))
                {
                    if (adHocAlternatePicking && alternateItemMode)
                    {
                        goto Step_Alternate_IdentifySSCCOrLocation;
                    }
                    goto Step_PickNotFullPallet;
                }


                query = BuildQuery.GetCodeNameItemsOrderByName(scannedItemCodeBaseComponent, gtinBaseComponent, Produmex.Sbo.Logex.Data.SboCommands.PackagingUnit.Sales, true, m_sboProviderService.GetDbTool());
                DataSet dsProductsBaseComponent = m_sboProviderService.RunView(false, null, null, query);
                if (GetNrOfRows(dsProductsBaseComponent) == 0)
                {
                    initialErrorKey = "MSG_ERROR_INCORRECT_PICK_PRODUCT";
                    goto Step_IdentifyBaseComponent_ReEntry;
                }

                //if we come here the dataset has rows
                scannedItemCodeBaseComponent = Convert.ToString(dsProductsBaseComponent.Tables[0].Rows[0][ProductDataSet.Columns.PRODUCT_CODE.Name]);


                if (scannedItemCodeBaseComponent != itemInfo.BaseItemInfo.ItemCode)
                {
                    initialErrorKey = "MSG_ERROR_INCORRECT_BASE_PRODUCT";
                    goto Step_IdentifyBaseComponent;
                }

                enteredItemCode = pickingItemInfo.ItemCode;
                enteredItemDescr = pickingItemInfo.ItemDescription;

                goto Step_GetBatchNumbers;
            }

            #endregion

        Step_IdentifyProduct:

            #region IDENTIFY PRODUCT

            gs1DataScanned = null;
            bcScanned = null;
            string gtin = null;
            string scannedItemCode = null;
			goBack = false;

            //session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.Picking.IIdentifyGS1DataPickingScreen),
            //    this.DefaultCultureInfo, BuildParamCollection(
            //    "InitialErrorKey", initialErrorKey,
            //    "TitleKey", "MSG_TITLE_IDENTIFY_PRODUCT",
            //    "BarcodeController", barcodeController,
            //    "AllowedButtons", disableSelectionOfItems || PmxOseCompany.Get().DisableItemSelectionInFlows ? null : new string[] { "MSG_BUTTON_SELECT_PRODUCT" },
            //    "ItemCode", pickingItemInfo.ItemCode,
            //    "ItemDescription", pickingItemInfo.ItemDescription,
            //    "ItemBarcode", pickingItemInfo.ItemBarcode,
            //    "ImagePath", itemInfo.PictureName,
            //    "StillToPick", pickingItemInfo.Quantity,
            //    "TotalQuantityToPick", pickingItemInfo.TotalQuantityToPick,
            //    "Uom", pickingItemInfo.Uom,
            //    "BatchNumber", (adHocAlternatePicking && alternateItemMode)? "":pickingItemInfo.BatchNumber,
            //    "BestBeforeDate", (adHocAlternatePicking && alternateItemMode)? null:pickingItemInfo.BestBeforeDate,
            //    "LocationCode", (adHocAlternatePicking && alternateItemMode)? enteredLocationCode:pickingItemInfo.LocationCode,
            //    "SSCC", (adHocAlternatePicking && alternateItemMode)? alternateSSCC:pickingItemInfo.SSCC,
            //    "IsAlternate", (adHocAlternatePicking && alternateItemMode),
            //    "GS1DataObject", gs1DataScanned));
            //msg = WaitForMessage();
            //initialErrorKey = null;

            //string lastButtonKey = null;
            //if (msg.Name == "Produmex.Foundation.SlimScreen.Interfaces.Picking.IIdentifyGS1DataPickingScreenReply.DataEntered")
            //{
            //    BarcodeContainer data = ExtractParameter<BarcodeContainer>(msg.Parameters, "data");
            //    if (data == null || string.IsNullOrEmpty(BarcodeHelper.GetProductIdentification(data)))
            //    {
            //        initialErrorKey = "MSG_ERROR_NO_PRODUCT_ON_BARCODE";
            //        goto Step_IdentifyProduct;
            //    }
            //    gs1DataScanned = data.GS1DataObject;

            //    gtin = BarcodeHelper.GetProductIdentification(data);
            //}
            //else if (msg.Name == "Produmex.Foundation.SlimScreen.Interfaces.Picking.IIdentifyGS1DataPickingScreenReply.ButtonClicked")
            //{
            //    lastButtonKey = ExtractParameter<string>(msg.Parameters, "buttonKey");
            //    if (lastButtonKey != "MSG_BUTTON_SELECT_PRODUCT")
            //    {
            //        initialErrorKey = KnownMessageKeys.MSG_ERROR_OPERATION_NOT_SUPPORTED;
            //        goto Step_IdentifyProduct;
            //    }
            //}
            //else if (msg.Name == "Produmex.Foundation.SlimScreen.Interfaces.Picking.IIdentifyGS1DataPickingScreenReply.BackRequested")
            //{
            //    goto Step_PickItem;
            //}
            //else if (msg.Name == "Produmex.Foundation.SlimScreen.Interfaces.Picking.IIdentifyGS1DataPickingScreenReply.StopRequested")
            //{
            //    return;
            //}
            //else
            //{
            //    initialErrorKey = KnownMessageKeys.MSG_ERROR_OPERATION_NOT_SUPPORTED;
            //    goto Step_IdentifyProduct;
            //}

			buttonList = new List<string>();

			if (!(disableSelectionOfItems || PmxOseCompany.Get().DisableItemSelectionInFlows))
			{
				buttonList.Add("MSG_BUTTON_SELECT_PRODUCT");
			}
			if (!shownPickLocationScreen)
			{
				buttonList.Add("MSG_BUTTON_OTHER_TASK");
			}

            session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.Picking.IIdentifyBarcodePickingScreen),
                this.DefaultCultureInfo, BuildParamCollection(
                "InitialErrorKey", initialErrorKey,
                "TitleKey", "MSG_TITLE_IDENTIFY_PRODUCT",
				"BarcodeController", barcodeController,
				"AllowedButtons", buttonList.ToArray(),
				"ItemCode", pickingItemInfo.ItemCode,
                "ItemDescription", pickingItemInfo.CustomItemDescription,
                "ItemBarcode", pickingItemInfo.ItemBarcode,
                "ImagePath", itemInfo.PictureName,
                "StillToPick", pickingItemInfo.Quantity,
                "TotalQuantityToPick", pickingItemInfo.TotalQuantityToPick,
                "Uom", pickingItemInfo.Uom,
                "BatchNumber", (adHocAlternatePicking && alternateItemMode) ? "" : pickingItemInfo.BatchNumber,
                "BestBeforeDate", (adHocAlternatePicking && alternateItemMode) ? null : pickingItemInfo.BestBeforeDate,
                "LocationCode", (adHocAlternatePicking && alternateItemMode) ? enteredLocationCode : pickingItemInfo.LocationCode,
                "SSCC", (adHocAlternatePicking && alternateItemMode) ? alternateSSCC : pickingItemInfo.SSCC,
                "IsAlternate", (adHocAlternatePicking && alternateItemMode)
                ));
            msg = WaitForMessage();
            initialErrorKey = null;

            string lastButtonKey = null;
            if (msg.Name == "Produmex.Foundation.SlimScreen.Interfaces.Picking.IIdentifyBarcodePickingScreenReply.DataEntered")
            {
                try
                {
                    string scannedValue = ExtractParameter<string>(msg.Parameters, "data");

                    bcScanned = barcodeController.GetScannedData(scannedValue, false, bcScanned);

                    if (bcScanned == null ||
                        (string.IsNullOrEmpty(BarcodeHelper.GetProductIdentification(bcScanned)) && string.IsNullOrEmpty(bcScanned.ItemCode)))
                    {
                        initialErrorKey = "MSG_ERROR_NO_PRODUCT_ON_BARCODE";
                        goto Step_IdentifyProduct;
                    }

                    gs1DataScanned = bcScanned.GS1DataObject;

                    gtin = BarcodeHelper.GetProductIdentification(bcScanned);
                    scannedItemCode = bcScanned.ItemCode;
                }
                catch (Exception ex)
                {
                    s_log.Error("", ex);
                    if (!(ex is ProdumexException) || ((ProdumexException)ex).Error == null)
                    {
                        // Inform the user with the error
                        session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
                            DefaultCultureInfo.Get(), BuildParamCollection(
                                "MessageKey", ExceptionHelper.Format(ExceptionHelper.Content.InnerExceptions, ex, DefaultCultureInfo.Get()),
                                "NoTranslationOfMessageKey", true,
                                "IsError", true,
                                "ShowButton", true));
                        WaitForMessage();

                    }
                    else
                    {
                        // Inform the user with the error
                        session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
                            DefaultCultureInfo.Get(), BuildParamCollection(
                                "PmxError", ((ProdumexException)ex).Error,
                                "ShowButton", true));
                        WaitForMessage();
                    }

                    goto Step_IdentifyProduct;
                }
            }
            else if (msg.Name == "Produmex.Foundation.SlimScreen.Interfaces.Picking.IIdentifyBarcodePickingScreenReply.ButtonClicked")
            {
                lastButtonKey = ExtractParameter<string>(msg.Parameters, "buttonKey");

				if (lastButtonKey == "MSG_BUTTON_OTHER_TASK")
				{
					goto Step_OtherTask;
				}

                if (lastButtonKey != "MSG_BUTTON_SELECT_PRODUCT")
                {
                    initialErrorKey = KnownMessageKeys.MSG_ERROR_OPERATION_NOT_SUPPORTED;
                    goto Step_IdentifyProduct;
                }
            }
            else if (msg.Name == "Produmex.Foundation.SlimScreen.Interfaces.Picking.IIdentifyBarcodePickingScreenReply.BackRequested")
            {
				goBack = true;
                goto Step_PickItem;
            }
            else if (msg.Name == "Produmex.Foundation.SlimScreen.Interfaces.Picking.IIdentifyBarcodePickingScreenReply.StopRequested")
            {
                return;
            }
            else
            {
                initialErrorKey = KnownMessageKeys.MSG_ERROR_OPERATION_NOT_SUPPORTED;
                goto Step_IdentifyProduct;
            }



            query = BuildQuery.GetProductsFromStockOrderByName(gtin, scannedItemCode, enteredLocationCode, null, false, Produmex.Sbo.Logex.Data.SboCommands.PackagingUnit.Sales, true, m_sboProviderService.GetDbTool());
            DataSet dsProducts = m_sboProviderService.RunView(false, null, null, query);
            int nrOfRows = GetNrOfRows(dsProducts);

            if (nrOfRows == 0 && lastButtonKey != "MSG_BUTTON_SELECT_PRODUCT")
            {
                initialErrorKey = "MSG_ERROR_INCORRECT_PICK_PRODUCT";
                goto Step_IdentifyProduct;
            }
            if (nrOfRows == 1 && lastButtonKey != "MSG_BUTTON_SELECT_PRODUCT")
            {
                enteredItemCode = Convert.ToString(dsProducts.Tables[0].Rows[0][ProductDataSet.Columns.PRODUCT_CODE.Name]);
                enteredItemDescr = Convert.ToString(dsProducts.Tables[0].Rows[0][ProductDataSet.Columns.PRODUCT_DESCRIPTION.Name]);
            }
            else
            {
                session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.ISelectProductScreen),
                    this.DefaultCultureInfo, BuildParamCollection(
                    "InitialErrorKey", initialErrorKey,
                    "TitleKey", "MSG_TITLE_SELECT_PRODUCT",
                    "ProductDS", dsProducts,
                    "AllowNoSelection", false));
                msg = WaitForMessage();
                initialErrorKey = null;

                if (msg.Name != "Produmex.Foundation.SlimScreen.Interfaces.ISelectProductScreenReply.ProductSelected")
                {
                    goto Step_IdentifyProduct;
                }

                enteredItemCode = ExtractParameter<string>(msg.Parameters, "itemCode");
                if (nrOfRows > 0)
                {
                    DataRow[] rows = dsProducts.Tables[0].Select(ProductDataSet.Columns.PRODUCT_CODE.Name + "='" + enteredItemCode + "'");
                    if (rows.Length > 0)
                    {
                        enteredItemDescr = Convert.ToString(rows[0][ProductDataSet.Columns.PRODUCT_DESCRIPTION.Name]);
                    }
                }
            }

            if (enteredItemCode != pickingItemInfo.ItemCode)
            {
                initialErrorKey = "MSG_ERROR_INCORRECT_PICK_PRODUCT";
                goto Step_IdentifyProduct;
            }

            #endregion

        Step_GetBatchNumbers:

            #region GET BATCH NUMBERS

            enteredBatchNumber = null;
            enteredQualityStatusCode = null;
            enteredBestBeforeDate = null;
            enteredItriKey = null;

            if (autoSelectItri && !(adHocAlternatePicking && alternateItemMode))
            {
                enteredBatchNumber = pickingItemInfo.BatchNumber;
                enteredBestBeforeDate = pickingItemInfo.BestBeforeDate;
                enteredItriKey = pickingItemInfo.ItemTransactionalInfoKey;
                enteredQualityStatusCode = pickingItemInfo.QualityStatusCode;
                goto Step_GetFreeQuantity;
            }

            if (gs1DataScanned != null)
            {
                query = BuildQuery.GetBatchNumbers((adHocAlternatePicking && alternateItemMode) ? alternateSSCC : null, enteredLocationCode, pickingItemInfo.ItemCode, gs1DataScanned.LotNumber, gs1DataScanned.BestBeforeDate);
            }
            else
            {
                query = BuildQuery.GetBatchNumbers((adHocAlternatePicking && alternateItemMode) ? alternateSSCC : null, enteredLocationCode, pickingItemInfo.ItemCode, null, null);
            }
            DataSet dsBatchNumbers = m_sboProviderService.RunView(false, null, null, query);


            if (GetNrOfRows(dsBatchNumbers) == 1
                && string.IsNullOrEmpty(initialErrorKey))
            {
                DataRow firstRow = dsBatchNumbers.Tables[0].Rows[0];
                enteredBatchNumber = firstRow[BatchNumberDataSet.Columns.BATCH_NUMBER.Name] as string;
                object bestBefore = firstRow[BatchNumberDataSet.Columns.BEST_BEFORE_DATE.Name];
                if (bestBefore is DateTime) enteredBestBeforeDate = Convert.ToDateTime(bestBefore);
                enteredItriKey = Convertor.ConvertToInt32Nullable(firstRow[BatchNumberDataSet.Columns.ITEMTRANSACTIONALINFOKEY.Name]);
                enteredQualityStatusCode = firstRow[BatchNumberDataSet.Columns.QUALITY_STATUS_CODE.Name] as string;
            }
            else if (!string.IsNullOrEmpty(initialErrorKey) &&
                adHocAlternatePicking && alternateItemMode &&
                GetNrOfRows(dsBatchNumbers) == 1)
            {
                goto Step_Alternate_IdentifySSCCOrLocation;
            }
            else
            {
                session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.ISelectBatchNumberScreen),
                    this.DefaultCultureInfo, BuildParamCollection(
                    "InitialErrorKey", initialErrorKey,
                    "TitleKey", "MSG_TITLE_SELECT_BATCH_NUMBER",
                    "BatchNumberDS", dsBatchNumbers));
                msg = WaitForMessage();
                initialErrorKey = null;

                if (msg.Name != "Produmex.Foundation.SlimScreen.Interfaces.ISelectBatchNumberScreenReply.BatchNumberSelected")
                {
                    goto Step_IdentifyBaseComponent;
                }

                enteredBatchNumber = ExtractParameter<string>(msg.Parameters, "batchNumber");
                enteredBestBeforeDate = ExtractParameter<DateTime?>(msg.Parameters, "bestBeforeDate");
                enteredItriKey = ExtractParameter<int?>(msg.Parameters, "itemTransactionalInfoKey");
                enteredQualityStatusCode = ExtractParameter<string>(msg.Parameters, "qualityStatusCode");
            }

            if (!(adHocAlternatePicking && alternateItemMode) || alternateCopyBatchNumbers)
            {

                if (enteredBatchNumber != pickingItemInfo.BatchNumber
                    || enteredBestBeforeDate != pickingItemInfo.BestBeforeDate
                    || enteredItriKey != pickingItemInfo.ItemTransactionalInfoKey
                    || enteredQualityStatusCode != pickingItemInfo.QualityStatusCode)
                {
                    initialErrorKey = "MSG_ERROR_INCORRECT_PICK_BATCH_AND_BEST_BEFORE";

                    if (gs1DataScanned != null &&
                        (gs1DataScanned.BestBeforeDate.HasValue || !string.IsNullOrEmpty(gs1DataScanned.LotNumber)))
                    {
                        //Scanned batch/bbd, and this is incorrect.
                        //So go back to scanning item, because a new barcode needs to be scanned.
                        goto Step_IdentifyBaseComponent;
                    }

                    goto Step_GetBatchNumbers;
                }
            }
            else
            {
                //all batches are allowed for the moment
                //But quality status not
                if (!listOfQualityStatusCanBeShipped.Contains(enteredQualityStatusCode))
                {
                    initialErrorKey = "MSG_ERROR_INCORRECT_PICK_QUALITY_STATUS";
                    goto Step_GetBatchNumbers;
                }

            }

            if (adHocAlternatePicking && alternateItemMode &&
                enteredItriKey.HasValue &&
                enteredItriKey != pickingItemInfo.ItemTransactionalInfoKey)
            {
                //Check if we need to limit batches based on batch attributes
                Collection<int> listOfAllowedItri = null;

                m_sboProviderService.InvokeMethodWithDbConnection<object>(false, false, null, null,
                delegate(PmxDbConnection conn, object[] parameters)
                {
                    PmxItemTransactionalBatchAttributeProvider itbaProv = new PmxItemTransactionalBatchAttributeProvider(conn);
                    listOfAllowedItri = itbaProv.GetListOfAllowedItriForPickListLine(pickingItemInfo.PickListDocEntry, pickingItemInfo.PickListLineID);
                    return null;
                });


                if (listOfAllowedItri != null &&
                    !listOfAllowedItri.Contains(enteredItriKey.Value))
                {
                    initialErrorKey = "MSG_ERROR_INCORRECT_PICK_BATCH_AND_BEST_BEFORE_DOES_NOT_MATCH_BATCH_ATTRIBUTE";
                    goto Step_GetBatchNumbers;
                }
            }

            #endregion

        Step_GetFreeQuantity:

            #region GET FREE QUANTITY


            double maxQuantity;

            if (!itemInfo.IsInventoryItem)
            {
                if (PmxOseCompany.Get().AllowOverPicking)
                {
                    maxQuantity = -1;
                }
                else
                {
                    maxQuantity = pickingItemInfo.Quantity;
                }

                goto Step_EnterQuantity;
            }

            if (PmxOseCompany.Get().AllowOverPicking &&
                !(alternateItemMode && adHocAlternatePicking))
            {
                PickingItemInfo tempInfo = RuntimeHelper.DeepCopy<PickingItemInfo>(pickingItemInfo);

                if ((adHocAlternatePicking && alternateItemMode))
                {
                    //Set correct data for the entered values
                    tempInfo.SSCC = alternateSSCC;
                    tempInfo.LogisticUnitIdentKey = alternateLUID;
                    tempInfo.LocationCode = enteredLocationCode;
                    tempInfo.ItemTransactionalInfoKey = enteredItriKey;
                    tempInfo.BatchNumber = enteredBatchNumber;
                    tempInfo.QualityStatusCode = enteredQualityStatusCode;
                    tempInfo.BestBeforeDate = enteredBestBeforeDate;
                }

                double freeQuantity = m_sboProviderService.InvokeMethodWithDbConnection<double>(false, false, null, null,
                    delegate(PmxDbConnection conn, object[] parameters)
                    {
                        PmxPickListProvider plProv = new PmxPickListProvider(conn);
                        return plProv.GetFreeQuantityForPickingMoreThanNeeded(pickListDocEntry, tempInfo);
                    });

                // Is this ( ( IsAlternateItem ? FreeQuantity :   NextItem.Quantity ) + FreeQuantity );
                // or      (   IsAlternateItem ? FreeQuantity : ( NextItem.Quantity   + FreeQuantity ) );

				maxQuantity = alternateItemMode && !adHocAlternatePicking && !string.IsNullOrEmpty(proposedPickingItemInfo.LocationCode)
                    ? freeQuantity / pickingItemInfo.QuantityPerUom
                    : pickingItemInfo.Quantity + (freeQuantity / pickingItemInfo.QuantityPerUom);

                maxQuantity = Math.Round(maxQuantity, pickingItemInfo.QuantityPerUom != 1 ? 6 : itemInfo.UomDecimals);

            }
            else
            {
                maxQuantity = pickingItemInfo.Quantity;
            }

            #endregion

        Step_EnterQuantity:

            #region ENTER QUANTITY

            weightDetails = null;

            listOfSerialNumbers = null;
            m_subFlowListOfSerialNumbers = null;

            m_subFlowQuantity = 0;
            m_subFlowQuantityUom2 = null;
            try
            {
                m_enterPickingQuantityHookFlow.BindParameters(
                    this.DefaultCultureInfo,
                    this.PmxOseCompany,
                    new ReadOnlyBinder<PickingItemInfo>("PickingItemInfo", pickingItemInfo),
                    new ReadOnlyBinder<double>("MaximumQuantity", maxQuantity),
                    new ReadWriteBinder<double>("QuantityEntered", this, "m_subFlowQuantity"),
                    new ReadWriteBinder<double?>("QuantityUom2Entered", this, "m_subFlowQuantityUom2"),
                    new ReadWriteBinder<IPreAndPostActionInTransaction>("PreAndPostActionObject", this, "m_preAndPostActionOnPickItem"),
                    new ReadWriteBinder<bool>("BackRequested", this, "m_subFlowBackRequested"));
                m_preAndPostActionOnPickItem = null;
                m_subFlowBackRequested = false;
                this.StartSubFlowAndWait(m_enterPickingQuantityHookFlow, true);

                if (m_subFlowBackRequested)
                {
                    goto Step_IdentifyBaseComponent;
                }
            }
            catch
            {
            }

            if (m_subFlowQuantity == 0)
            {
                weightDetails = itemInfo.PickingRecordsWeightDetails ? new List<KeyValuePair<double, double>>() : null;

                if (((!multiPicking && string.IsNullOrEmpty(targetMoveableLocation))
                    && (itemInfo.HasSerialNumber || itemInfo.HasSAPSerialNumber)) ||
                    itemInfo.TrackLocationSerialNumber || (pickingItemInfo.ListOfSerialNumbers != null && pickingItemInfo.ListOfSerialNumbers.Count > 0))
                {
                    //enter serial number


                    PickingItemInfo tempInfo = RuntimeHelper.DeepCopy<PickingItemInfo>(pickingItemInfo);

                    if ((adHocAlternatePicking && alternateItemMode))
                    {
                        //Set correct data for the entered values
                        tempInfo.SSCC = alternateSSCC;
                        tempInfo.LogisticUnitIdentKey = alternateLUID;
                        tempInfo.LocationCode = enteredLocationCode;
                        tempInfo.ItemTransactionalInfoKey = enteredItriKey;
                        tempInfo.BatchNumber = enteredBatchNumber;
                        tempInfo.QualityStatusCode = enteredQualityStatusCode;
                        tempInfo.BestBeforeDate = enteredBestBeforeDate;
                    }


                    //show screen to ask user what to do.
                    buttonKeys = new string[2] { "MSG_BUTTON_SERIAL_NUMBER", "MSG_BUTTON_SERIAL_NUMBER_RANGE" };

                    ToolBarButtons toolBarButtons = ToolBarButtons.Back;

                    session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.ISelectButtonScreen),
                            DefaultCultureInfo.Get(), BuildParamCollection(
                            "InitialErrorKey", initialErrorKey,
                            "TitleKey", "MSG_TITLE_ENTER_QUANTITY_SERIAL_NUMBER_TASK",
                            "AllowedButtons", buttonKeys,
                            "EnabledToolBarButtons", toolBarButtons
                            ));
                    msg = WaitForMessage();
                    initialErrorKey = null;
                    if (msg.Name.EndsWith(".ButtonClicked"))
                    {
                        string filterButtonKey = ExtractParameter<string>(msg.Parameters, "buttonKey");

                        if (filterButtonKey == "MSG_BUTTON_SERIAL_NUMBER" ||
                            filterButtonKey == "MSG_BUTTON_SERIAL_NUMBER_RANGE")
                        {
                            #region ENTER SERIAL NUMBERS

                            Collection<SerialNumberInfo> listOfSerialNumbersToUse = new Collection<SerialNumberInfo>();

                            if (itemInfo.HasSerialNumber || itemInfo.HasSAPSerialNumber)
                            {
                                if (!tempInfo.IsConsolidated)
                                {
                                    //Group all linked serial numbers
                                    if (tempInfo.ListOfSerialNumbers != null)
                                    {
                                        listOfSerialNumbersToUse = tempInfo.ListOfSerialNumbers;
                                    }
                                }
                                else
                                {

                                    if (tempInfo.ListOfPickingItemInfos != null)
                                    {
                                        foreach (PickingItemInfo pickInfo in tempInfo.ListOfPickingItemInfos)
                                        {
                                            if (pickInfo.ListOfSerialNumbers != null && pickInfo.ListOfSerialNumbers.Count > 0)
                                            {
                                                //consolidated picking with linked serial numbers is not allowed (We should never come here!)
                                                throw new ProdumexException(Produmex.Sbo.Logex.Data.ProdumexErrors.SerialNumbersLinkedToPickListSoNoConsolidationIsAllowed(s_log, waveKey));
                                            }
                                        }
                                    }
                                }
                            }



                            bool isSerialNumberRange = (filterButtonKey == "MSG_BUTTON_SERIAL_NUMBER_RANGE");

                            SubScannedBarcodeContainer = bcScanned;

                            m_enterQuantitySubFlow.BindParameters(
                                       this.DefaultCultureInfo,
                                       this.PmxOseCompany,
                                       new ReadOnlyBinder<string>("TitleKey", "MSG_TITLE_ENTER_QUANTITY"),
                                       new ReadOnlyBinder<PmxItemInfo>("ItemInfo", itemInfo),
                                       new ReadOnlyBinder<string>("BatchNumber", tempInfo.BatchNumber),
                                       new ReadOnlyBinder<DateTime?>("BestBeforeDate", tempInfo.BestBeforeDate),
                                       new ReadOnlyBinder<int?>("ItriKey", tempInfo.ItemTransactionalInfoKey),
                                       new ReadOnlyBinder<string>("LocationCode", tempInfo.LocationCode),
                                       new ReadOnlyBinder<string>("SSCC", tempInfo.SSCC),

                                       new ReadOnlyBinder<string>("DocNum", waveKey.ToString()),
                                       new ReadOnlyBinder<bool>("UsePickingInfo", true),
                                       new ReadOnlyBinder<bool>("IsAlternatePicking", (adHocAlternatePicking && alternateItemMode)),

                                       new ReadOnlyBinder<bool>("AllowZero", false),
                                       new ReadOnlyBinder<double>("InitialQuantity", pickingItemInfo.Quantity),
                                       new ReadOnlyBinder<string>("MessageKeyStillQuantity", null),
                                       new ReadOnlyBinder<double>("StillToQuantity", pickingItemInfo.Quantity),
                                       new ReadOnlyBinder<double>("MaximumQuantity", maxQuantity),
                                       new ReadOnlyBinder<double>("MinimumQuantity", -1),
                                       new ReadOnlyBinder<double>("TotalQuantityToPick", pickingItemInfo.TotalQuantityToPick),
                                       new ReadOnlyBinder<string>("UOM", pickingItemInfo.Uom),
                                       new ReadOnlyBinder<int>("NumberOfDigitsAfterDecimal", (itemInfo != null) ? itemInfo.UomDecimals : 0),

                                       new ReadOnlyBinder<double>("InnerQuantity", pickingItemInfo.QuantityPerUom),
                                       new ReadOnlyBinder<string>("InnerUOM", pickingItemInfo.InventoryUom),
                                       new ReadOnlyBinder<bool>("ForceInputSerialNumbers", true),//(!multiPicking && string.IsNullOrEmpty(targetMoveableLocation)) ||  itemInfo.TrackLocationSerialNumber),
                                       new ReadOnlyBinder<string>("ObjectType", Produmex.Sbo.Logex.Data.BusinessObjects.Definitions.PmxObjectTypes.PmxPickList),
                                       new ReadOnlyBinder<int?>("DocEntry", pickingItemInfo.PickListDocEntry),
                                       new ReadOnlyBinder<bool>("AutoFillQuantityForPackagingTypes", autoFillQuantityOnPackagingType),

                                       new ReadOnlyBinder<bool>("UseSerialNumberRange", isSerialNumberRange),

                                       new ReadOnlyBinder<int?>("LUID", tempInfo.LogisticUnitIdentKey),
                                       new ReadOnlyBinder<BarcodeContainer>("ScannedBarcodeContainer", SubScannedBarcodeContainer),

                                       new ReadOnlyBinder<Collection<SerialNumberInfo>>("ListOfSerialNumbersToUse", listOfSerialNumbersToUse),
                                       new ReadOnlyBinder<int?>("LineNum", pickingItemInfo.PickListLineID),
                                       new ReadOnlyBinder<double?>("StillToQuantityUom2", pickingItemInfo.QuantityUom2),

                                       new ReadWriteBinder<bool>("BackRequested", this, "m_subFlowBackRequested"),
                                       new ReadWriteBinder<double>("QuantityEntered", this, "m_subFlowQuantity"),
                                       new ReadWriteBinder<double?>("QuantityUom2Entered", this, "m_subFlowQuantityUom2"),
                                       new ReadWriteBinder<Collection<string>>("ListOfSerialNumbers", this, "m_subFlowListOfSerialNumbers"),
                                       ReadOnlyBinder.Make("WeightDetails",weightDetails)
                                );


                            #endregion

                        }

                    }
                    else
                    {
                        //go back
                        goto Step_IdentifyBaseComponent;
                    }


                }
                else
                {

                    bool overrideAskingWeight = false;

                    if (multiPicking || !string.IsNullOrEmpty(targetMoveableLocation) || alwaysSetToStatusPicked)
                    {
                        overrideAskingWeight = true;
                    }

                    SubScannedBarcodeContainer = bcScanned;

                    m_enterQuantitySubFlow.BindParameters(
                        this.DefaultCultureInfo,
                        this.PmxOseCompany,
                        new ReadOnlyBinder<string>("TitleKey", "MSG_TITLE_ENTER_QUANTITY"),
                        new ReadOnlyBinder<PmxItemInfo>("ItemInfo", itemInfo),
                        new ReadOnlyBinder<string>("BatchNumber", (adHocAlternatePicking && alternateItemMode) ? enteredBatchNumber : pickingItemInfo.BatchNumber),
                        new ReadOnlyBinder<DateTime?>("BestBeforeDate", (adHocAlternatePicking && alternateItemMode) ? enteredBestBeforeDate : pickingItemInfo.BestBeforeDate),
                        new ReadOnlyBinder<int?>("ItriKey", (adHocAlternatePicking && alternateItemMode) ? enteredItriKey : pickingItemInfo.ItemTransactionalInfoKey),
                        new ReadOnlyBinder<string>("LocationCode", (adHocAlternatePicking && alternateItemMode) ? enteredLocationCode : pickingItemInfo.LocationCode),
                        new ReadOnlyBinder<string>("SSCC", (adHocAlternatePicking && alternateItemMode) ? alternateSSCC : pickingItemInfo.SSCC),

                        new ReadOnlyBinder<string>("DocNum", waveKey.ToString()),
                        new ReadOnlyBinder<bool>("UsePickingInfo", true),
                        new ReadOnlyBinder<bool>("IsAlternatePicking", (adHocAlternatePicking && alternateItemMode)),

                        new ReadOnlyBinder<bool>("AllowZero", false),
                        new ReadOnlyBinder<double>("InitialQuantity", pickingItemInfo.Quantity),
                        new ReadOnlyBinder<string>("MessageKeyStillQuantity", null),
                        new ReadOnlyBinder<double>("StillToQuantity", pickingItemInfo.Quantity),
                        new ReadOnlyBinder<double>("MaximumQuantity", maxQuantity),
                        new ReadOnlyBinder<double>("MinimumQuantity", -1),
                        new ReadOnlyBinder<double>("TotalQuantityToPick", pickingItemInfo.TotalQuantityToPick),
                        new ReadOnlyBinder<string>("UOM", pickingItemInfo.Uom),
                        new ReadOnlyBinder<int>("NumberOfDigitsAfterDecimal", (itemInfo != null) ? itemInfo.UomDecimals : 0),

                        new ReadOnlyBinder<double>("InnerQuantity", pickingItemInfo.QuantityPerUom),
                        new ReadOnlyBinder<string>("InnerUOM", pickingItemInfo.InventoryUom),
                        new ReadOnlyBinder<bool>("ForceInputSerialNumbers", !multiPicking && string.IsNullOrEmpty(targetMoveableLocation)),
                        new ReadOnlyBinder<string>("ObjectType", Produmex.Sbo.Logex.Data.BusinessObjects.Definitions.PmxObjectTypes.PmxPickList),
                        new ReadOnlyBinder<int?>("DocEntry", pickingItemInfo.PickListDocEntry),
                        new ReadOnlyBinder<bool>("AutoFillQuantityForPackagingTypes", autoFillQuantityOnPackagingType),
                        new ReadOnlyBinder<bool>("OverrideAskingWeight", overrideAskingWeight),
                        new ReadOnlyBinder<BarcodeContainer>("ScannedBarcodeContainer", SubScannedBarcodeContainer), 
                        new ReadOnlyBinder<double?>("CatchWeightPieces", null),
                        new ReadOnlyBinder<bool>("HidePickingLocation", !itemInfo.IsInventoryItem),
                        new ReadOnlyBinder<double?>("StillToQuantityUom2", pickingItemInfo.QuantityUom2),
						new ReadOnlyBinder<bool>("SetInitialQuantity", true),

                        new ReadWriteBinder<bool>("BackRequested", this, "m_subFlowBackRequested"),
                        new ReadWriteBinder<double>("QuantityEntered", this, "m_subFlowQuantity"),
                        new ReadWriteBinder<double?>("QuantityUom2Entered", this, "m_subFlowQuantityUom2"),
                        new ReadWriteBinder<Collection<string>>("ListOfSerialNumbers", this, "m_subFlowListOfSerialNumbers"),
                        ReadOnlyBinder.Make("WeightDetails", weightDetails)
                    );
                }
                m_subFlowBackRequested = false;
                this.StartSubFlowAndWait(m_enterQuantitySubFlow, true);

                if (m_subFlowBackRequested)
                {
                    if (itemInfo.IsInventoryItem)
                    {
                        goto Step_IdentifyBaseComponent;
                    }
                    else
                    {
                        goto Step_ShowNextItems;
                    }
                }

                listOfSerialNumbers = m_subFlowListOfSerialNumbers;

            }

            #endregion

        Step_ScanMultiPickingMovLoc:

            #region SCAN MULTI PICKING MOV LOC

            if (!multiPicking) goto Step_SetPickingItemInfo;

            session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.Picking.IIdentifyLocationScreen),
                this.DefaultCultureInfo, BuildParamCollection(
                "InitialErrorKey", initialErrorKey,
                "TitleKey", "MSG_TITLE_IDENTIFY_MOVEABLE_LOCATION",
                "ItemCode", pickingItemInfo.ItemCode,
                "ItemDescription", pickingItemInfo.CustomItemDescription,
                "ItemBarcode", pickingItemInfo.ItemBarcode,
                "ImagePath", itemInfo.PictureName,
                "StillToPick", m_subFlowQuantity,
                "TotalQuantityToPick", proposedPickingItemInfo.TotalQuantityToPick,
                "Uom", pickingItemInfo.Uom,
                "BatchNumber", pickingItemInfo.BatchNumber,
                "BestBeforeDate", pickingItemInfo.BestBeforeDate,
                "SSCC", pickingItemInfo.SSCC,
                "LocationCode", multiPickingPickListInfos[pickingItemInfo.PickListDocEntry].MovableLocation));
            msg = WaitForMessage();
            initialErrorKey = null;

            if (!msg.Name.EndsWith(".IIdentifyLocationScreenReply.LocationEntered"))
            {
                initialErrorKey = KnownMessageKeys.MSG_ERROR_OPERATION_NOT_SUPPORTED;
                goto Step_ScanMultiPickingMovLoc;
            }
            if (String.Compare(ExtractParameter<string>(msg.Parameters, "locationCode"), multiPickingPickListInfos[pickingItemInfo.PickListDocEntry].MovableLocation, true) != 0)
            {
                initialErrorKey = "MSG_ERROR_INCORRECT_MOVABLE_LOCATION";
                goto Step_ScanMultiPickingMovLoc;
            }

            #endregion

        Step_SetPickingItemInfo:

            #region SET PICKING ITEM INFO AND PERFORM PICKING


            PickingItemInfo pickingItemInfoToSend = RuntimeHelper.DeepCopy<PickingItemInfo>(pickingItemInfo);

            if (adHocAlternatePicking && alternateItemMode)
            {
                pickingItemInfoToSend.AdHocAlternatePickingOriginalPickingItemInfo = pickingItemInfo;
                pickingItemInfoToSend.IsFromAdHocAlternatePicking = true;

                //Set correct data for the entered values
                pickingItemInfoToSend.SSCC = alternateSSCC;
                pickingItemInfoToSend.LogisticUnitIdentKey = alternateLUID;
                pickingItemInfoToSend.LocationCode = enteredLocationCode;
                pickingItemInfoToSend.ItemTransactionalInfoKey = enteredItriKey;
                pickingItemInfoToSend.BatchNumber = enteredBatchNumber;
                pickingItemInfoToSend.QualityStatusCode = enteredQualityStatusCode;
                pickingItemInfoToSend.BestBeforeDate = enteredBestBeforeDate;


            }


			if (!(pickingItemInfoToSend.IsFullPallet || pickingItemInfoToSend.IsFullMasterSSCC))
            {
                pickingItemInfoToSend.Quantity = m_subFlowQuantity;
                pickingItemInfoToSend.QuantityUom2 = m_subFlowQuantityUom2;
                pickingItemInfoToSend.ListOfWeightDetails = weightDetails;
            }
			else if (!pickingItemInfoToSend.IsFullMasterSSCC)
            {
				Collection<PickingItemInfo> listOfPickItemsToPerformActions = new Collection<PickingItemInfo>();

				if (!pickingItemInfoToSend.IsFullMixedPallet)
				{
					//No mixed pallet, so just add the pick object.
					//This will make sure the system performs the same tasks as before the support for mixed full pallets
					listOfPickItemsToPerformActions.Add(pickingItemInfoToSend);
				}
				else
				{
					//This is a mixed full pallet, so we need to loop through all items to see if we need to enter serial numbers or catch weight
					listOfPickItemsToPerformActions = pickingItemInfoToSend.ListOfPickingItemInfos;
				}

				foreach (PickingItemInfo tempPickItem in listOfPickItemsToPerformActions)
				{
					PmxItemInfo subItemInfo = ExecuteMethodWithDbConnection<PmxItemInfo>(out pmxError, false, false,
							delegate(PmxDbConnection conn, object[] parameters)
							{
								PmxItemAllConnectionsProvider itemProv = new PmxItemAllConnectionsProvider(conn);
								return itemProv.GetCachedItemInfo(tempPickItem.ItemCode);
							});

					//this is for a full pallet, so we need to enter serial numbers if needed
					if (subItemInfo.HasSerialNumber || subItemInfo.HasSAPSerialNumber)
					{
						#region SERIAL NUMBERS

						listOfSerialNumbers = null;
						m_subFlowListOfSerialNumbers = null;

						bool askSerialNumbers = true;

						//first try to get serial numbers from stock
						if (subItemInfo.TrackLocationSerialNumber)
						{
							this.ExecuteMethodWithDbConnection<object>(out pmxError, false, false,
								delegate(PmxDbConnection conn, object[] parameters)
								{
									PmxSerialNumberProvider senuProv = new PmxSerialNumberProvider(conn);
									Dictionary<string, int> temp = senuProv.GetSerialNumberKeyList(tempPickItem.ItemCode, (Collection<SerialNumberInfo>)null, tempPickItem.ItemTransactionalInfoKey, true, null /*pmxWarehouseCode*/, tempPickItem.LogisticUnitIdentKey);

									if (Convert.ToDouble(temp.Keys.Count) == tempPickItem.Quantity)
									{
										askSerialNumbers = false;
										listOfSerialNumbers = new Collection<string>();
										//fill list of serial numbers
										foreach (string serNum in temp.Keys)
										{
											listOfSerialNumbers.Add(serNum);
										}
									}

									return null;
								});

							if (pickingItemInfo.IsFullMixedPallet)
							{
								//Store serial numbers in actual sub picking item info object.
								tempPickItem.ListOfPickedSerialNumbers = listOfSerialNumbers;
							}

						}

						if (askSerialNumbers)
						{
							//HOOK
							m_enterSerialNumbersForFullPalletPickingHookFlow.BindParameters(
								  this.DefaultCultureInfo,
								  this.PmxOseCompany,
								  new ReadOnlyBinder<int?>("ItriKey", tempPickItem.ItemTransactionalInfoKey),
								  new ReadOnlyBinder<int?>("Luid", tempPickItem.LogisticUnitIdentKey),
								  new ReadOnlyBinder<Collection<string>>("ListOfSerialNumbers", this, "m_subFlowListOfSerialNumbers"),
								  new ReadWriteBinder<bool>("SerialNumbersSelected", this, "m_subFlowSerialNumbersSelected"),
								  new ReadWriteBinder<bool>("BackRequested", this, "m_subFlowBackRequested")
							  );

							m_subFlowListOfSerialNumbers = null;
							m_subFlowSerialNumbersSelected = false;
							m_subFlowBackRequested = false;

							this.StartSubFlowAndWait(m_enterSerialNumbersForFullPalletPickingHookFlow, true);

							if (m_subFlowBackRequested)
							{
								goto Step_FullPalletIdentifySSCC;
							}

							if (m_subFlowSerialNumbersSelected && m_subFlowListOfSerialNumbers != null && m_subFlowListOfSerialNumbers.Count == pickingItemInfoToSend.Quantity)
							{
								askSerialNumbers = false;
								listOfSerialNumbers = m_subFlowListOfSerialNumbers;
							}

							if (pickingItemInfo.IsFullMixedPallet)
							{
								//Store serial numbers in actual sub picking item info object.
								tempPickItem.ListOfPickedSerialNumbers = listOfSerialNumbers;
							}
						}


						if (askSerialNumbers)
						{


							//enter serial number

							//show screen to ask user what to do.
							buttonKeys = new string[2] { "MSG_BUTTON_SERIAL_NUMBER", "MSG_BUTTON_SERIAL_NUMBER_RANGE" };

							ToolBarButtons toolBarButtons = ToolBarButtons.Back;

							session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.ISelectButtonScreen),
									DefaultCultureInfo.Get(), BuildParamCollection(
									"InitialErrorKey", initialErrorKey,
									"TitleKey", "MSG_TITLE_ENTER_QUANTITY_SERIAL_NUMBER_TASK",
									"AllowedButtons", buttonKeys,
									"EnabledToolBarButtons", toolBarButtons
									));
							msg = WaitForMessage();
							initialErrorKey = null;
							if (msg.Name.EndsWith(".ButtonClicked"))
							{
								string filterButtonKey = ExtractParameter<string>(msg.Parameters, "buttonKey");

								if (filterButtonKey == "MSG_BUTTON_SERIAL_NUMBER" ||
									filterButtonKey == "MSG_BUTTON_SERIAL_NUMBER_RANGE")
								{
									#region ENTER SERIAL NUMBERS


									Collection<SerialNumberInfo> listOfSerialNumbersToUse = new Collection<SerialNumberInfo>();

									if (subItemInfo.HasSerialNumber || subItemInfo.HasSAPSerialNumber)
									{
										if (!pickingItemInfoToSend.IsConsolidated)
										{
											//Group all linked serial numbers
											if (pickingItemInfoToSend.ListOfSerialNumbers != null)
											{
												listOfSerialNumbersToUse = pickingItemInfoToSend.ListOfSerialNumbers;
											}
										}
										else
										{
											if (pickingItemInfoToSend.ListOfPickingItemInfos != null)
											{
												foreach (PickingItemInfo pickInfo in pickingItemInfoToSend.ListOfPickingItemInfos)
												{
													if (pickInfo.ListOfSerialNumbers != null && pickInfo.ListOfSerialNumbers.Count > 0)
													{
														//consolidated picking with linked serial numbers is not allowed (We should never come here!)
														throw new ProdumexException(Produmex.Sbo.Logex.Data.ProdumexErrors.SerialNumbersLinkedToPickListSoNoConsolidationIsAllowed(s_log, waveKey));
													}
												}
											}
										}
									}

									bool isSerialNumberRange = (filterButtonKey == "MSG_BUTTON_SERIAL_NUMBER_RANGE");

									SubScannedBarcodeContainer = bcScanned;

									m_enterQuantitySubFlow.BindParameters(
											   this.DefaultCultureInfo,
											   this.PmxOseCompany,
											   new ReadOnlyBinder<string>("TitleKey", "MSG_TITLE_ENTER_QUANTITY"),
											   new ReadOnlyBinder<PmxItemInfo>("ItemInfo", subItemInfo),
											   new ReadOnlyBinder<string>("BatchNumber", tempPickItem.BatchNumber),
											   new ReadOnlyBinder<DateTime?>("BestBeforeDate", tempPickItem.BestBeforeDate),
											   new ReadOnlyBinder<int?>("ItriKey", tempPickItem.ItemTransactionalInfoKey),
											   new ReadOnlyBinder<string>("LocationCode", tempPickItem.LocationCode),
											   new ReadOnlyBinder<string>("SSCC", tempPickItem.SSCC),

											   new ReadOnlyBinder<string>("DocNum", waveKey.ToString()),
											   new ReadOnlyBinder<bool>("UsePickingInfo", true),

											   new ReadOnlyBinder<bool>("AllowZero", false),
											   new ReadOnlyBinder<double>("InitialQuantity", tempPickItem.Quantity),
											   new ReadOnlyBinder<string>("MessageKeyStillQuantity", null),
											   new ReadOnlyBinder<double>("StillToQuantity", tempPickItem.Quantity),
											   new ReadOnlyBinder<double>("MaximumQuantity", tempPickItem.Quantity),
											   new ReadOnlyBinder<double>("MinimumQuantity", tempPickItem.Quantity),
											   new ReadOnlyBinder<double>("TotalQuantityToPick", tempPickItem.TotalQuantityToPick),
											   new ReadOnlyBinder<string>("UOM", tempPickItem.Uom),
											   new ReadOnlyBinder<int>("NumberOfDigitsAfterDecimal", (subItemInfo != null) ? subItemInfo.UomDecimals : 0),

											   new ReadOnlyBinder<double>("InnerQuantity", tempPickItem.QuantityPerUom),
											   new ReadOnlyBinder<string>("InnerUOM", tempPickItem.InventoryUom),
											   new ReadOnlyBinder<bool>("ForceInputSerialNumbers", true),
											   new ReadOnlyBinder<string>("ObjectType", Produmex.Sbo.Logex.Data.BusinessObjects.Definitions.PmxObjectTypes.PmxPickList),
											   new ReadOnlyBinder<int?>("DocEntry", tempPickItem.PickListDocEntry),
											   new ReadOnlyBinder<bool>("AutoFillQuantityForPackagingTypes", autoFillQuantityOnPackagingType),
											   new ReadOnlyBinder<BarcodeContainer>("ScannedBarcodeContainer", SubScannedBarcodeContainer),

											   new ReadOnlyBinder<bool>("UseSerialNumberRange", isSerialNumberRange),
											   new ReadOnlyBinder<int?>("LUID", tempPickItem.LogisticUnitIdentKey),

											   new ReadOnlyBinder<Collection<SerialNumberInfo>>("ListOfSerialNumbersToUse", listOfSerialNumbersToUse),
                                               new ReadOnlyBinder<int?>("LineNum", tempPickItem.PickListLineID),

                                               new ReadOnlyBinder<double?>("StillToQuantityUom2", tempPickItem.QuantityUom2),

                                               new ReadWriteBinder<bool>("BackRequested", this, "m_subFlowBackRequested"),
                                               new ReadWriteBinder<double>("QuantityEntered", this, "m_subFlowQuantity"),
                                               new ReadWriteBinder<double?>("QuantityUom2Entered", this, "m_subFlowQuantityUom2"),
											   new ReadWriteBinder<Collection<string>>("ListOfSerialNumbers", this, "m_subFlowListOfSerialNumbers")

										);


									#endregion

								}

							}
							else
							{
								//go back
								goto Step_FullPalletIdentifySSCC;
							}

							m_subFlowBackRequested = false;
							this.StartSubFlowAndWait(m_enterQuantitySubFlow, true);

							if (m_subFlowBackRequested)
							{
								goto Step_FullPalletIdentifySSCC;
							}

							listOfSerialNumbers = m_subFlowListOfSerialNumbers;

							if (pickingItemInfo.IsFullMixedPallet)
							{
								//Store serial numbers in actual sub picking item info object.
								tempPickItem.ListOfPickedSerialNumbers = listOfSerialNumbers;
							}
						}

						#endregion
					}

					if (subItemInfo.IsCatchWeightItem &&
						(subItemInfo.UomToUseForSales == PmxUomToUse.UOM1AndUOM2 ||
						subItemInfo.UomToUseForSales == PmxUomToUse.UOM2))
					{
						#region CATCH WEIGHT

						if (alwaysSetToStatusPicked || !String.IsNullOrEmpty(targetMoveableLocation))
						{
							//We need to go through packing, so ask for weight on packing flow.
							m_subFlowQuantityUom2 = PmxItemAllConnectionsProvider.CalculateWeightFromPieces(tempPickItem.Quantity, subItemInfo);
						}
						else
						{

							SubScannedBarcodeContainer = bcScanned;

							//We need to ask for the weight!
							m_enterQuantitySubFlow.BindParameters(
							    this.DefaultCultureInfo,
							    this.PmxOseCompany,
							    new ReadOnlyBinder<string>("TitleKey", "MSG_TITLE_ENTER_QUANTITY"),
							    new ReadOnlyBinder<PmxItemInfo>("ItemInfo", subItemInfo),
							    new ReadOnlyBinder<string>("BatchNumber", (adHocAlternatePicking && alternateItemMode) ? enteredBatchNumber : tempPickItem.BatchNumber),
							    new ReadOnlyBinder<DateTime?>("BestBeforeDate", (adHocAlternatePicking && alternateItemMode) ? enteredBestBeforeDate : tempPickItem.BestBeforeDate),
							    new ReadOnlyBinder<int?>("ItriKey", (adHocAlternatePicking && alternateItemMode) ? enteredItriKey : tempPickItem.ItemTransactionalInfoKey),
							    new ReadOnlyBinder<string>("LocationCode", (adHocAlternatePicking && alternateItemMode) ? enteredLocationCode : tempPickItem.LocationCode),
							    new ReadOnlyBinder<string>("SSCC", (adHocAlternatePicking && alternateItemMode) ? alternateSSCC : tempPickItem.SSCC),

							    new ReadOnlyBinder<string>("DocNum", waveKey.ToString()),
							    new ReadOnlyBinder<bool>("UsePickingInfo", true),
							    new ReadOnlyBinder<bool>("IsAlternatePicking", (adHocAlternatePicking && alternateItemMode)),

							    new ReadOnlyBinder<bool>("AllowZero", false),
							    new ReadOnlyBinder<double>("InitialQuantity", tempPickItem.Quantity),
							    new ReadOnlyBinder<string>("MessageKeyStillQuantity", null),
							    new ReadOnlyBinder<double>("StillToQuantity", tempPickItem.Quantity),
							    new ReadOnlyBinder<double>("MaximumQuantity", tempPickItem.Quantity),
							    new ReadOnlyBinder<double>("MinimumQuantity", tempPickItem.Quantity),
							    new ReadOnlyBinder<double>("TotalQuantityToPick", tempPickItem.TotalQuantityToPick),
							    new ReadOnlyBinder<string>("UOM", tempPickItem.Uom),
							    new ReadOnlyBinder<int>("NumberOfDigitsAfterDecimal", (subItemInfo != null) ? subItemInfo.UomDecimals : 0),

							    new ReadOnlyBinder<double>("InnerQuantity", tempPickItem.QuantityPerUom),
							    new ReadOnlyBinder<string>("InnerUOM", tempPickItem.InventoryUom),
							    new ReadOnlyBinder<bool>("ForceInputSerialNumbers", !multiPicking && string.IsNullOrEmpty(targetMoveableLocation)),
							    new ReadOnlyBinder<string>("ObjectType", Produmex.Sbo.Logex.Data.BusinessObjects.Definitions.PmxObjectTypes.PmxPickList),
							    new ReadOnlyBinder<int?>("DocEntry", tempPickItem.PickListDocEntry),
							    new ReadOnlyBinder<bool>("AutoFillQuantityForPackagingTypes", autoFillQuantityOnPackagingType),
							    new ReadOnlyBinder<BarcodeContainer>("ScannedBarcodeContainer", SubScannedBarcodeContainer),
								    //new ReadOnlyBinder<double?>("MaximumQuantityUom2", pickingItemInfo.QuantityUom2),
							    new ReadOnlyBinder<double?>("CatchWeightPieces", tempPickItem.Quantity * tempPickItem.QuantityPerUom),
							    new ReadOnlyBinder<double?>("StillToQuantityUom2", tempPickItem.QuantityUom2),

                                new ReadWriteBinder<bool>("BackRequested", this, "m_subFlowBackRequested"),
							    new ReadWriteBinder<double>("QuantityEntered", this, "m_subFlowQuantity"),
                                new ReadWriteBinder<double?>("QuantityUom2Entered", this, "m_subFlowQuantityUom2"),
                                new ReadWriteBinder<Collection<string>>("ListOfSerialNumbers", this, "m_subFlowListOfSerialNumbers")

							);

							m_subFlowBackRequested = false;
							this.StartSubFlowAndWait(m_enterQuantitySubFlow, true);

							if (m_subFlowBackRequested)
							{
								goto Step_FullPalletIdentifySSCC;
							}
						}

						tempPickItem.QuantityUom2 = m_subFlowQuantityUom2;

						#endregion
					}
				}

				if (pickingItemInfo.IsFullMixedPallet)
				{
					//Clear global list of serial numbers, as the mixed pallet will have the serial numbers in each sub picking info object.
					listOfSerialNumbers = null;
				}
            }
            else
            {
                //Full master SSCC
                //If we only scanned a sub SSCC, pick the sub SSCC.
                if (!targetMasterLuid.HasValue)
                {
                    foreach (PickingItemInfo info in pickingItemInfo.ListOfPickingItemInfos)
                    {
                        if (info.LogisticUnitIdentKey == targetLuid)
                        {
                            pickingItemInfoToSend = RuntimeHelper.DeepCopy<PickingItemInfo>(info);
                            break;
                        }
                    }
                }
            }

            if (!String.IsNullOrEmpty(targetMoveableLocation))
            {
                dstLocationCode = targetMoveableLocation;
            }
            else if (!string.IsNullOrEmpty(targetLogCarrierStorLocCode))
            {
                dstLocationCode = targetLogCarrierStorLocCode;
            }
            else
            {
                dstLocationCode = pickingItemInfo.DestinationStorLocCode;
            }

            bool locationIsLocked = m_sboProviderService.InvokeMethodWithDbConnection<bool>(false, false, null, null,
                delegate(PmxDbConnection conn, object[] parameters)
                {
                    PmxOseStorageLocationProvider locProv = new PmxOseStorageLocationProvider(conn);
                    return locProv.IsLocationLocked(dstLocationCode);
                });
            if (locationIsLocked)
            {
                session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
                    this.DefaultCultureInfo, BuildParamCollection(
                    "MessageKey", "MSG_ERROR_LOCATION_LOCKED",
                    "ShowButton", true));
                msg = WaitForMessage();
                goto Step_SelectWave;
            }

            bool doRetry = false;
            do
            {
				int? targetLUIDToStartWith = targetLuid;
				int? targetMasterLUIDToStartWith = targetMasterLuid;
                doRetry = false;
                this.ExecuteMethodWithDbConnection<int?>(out pmxError, false, false,
                delegate(PmxDbConnection conn, object[] parameters)
                {
                    if (m_preAndPostActionOnPickItem != null) m_preAndPostActionOnPickItem.PreAction(conn);

                    PmxPickListProvider plProv = new PmxPickListProvider(conn);

                    PickedResultInfo pickResult = plProv.PickItem(pickingItemInfoToSend, pickListDocEntry, targetLuid, dstLocationCode, targetLogCarrierCode, pickingItemInfoToSend.IsFullPallet, !String.IsNullOrEmpty(targetMoveableLocation), listOfSerialNumbers, pickingItemInfoToSend.IsFullMasterSSCC, targetMasterLuid, createMasterSSCC);

                  
                    if (m_preAndPostActionOnPickItem != null) m_preAndPostActionOnPickItem.PostAction(conn);
					
					//Only store target LUID when not multipicking
					if (!multiPicking)
					{
						targetLuid = pickResult.LUID;
						//if (pickingItemInfoToSend.IsFullMasterSSCC)
						{
							targetMasterLuid = pickResult.MasterLUID;
						}
					}
					return targetLuid;

                    
                });
                m_preAndPostActionOnPickItem = null;
                if (pmxError != null)
                {
					//Error occured, so reset the target luid's
					targetLuid = targetLUIDToStartWith;
					targetMasterLuid = targetMasterLUIDToStartWith;

                    session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
                        this.DefaultCultureInfo, BuildParamCollection(
                        "PmxError", pmxError,
                        "ShowButton", true));
                    msg = WaitForMessage();

                    //retry?
                    session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IDecisionScreen),
                        DefaultCultureInfo.Get(), BuildParamCollection(
                            "TitleKey", "MSG_TITLE_RETRY_UPDATE",
                            "MessageKey", "MSG_QUESTION_AN_ERROR_OCCURED_WANT_RETRY"));
                    msg = WaitForMessage();

                    if (msg.Name.ToLower().EndsWith(".yes"))
                    {
                        doRetry = true;
                    }
                    else
                    {
                        if (alternateItemMode && adHocAlternatePicking)
                        {
                            goto Step_Alternate_IdentifySSCCOrLocation;
                        }
                        goto Step_SelectWave;
                    }
                }

            } while (doRetry);


            if (maxTimeShowMsgItemHasBeenPicked >= 0)
            {
                session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
                    this.DefaultCultureInfo, BuildParamCollection(
                    "MessageKey", "MSG_ITEMS_PICKED",
                    "ShowButton", true,
                    "MaxTimeForMessage", maxTimeShowMsgItemHasBeenPicked));
                msg = WaitForMessage();
            }


            #region AFTER ITEM PICKED HOOK FLOW

            m_afterItemPickedHookScript.BindParameters(
                  new ReadOnlyBinder<CultureInfo>("DefaultCultureInfo", this.DefaultCultureInfo),
                  new ReadOnlyBinder<PmxOseCompany>("PmxOseCompany", this.PmxOseCompany),
                  new ReadOnlyBinder<int?>("LUID", targetLuid),
                  new ReadOnlyBinder<Collection<PickingItemInfo>>("PickedItemInfos", new Collection<PickingItemInfo>() { pickingItemInfoToSend }),
                  new ReadOnlyBinder<Collection<string>>("ListOfSerialNumbers", listOfSerialNumbers),
                  new ReadOnlyBinder<int>("DocEntry", waveKey),
                  new ReadOnlyBinder<string>("ObjType", Produmex.Sbo.Logex.Data.BusinessObjects.Definitions.PmxObjectTypes.PmxWave)
                  );

            try
            {
                //call flow
                this.StartSubFlowAndWait(m_afterItemPickedHookScript, true);
            }
            catch (Exception ex)
            {
                s_log.Error("", ex);
                if (!(ex is ProdumexException) || ((ProdumexException)ex).Error == null)
                {
                    // Inform the user with the error
                    session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
                        DefaultCultureInfo.Get(), BuildParamCollection(
                            "MessageKey", ExceptionHelper.Format(ExceptionHelper.Content.InnerExceptions, ex, DefaultCultureInfo.Get()),
                            "NoTranslationOfMessageKey", true,
                            "IsError", true,
                            "ShowButton", true));
                    WaitForMessage();

                }
                else
                {
                    // Inform the user with the error
                    session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
                        DefaultCultureInfo.Get(), BuildParamCollection(
                            "PmxError", ((ProdumexException)ex).Error,
                            "ShowButton", true));
                    WaitForMessage();
                }
            }

            #endregion


            PmxOperationTypeConvertor convertor = new PmxOperationTypeConvertor();

            if (itemInfo.IsInventoryItem)
            {
                m_cycleCountSubFlow.BindParameters(
                    this.DefaultCultureInfo,
                    this.PmxOseCompany,
                    new ReadOnlyBinder<string>("LocationCode", pickingItemInfo.LocationCode),
                    new ReadOnlyBinder<int>("OtherOperationType", m_otherOperationTypeCycleCountingFlow));

                this.StartSubFlowAndWait(m_cycleCountSubFlow, true);
            }



            //check if alternate pick location was used, and if cycle count is allowed for that
            if (allowCycleCountOfOriginalPickLocationOnAlternatePicking &&
                pickingItemInfo.LocationCode != proposedPickingItemInfo.LocationCode &&
                !string.IsNullOrEmpty(proposedPickingItemInfo.LocationCode))
            {
                m_cycleCountSubFlow.BindParameters(
                this.DefaultCultureInfo,
                this.PmxOseCompany,
                new ReadOnlyBinder<string>("LocationCode", proposedPickingItemInfo.LocationCode),
                new ReadOnlyBinder<int>("OtherOperationType", m_otherOperationTypeAlternateCycleCountingFlow));

                this.StartSubFlowAndWait(m_cycleCountSubFlow, true);
            }

            // alternateItemMode = false;  // Will be done by or after Step_GetPickingItemInfo
            targetLogCarrierCanBeFull = true;

            //clear targetLogCarrierCode
            targetLogCarrierCode = null;
            ssccToPickOnHasStock = true;

            if (pickingItemInfo.IsFullPallet)
            {
                goto Step_LuidFull;
            }

            goto Step_GetPickingItemInfo;

            #endregion

        Step_OtherTask:

            #region OTHER TASK

            buttonKeyList = new List<string>();

            if (!String.IsNullOrEmpty(targetMoveableLocation))
            {
                buttonKeyList.Add("MSG_BUTTON_USE_OTHER_MOVEABLE_LOCATION");
            }

            if (targetLogCarrierCanBeFull)
            {
                buttonKeyList.Add("MSG_BUTTON_LOGISTIC_CARRIER_IS_FULL");
            }
            else if (targetMasterLuid.HasValue)
            {
                buttonKeyList.Add("MSG_BUTTON_MASTER_LUID_FULL");
            }

            if (!(pickingItemInfo.ListOfSerialNumbers != null &&
                pickingItemInfo.ListOfSerialNumbers.Count > 0 &&
                itemInfo.TrackLocationSerialNumber))
            {
                buttonKeyList.Add("MSG_BUTTON_ALTERNATE_ITEM");
            }

            buttonKeyList.Add("MSG_BUTTON_SKIP_ITEM");

            //session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.Picking.IShowNextItemScreen),
            //    this.DefaultCultureInfo, BuildParamCollection(
            //    "InitialErrorKey", initialErrorKey,
            //    "TitleKey", "MSG_TITLE_SELECT_TASK",
            //    "ItemCode", pickingItemInfo.ItemCode,
            //    "ItemDescription", pickingItemInfo.ItemDescription,
            //    "ItemBarcode", pickingItemInfo.ItemBarcode,
            //    "ImagePath", itemInfo.PictureName,
            //    "StillToPick", pickingItemInfo.Quantity,
            //    "TotalQuantityToPick", pickingItemInfo.TotalQuantityToPick,
            //    "Uom", pickingItemInfo.Uom,
            //    "BatchNumber", pickingItemInfo.BatchNumber,
            //    "BestBeforeDate", pickingItemInfo.BestBeforeDate,
            //    "LocationCode", pickingItemInfo.LocationCode,
            //    "ButtonKeys", buttonKeyList.ToArray()));

            session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.ISelectButtonScreen),
                this.DefaultCultureInfo, BuildParamCollection(
                "InitialErrorKey", initialErrorKey,
                "TitleKey", "MSG_TITLE_SELECT_TASK",
                "AllowedButtons", buttonKeyList.ToArray()));


            msg = WaitForMessage();
            initialErrorKey = null;
            if (msg.Name.EndsWith(".ButtonClicked"))
            {
                string buttonKey = ExtractParameter<string>(msg.Parameters, "buttonKey");
                if (buttonKey == "MSG_BUTTON_ALTERNATE_ITEM")
                {
                    goto Step_AlternateItem;
                }
                if (buttonKey == "MSG_BUTTON_LOGISTIC_CARRIER_IS_FULL" ||
                    buttonKey == "MSG_BUTTON_MASTER_LUID_FULL")
                {
                    if (createMasterSSCC &&
                        buttonKey != "MSG_BUTTON_MASTER_LUID_FULL")
                    {
                        //We are creating a master SSCC, so we need to ask if the Master SSCC is full, or if the sub LUID is full
                        buttonKeys = new string[2] { "MSG_BUTTON_MASTER_LUID_FULL", "MSG_BUTTON_SUB_SSCC_FULL" };

                        ToolBarButtons toolBarButtons = ToolBarButtons.Back;

                        session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.ISelectButtonScreen),
                                DefaultCultureInfo.Get(), BuildParamCollection(
                                "InitialErrorKey", initialErrorKey,
                                "TitleKey", "MSG_TITLE_SELECT_TASK",
                                "AllowedButtons", buttonKeys,
                                "EnabledToolBarButtons", toolBarButtons
                                ));
                        msg = WaitForMessage();
                        initialErrorKey = null;
                        if (msg.Name.EndsWith(".ButtonClicked"))
                        {
                            string filterButtonKey = ExtractParameter<string>(msg.Parameters, "buttonKey");

                            if (filterButtonKey == "MSG_BUTTON_MASTER_LUID_FULL")
                            {
                                masterLUIDFull = true;
                            }
                           
                            goto Step_LuidFull;
                           

                        }
                        else
                        {
                            //go back
                            goto Step_OtherTask;
                        }
                    }

                    session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IDecisionScreen),
                        this.DefaultCultureInfo, BuildParamCollection(
                        "TitleKey", "MSG_TITLE_CONFIRM_FULL_LUID",
                        "MessageKey", "MSG_QUESTION_CONFIRM_FULL_LUID"));
                    msg = WaitForMessage();

                    if (msg.Name == "Produmex.Foundation.SlimScreen.Interfaces.IDecisionScreenReply.Yes")
                    {
                        if (buttonKey == "MSG_BUTTON_MASTER_LUID_FULL")
                        {
                            masterLUIDFull = true;
                        }
                        goto Step_LuidFull;
                    }

                    goto Step_OtherTask;
                }
                if (buttonKey == "MSG_BUTTON_SKIP_ITEM")
                {
                    goto Step_SkipItem;
                }
                if (buttonKey == "MSG_BUTTON_USE_OTHER_MOVEABLE_LOCATION")
                {
                    targetMoveableLocation = null;
                    goto Step_IdentifyMoveableLocation;
                }
            }

            goto Step_PickItem;

            #endregion

        Step_AlternateItem:

            #region SELECT ALTERNATE ITEM

            if (!adHocAlternatePicking)
            {

                Collection<PickingItemInfo> alternateItems = this.ExecuteMethodWithDbConnection<Collection<PickingItemInfo>>(out pmxError, false, false,
                    delegate(PmxDbConnection conn, object[] parameters)
                    {
                        PmxPickListProvider plProv = new PmxPickListProvider(conn);
                        return plProv.GetAlternateItems(pickListDocEntry, proposedPickingItemInfo);
                    });
                if (pmxError != null)
                {
                    session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
                        this.DefaultCultureInfo, BuildParamCollection(
                        "PmxError", pmxError,
                        "ShowButton", true));
                    msg = WaitForMessage();

                    goto Step_OtherTask;
                }

                DataSet dsAlternateItems = BuildAlternateList(alternateItems);

                //DataSet dsAlternateItems = new DataSet();
                //DataTable tblAlternateItems = dsAlternateItems.Tables.Add("AlternateItems");
                //tblAlternateItems.Columns.Add(AlternateItemDataSet.Columns.INTERNAL_KEY.Name, typeof(int));
                //tblAlternateItems.Columns.Add(AlternateItemDataSet.Columns.STORAGE_LOCATION.Name);
                //tblAlternateItems.Columns.Add(AlternateItemDataSet.Columns.QUANTITY.Name, typeof(double));
                //tblAlternateItems.Columns.Add(AlternateItemDataSet.Columns.ITEMCODE.Name);
                //tblAlternateItems.Columns.Add(AlternateItemDataSet.Columns.BESTBEFOREDATE.Name, typeof(DateTime));
                //tblAlternateItems.Columns.Add(AlternateItemDataSet.Columns.BATCHNUMBER.Name);
                //tblAlternateItems.Columns.Add(AlternateItemDataSet.Columns.UOM.Name);
                //tblAlternateItems.Columns.Add(AlternateItemDataSet.Columns.ITEM_DESCRIPTION.Name);
                //tblAlternateItems.Columns.Add(AlternateItemDataSet.Columns.ITEM_BARCODE.Name);
                //tblAlternateItems.Columns.Add(AlternateItemDataSet.Columns.SSCC.Name);
                //tblAlternateItems.Columns.Add(AlternateItemDataSet.Columns.UOM2.Name);
                //tblAlternateItems.Columns.Add(AlternateItemDataSet.Columns.QUANTITY_UOM2.Name, typeof(double));
                //tblAlternateItems.Columns.Add(AlternateItemDataSet.Columns.BATCHNUMBER2.Name);

                //for (int i = 0; i < alternateItems.Count; i++)
                //{
                //    DataRow row = tblAlternateItems.Rows.Add(
                //        i,
                //        alternateItems[i].LocationCode,
                //        alternateItems[i].Quantity,
                //        alternateItems[i].ItemCode,
                //        alternateItems[i].BestBeforeDate,
                //        alternateItems[i].BatchNumber,
                //        alternateItems[i].Uom,
                //        alternateItems[i].ItemDescription,
                //        alternateItems[i].ItemBarcode,
                //        alternateItems[i].IsFullMasterSSCC ? alternateItems[i].MasterSSCC : alternateItems[i].IsFullPallet ? alternateItems[i].SSCC : "",
                //        alternateItems[i].Uom2,
                //        alternateItems[i].QuantityUom2,
                //        alternateItems[i].BatchNumber2
                //        );
                //}

                session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.Picking.ISelectAlternateItemScreen),
                    this.DefaultCultureInfo, BuildParamCollection(
                    "InitialErrorKey", initialErrorKey,
                    "TitleKey", "MSG_TITLE_SELECT_ALTERNATE_ITEM",
                    "DSAlternateItem", dsAlternateItems));
                msg = WaitForMessage();
                initialErrorKey = null;

                if (msg.Name != "Produmex.Foundation.SlimScreen.Interfaces.Picking.ISelectAlternateItemScreenReply.AlternateItemSelected")
                {
                    goto Step_OtherTask;
                }

                int altItemKey = ExtractParameter<int>(msg.Parameters, "internalKey");
                alternateItemMode = true;
                pickingItemInfo = alternateItems[altItemKey];

            }
            else
            {
                alternateItemMode = true;
            }

            goto Step_WorkWithPickingItemInfo;

            #endregion

        Step_LuidFull:

            #region LUID IS FULL

            if (!String.IsNullOrEmpty(targetMoveableLocation))
            {
                // goto Step_ContinueWithNextItem;
                goto Step_GetPickingItemInfo;
            }

            if (targetLuid.HasValue &&
              !createMasterSSCC)
            {
                m_sboProviderService.InvokeMethodWithDbConnection<object>(false, false, null, null, delegate (PmxDbConnection dbConn, object[] parameters)
                {
                    DbCommand dbCmd = dbConn.CreateCommand(BuildQuery.SetLogUnitReadyForShipping(targetLuid.Value, true, dbConn.DbTool));
                    dbCmd.ExecuteNonQuery();

                    return null;
                });
            }
            else if (masterLUIDFull &&
                targetMasterLuid.HasValue)
            {
                m_sboProviderService.InvokeMethodWithDbConnection<object>(false, false, null, null, delegate (PmxDbConnection dbConn, object[] parameters)
                {
                    DbCommand dbCmd = dbConn.CreateCommand(BuildQuery.SetMasterLogUnitReadyForShipping(targetMasterLuid.Value, true, dbConn.DbTool));
                    dbCmd.ExecuteNonQuery();

                    return null;
                });
            }



            m_subLUID = -1;
            if (targetLuid.HasValue &&
                !targetMasterLuid.HasValue)
            {
                m_subLUID = targetLuid.Value;
            }

            if (targetMasterLuid.HasValue &&
                masterLUIDFull)
            {
                //Master LUID is full, so now ask measures for the sub flow
                m_subLUID = targetMasterLuid.Value;
            }

            if(m_subLUID != -1)
            {

                try
                {

                    //call flow
                    this.StartSubFlowAndWait(m_enterLUIDDataFlow, true);
                }
                catch (Exception ex)
                {
                    s_log.Error("", ex);
                    if (!(ex is ProdumexException) || ((ProdumexException)ex).Error == null)
                    {
                        // Inform the user with the error
                        session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
                            DefaultCultureInfo.Get(), BuildParamCollection(
                                "MessageKey", ExceptionHelper.Format(ExceptionHelper.Content.InnerExceptions, ex, DefaultCultureInfo.Get()),
                                "NoTranslationOfMessageKey", true,
                                "IsError", true,
                                "ShowButton", true));
                        WaitForMessage();

                    }
                    else
                    {
                        // Inform the user with the error
                        session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
                            DefaultCultureInfo.Get(), BuildParamCollection(
                                "PmxError", ((ProdumexException)ex).Error,
                                "ShowButton", true));
                        WaitForMessage();
                    }
                }
            

            #region START AFTER LOGISTIC UNIT FULL HOOKSCRIPT


                m_luidAfterLogisticUnitFullHookFlow = m_subLUID;

                m_afterLogisticUnitFullHookScript.BindParameters(
                    new ReadOnlyBinder<CultureInfo>("DefaultCultureInfo", this.DefaultCultureInfo),
                    new ReadOnlyBinder<PmxOseCompany>("PmxOseCompany", this.PmxOseCompany),
                    new ReadOnlyBinder<int>("LUID", this, "m_luidAfterLogisticUnitFullHookFlow")
                    );

                try
                {
                    //call flow
                    this.StartSubFlowAndWait(m_afterLogisticUnitFullHookScript, true);
                    m_luidAfterLogisticUnitFullHookFlow = 0;
                }
                catch (Exception ex)
                {
                    s_log.Error("", ex);
                    if (!(ex is ProdumexException) || ((ProdumexException)ex).Error == null)
                    {
                        // Inform the user with the error
                        session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
                            DefaultCultureInfo.Get(), BuildParamCollection(
                                "MessageKey", ExceptionHelper.Format(ExceptionHelper.Content.InnerExceptions, ex, DefaultCultureInfo.Get()),
                                "NoTranslationOfMessageKey", true,
                                "IsError", true,
                                "ShowButton", true));
                        WaitForMessage();

                    }
                    else
                    {
                        // Inform the user with the error
                        session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
                            DefaultCultureInfo.Get(), BuildParamCollection(
                                "PmxError", ((ProdumexException)ex).Error,
                                "ShowButton", true));
                        WaitForMessage();
                    }
                }

            #endregion

            }


            query = BuildQuery.GetSSCCInfo(targetLuid, targetMasterLuid, m_sboProviderService.GetDbTool());
            DataSet dsSSCC = m_sboProviderService.RunView(false, null, null, query);
            DataRow rowSSCC = dsSSCC.Tables[0].Rows[0];
            string infoSSCC = rowSSCC[SSCCPackingDataSet.Columns.SSCC.Name] as string;
            string infoCardCode = rowSSCC[SSCCPackingDataSet.Columns.CARD_CODE.Name] as string;
            string infoCardName = rowSSCC[SSCCPackingDataSet.Columns.CARD_NAME.Name] as string;
            string infoShippingDock = rowSSCC[SSCCPackingDataSet.Columns.LOCATION_CODE.Name] as string;

            if (!waveSameDelivery &&
                allowMultipleCustomersOnSSCC)
            {
                //no customer on label, because SSCC can contains items for multiple SSCC's
                infoCardCode = "";
                infoCardName = "";
            }

            Step_PrintOnLuidFull:

            #region PRINT ON LUID FULL

            int? luidToPrint = targetLuid;

            if (targetMasterLuid.HasValue &&
                pickingItemInfo.IsFullMasterSSCC)
            {
                luidToPrint = targetMasterLuid;
            }

         
            if (luidToPrint.HasValue)
            {
                nrOfDocsPrinted = this.ExecuteMethodWithDbConnection<int>(out pmxError, false, false,
                    delegate(PmxDbConnection conn, object[] parameters)
                    {
                        PmxPrintReportProvider printProv = new PmxPrintReportProvider(conn);

                        if (this.ProductionPicking)
                        {
                            return printProv.PrintReportsForEvent(PmxPrintReportEventType.PickingForProductionNewLogisticUnitFull,
                                dstLocationCode, session.DeviceInstanceId, luidToPrint.Value, out askReprint);
                        }
                        else
                        {
                            return printProv.PrintReportsForEvent(PmxPrintReportEventType.PickingNewLogisticUnitFull,
                                dstLocationCode, session.DeviceInstanceId, luidToPrint.Value, out askReprint);
                        }
                    });

            }

            //Print master LUID
            if (targetMasterLuid.HasValue &&
                masterLUIDFull)
            {
                luidToPrint = targetMasterLuid;

                if (luidToPrint.HasValue)
                {
                    nrOfDocsPrinted = this.ExecuteMethodWithDbConnection<int>(out pmxError, false, false,
                     delegate (PmxDbConnection conn, object[] parameters)
                     {
                         PmxPrintReportProvider printProv = new PmxPrintReportProvider(conn);

                         if (this.ProductionPicking)
                         {
                             return printProv.PrintReportsForEvent(PmxPrintReportEventType.PickingForProductionNewLogisticUnitFull,
                                 dstLocationCode, session.DeviceInstanceId, luidToPrint.Value, out askReprint);
                         }
                         else
                         {
                             return printProv.PrintReportsForEvent(PmxPrintReportEventType.PickingNewLogisticUnitFull,
                                 dstLocationCode, session.DeviceInstanceId, luidToPrint.Value, out askReprint);
                         }
                     });

                }


            }



            bool showPalletInfo = true;

            if (nrOfDocsPrinted > 0 || pmxError != null)
            {

                if (pmxError != null)
                {
                    session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
                        this.DefaultCultureInfo, BuildParamCollection(
                        "PmxError", pmxError,
                        "ShowButton", true));
                    msg = WaitForMessage();

                    //in case of error for printing, still proceed like normal
                    //goto Step_SelectWave;
                }

                if (askReprint)
                {
                    showPalletInfo = false;

                    session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.Picking.IPickingDecisionScreen),
                        this.DefaultCultureInfo, BuildParamCollection(
                        "TitleKey", "MSG_TITLE_REPRINT_LABEL_PICKING",
                        "MessageKey", "MSG_QUESTION_REPRINT_LABEL",
                        "SSCC", infoSSCC,
                        "CardCode", infoCardCode,
                        "CardName", infoCardName,
                        "ShippingDock", infoShippingDock));
                    msg = WaitForMessage();

                    if (msg.Name == "Produmex.Foundation.SlimScreen.Interfaces.Picking.IPickingDecisionScreenReply.Yes")
                    {
                        goto Step_PrintOnLuidFull;
                    }
                }
            }

            if (showPalletInfo)
            {
                session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.Picking.IShowPickedPalletScreen),
                    this.DefaultCultureInfo, BuildParamCollection(
                    "TitleKey", "MSG_TITLE_PICKED_PALLET",
                    "SSCC", infoSSCC,
                    "CardCode", infoCardCode,
                    "CardName", infoCardName,
                    "ShippingDock", infoShippingDock));
                msg = WaitForMessage();
            }

            #endregion

            Step_CountPickedColli:

            m_countPickedColliSubFlow.BindParameters(
                this.DefaultCultureInfo,
                this.PmxOseCompany,
                new ReadOnlyBinder<int>("WaveKey", waveKey),
                new ReadOnlyBinder<int>("PickListDocEntry", pickListDocEntry),
                new ReadOnlyBinder<int?>("LUID", targetLuid),
                new ReadWriteBinder<bool>("CanProceed", this, "m_subFlowSuccess"));
            m_subFlowSuccess = false;
            this.StartSubFlowAndWait(m_countPickedColliSubFlow, true);







            if (!m_subFlowSuccess)
            {
                goto Step_SelectWave;
            }




            // Clean up LU info
            targetLuid = null;
            if (masterLUIDFull)
            {
                targetMasterLuid = null;
                masterLUIDFull = false;
            }
            targetLogCarrierCode = null;
            targetLogCarrierCanBeFull = false;

            goto Step_GetPickingItemInfo;

            #endregion

        Step_SkipItem:

            #region SKIP ITEM

            query = BuildQuery.GetPickReasons();
            DataSet dsReasons = m_sboProviderService.RunView(false, null, null, query);

            reasonsCodesRequiresFreeText = new Dictionary<string, bool>();
            foreach (DataRow row in dsReasons.Tables[0].Rows)
            {
                reasonsCodesRequiresFreeText.Add(
                    row[ReasonDataSet.Columns.REASON_CODE.Name].ToString(),
                    (SboBooleanConvertor.ObjectToBooleanOrNull(row[ReasonDataSet.Columns.REASON_REQUIRES_FREE_TEXT.Name]).HasValue ? (SboBooleanConvertor.ObjectToBooleanOrNull(row[PmxReasonTable.Columns.RequiresFreeText.NAME])).Value : false)
                    );
            }

            session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.ISelectReasonScreen),
                this.DefaultCultureInfo, BuildParamCollection(
                "InitialErrorKey", initialErrorKey,
                "TitleKey", "MSG_TITLE_SELECT_REASON",
                "MessageKey", "MSG_SELECT_PICKING_REASON",
                "ReasonDS", dsReasons));
            msg = WaitForMessage();
            initialErrorKey = null;

            if (msg.Name != "Produmex.Foundation.SlimScreen.Interfaces.ISelectReasonScreenReply.ReasonSelected")
            {
                goto Step_OtherTask;
            }

            reasonCode = ExtractParameter<string>(msg.Parameters, "reasonCode");

            DataRow[] temp_rows = dsReasons.Tables[0].Select(String.Format("{0} = '{1}'", ReasonDataSet.Columns.REASON_CODE.Name, reasonCode));
            if (temp_rows.Length > 0)
            {
                reasonName = temp_rows[0][ReasonDataSet.Columns.REASON_NAME.Name] as string;
            }

            #region ENTER REASON FREE TEXT

            if (reasonsCodesRequiresFreeText.ContainsKey(reasonCode) &&
                reasonsCodesRequiresFreeText[reasonCode])
            {


                // Show screen with list
                session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IEnterStringValueScreen),
                    DefaultCultureInfo.Get(), BuildParamCollection(
                        "InitialErrorKey", initialErrorKey,
                        "TitleKey", "MSG_TITLE_ENTER_REASON",
                        "Information", "MSG_ENTER_FREE_REASON_DEFAULT",
                        "Parameters", new object[] { reasonName },
                        "AllowToGoBack", true,
                        "ForceDataEntry", true,
                        "AllowMultiLine", true,
                        "MinimumNumberOfCharacters", Definitions.MIMIMUM_NUMBER_OF_CHARACTERS_FOR_REASON
                        ));

                msg = WaitForMessage();

                initialErrorKey = null;

                if (msg.Name.EndsWith(".StringEntered"))
                {
                    reasonFreeText = ExtractParameter<string>(msg.Parameters, "stringValue");

                }
                else
                {
                    //go back
                    goto Step_SkipItem;
                }
            }
            #endregion



            this.ExecuteMethodWithDbConnection<object>(out pmxError, false, false,
                delegate(PmxDbConnection conn, object[] parameters)
                {
                    PmxPickListProvider plProv = new PmxPickListProvider(conn);
                    plProv.SetPickListLineToPicked(pickListDocEntry, pickingItemInfo.PickListLineID, new ReasonInfo(reasonCode, reasonFreeText));
                    //plProv.ForceCloseLine(pickListDocEntry, new Collection<int>() { pickingItemInfo.PickListLineID });

                    return null;
                });
            if (pmxError != null)
            {
                session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
                    this.DefaultCultureInfo, BuildParamCollection(
                    "PmxError", pmxError,
                    "ShowButton", true));
                msg = WaitForMessage();

                goto Step_SkipItem;
            }

            alternateItemMode = false;

            goto Step_GetPickingItemInfo;

            #endregion

        Step_PickFullPallet:

            #region PICK FULL SSCC


            if (targetLuid.HasValue && String.IsNullOrEmpty(targetMoveableLocation))
            {
                // We are using an LU
                session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IDecisionScreen),
                    this.DefaultCultureInfo, BuildParamCollection(
                    "TitleKey", "MSG_TITLE_PRINT_PICK_LABEL",
                    "MessageKey", "MSG_QUESTION_PRINT_PICK_LABEL"));
                msg = WaitForMessage();
                if (msg.Name == "Produmex.Foundation.SlimScreen.Interfaces.IDecisionScreenReply.Yes")
                {
                    goto Step_LuidFull;
                }

                goto Step_CountPickedColli;
            }

            // Clean up obsolete LU info
            targetLuid = null;
            if (masterLUIDFull)
            {
                targetMasterLuid = null;
            }
            targetLogCarrierCode = null;
            targetLogCarrierCanBeFull = false;

        Step_FullPalletIdentifySSCC:


            session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.Picking.IIdentifyFullPalletPickingScreen),
                this.DefaultCultureInfo, BuildParamCollection(
                "InitialErrorKey", initialErrorKey,
                "TitleKey", "MSG_TITLE_IDENTIFY_SSCC",
                "SSCC", pickingItemInfo.SSCC,
                "LocationCode", pickingItemInfo.LocationCode,
                "ItemCode", pickingItemInfo.ItemCode,
                "ItemDescription", pickingItemInfo.CustomItemDescription,
                "ItemBarcode", pickingItemInfo.ItemBarcode,
                "ImagePath", itemInfo.PictureName,
                "StillToPick", pickingItemInfo.Quantity,  // take from proposedPickingItemInfo ?
                "Uom", pickingItemInfo.Uom,
                "BatchNumber", pickingItemInfo.BatchNumber,
                "BestBeforeDate", pickingItemInfo.BestBeforeDate,
                "ShowFullNextItemData", true,
                "AllowedButtons", new string[] { "MSG_BUTTON_OTHER_TASK" }));
            msg = WaitForMessage();
            initialErrorKey = null;

            if (msg.Name == "Produmex.Foundation.SlimScreen.Interfaces.Picking.IIdentifyFullPalletPickingScreenReply.ButtonClicked")
            {
                goto Step_OtherTask;
            }
            if (msg.Name != "Produmex.Foundation.SlimScreen.Interfaces.Picking.IIdentifyFullPalletPickingScreenReply.SSCCEntered")
            {
                goto Step_ShowNextItems;
            }

            //Check if SSCC is on valid location
            if (string.IsNullOrEmpty(pickingItemInfo.LocationCode))
            {
                initialErrorKey = "MSG_ERROR_PICK_LOCATION_UNKNOWN_FOR_FULL_PALLET";
                goto Step_FullPalletIdentifySSCC;
            }

            string enteredSSCC = ExtractParameter<string>(msg.Parameters, "sscc");
            if (enteredSSCC != pickingItemInfo.SSCC)
            {
                initialErrorKey = "MSG_ERROR_INCORRECT_SSCC_TO_PICK";
                goto Step_FullPalletIdentifySSCC;
            }

            targetLuid = pickingItemInfo.LogisticUnitIdentKey;

            goto Step_SetPickingItemInfo;

            #endregion

        Step_PickFullMasterSSCC:

            #region PICK FULL MASTER SSCC


            if (targetLuid.HasValue && String.IsNullOrEmpty(targetMoveableLocation))
            {
                // We are using an LU
                session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IDecisionScreen),
                    this.DefaultCultureInfo, BuildParamCollection(
                    "TitleKey", "MSG_TITLE_PRINT_PICK_LABEL",
                    "MessageKey", "MSG_QUESTION_PRINT_PICK_LABEL"));
                msg = WaitForMessage();
                if (msg.Name == "Produmex.Foundation.SlimScreen.Interfaces.IDecisionScreenReply.Yes")
                {
                    goto Step_LuidFull;
                }

                goto Step_CountPickedColli;
            }

            // Clean up obsolete LU info
            targetLuid = null;
            targetMasterLuid = null;
            targetLogCarrierCode = null;
            targetLogCarrierCanBeFull = false;

        Step_FullMasterSSCCIdentifySSCC:


            DataSet dsSubItems = BuildAlternateList(pickingItemInfo.ListOfPickingItemInfos);


            session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.Picking.IIdentifyFullMasterSSCCPickingScreen),
                this.DefaultCultureInfo, BuildParamCollection(
                "InitialErrorKey", initialErrorKey,
                "TitleKey", "MSG_TITLE_IDENTIFY_MASTER_SSCC",
				"BarcodeController", barcodeController,
                "MasterSSCC", pickingItemInfo.MasterSSCC,
                "LocationCode", pickingItemInfo.LocationCode,
                "DSSubItems", dsSubItems,
                "AllowedButtons", new string[] { "MSG_BUTTON_OTHER_TASK" }
                ));
            msg = WaitForMessage();
            initialErrorKey = null;

            if (msg.Name == "Produmex.Foundation.SlimScreen.Interfaces.Picking.IIdentifyFullMasterSSCCPickingScreenReply.ButtonClicked")
            {
                goto Step_OtherTask;
            }
            if (msg.Name != "Produmex.Foundation.SlimScreen.Interfaces.Picking.IIdentifyFullMasterSSCCPickingScreenReply.SSCCEntered")
            {
                goto Step_ShowNextItems;
            }

            //Check if SSCC is on valid location
            if (string.IsNullOrEmpty(pickingItemInfo.LocationCode))
            {
                initialErrorKey = "MSG_ERROR_PICK_LOCATION_UNKNOWN_FOR_FULL_PALLET";
                goto Step_FullMasterSSCCIdentifySSCC;
            }

            string enteredMasterSSCC = ExtractParameter<string>(msg.Parameters, "sscc");
            if (enteredMasterSSCC != pickingItemInfo.MasterSSCC)
            {
                bool foundSubSSCC = false;
                //check if we find a sub SSCC that has been scanned
                foreach (PickingItemInfo tempInfo in pickingItemInfo.ListOfPickingItemInfos)
                {
                    if (enteredMasterSSCC == tempInfo.SSCC)
                    {
                        foundSubSSCC = true;
                        targetLuid = tempInfo.LogisticUnitIdentKey;
                        break;
                    }
                }
                if (!foundSubSSCC)
                {
                    initialErrorKey = "MSG_ERROR_INCORRECT_SSCC_TO_PICK";
                    goto Step_FullMasterSSCCIdentifySSCC;
                }
            }
            else
            {
                targetLuid = null;
                targetMasterLuid = pickingItemInfo.MasterLUID;
            }

            goto Step_SetPickingItemInfo;

            #endregion

        Step_IsPickListPicked:

            #region IS PICK LIST PICKED AND/OR PACKED


            //bool isPickListClosed = m_sboProviderService.InvokeMethodWithDbConnection<bool>(false, false, null, null,
            // delegate(PmxDbConnection conn, object[] parameters)
            // {
            //     PmxPickListProvider plProv = new PmxPickListProvider(conn);
            //     return plProv.IsPickListClosed(pickListDocEntry);
            // });

            //if (isPickListClosed)
            //{
            //    initialErrorKey = "MSG_ERROR_PICKLIST_CLOSED_DURING_PICKING";
            //    goto Step_LogStopWave;
            //}

			bool tempAskReprint = false;

            bool isPickListPicked = m_sboProviderService.InvokeMethodWithDbConnection<bool>(false, false, null, null,
                delegate(PmxDbConnection conn, object[] parameters)
                {
                    PmxPickListProvider plProv = new PmxPickListProvider(conn);
                    return plProv.IsPickListPicked(pickListDocEntry);
                });




            if (targetLuid.HasValue && String.IsNullOrEmpty(targetMoveableLocation))
            {
                // We are using an LU, so first execute code "LU is full" (and come back to this step afterwards)
                goto Step_LuidFull;
            }

            if (!isPickListPicked)
            {
                initialErrorKey = "MSG_ERROR_NO_ITEM_TO_PICK";
                goto Step_LogStopWave;
                //goto Step_SelectWave;
            }


            if (this.CrossDockingWaveKey != null)
            {
                crossDockingCompleted = true;
            }

            if (isProductionPicking)
            {
                //No shipping needed for production pick lists
                goto Step_LogStopWave;
            }

            bool isPickListPacked = m_sboProviderService.InvokeMethodWithDbConnection<bool>(false, false, null, null,
                delegate(PmxDbConnection conn, object[] parameters)
                {
                    PmxPickListProvider plProv = new PmxPickListProvider(conn);
                    return plProv.IsPickListPacked(pickListDocEntry);
                });

            if (!isPickListPacked)
            {
                goto Step_LogStopWave;
                //goto Step_SelectWave;
            }

		Step_PrintPackList:
            bool isWavePacked = m_sboProviderService.InvokeMethodWithDbConnection<bool>(false, false, null, null,
                delegate(PmxDbConnection conn, object[] parameters)
                {
                    PmxPickListProvider plProv = new PmxPickListProvider(conn);
                    return plProv.IsWavePacked(waveKey);
                });

            if (!isWavePacked)
            {
                goto Step_LogStopWave;
                //goto Step_SelectWave;
            }

			PmxPrintReportEventType printEvent = PmxPrintReportEventType.PickingAfterWavePacked;

			//execute print event
			nrOfDocsPrinted += this.ExecuteMethodWithDbConnection<int>(out pmxError, true, false,
			  delegate(PmxDbConnection conn, object[] parameters)
			  {
				  PmxPrintReportProvider printProv = new PmxPrintReportProvider(conn);
				  return printProv.PrintReportsForEvent(printEvent,
					  dstLocationCode, session.DeviceInstanceId, waveKey, out tempAskReprint);
			  });
			if (pmxError != null)
			{
				session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
					this.DefaultCultureInfo, BuildParamCollection(
					"PmxError", pmxError,
					"ShowButton", true));
				msg = WaitForMessage();

				//ask user if he wants to try again after error
				session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IDecisionScreen),
					DefaultCultureInfo.Get(), BuildParamCollection(
						"TitleKey", "MSG_TITLE_ERROR_TRY_AGAIN",
						"MessageKey", "MSG_QUESTION_ERROR_TRY_AGAIN"));
				msg = WaitForMessage();

				if (msg.Name.EndsWith(".Yes"))
				{
					goto Step_PrintPackList;
				}
			}

			if (tempAskReprint)
			{
				askReprint = true;

				session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IDecisionScreen),
					DefaultCultureInfo.Get(), BuildParamCollection(
						"TitleKey", "MSG_TITLE_REPRINT_DOCUMENT",
						"MessageKey", "MSG_QUESTION_REPRINT_DOCUMENT"
																	  ));
				msg = WaitForMessage();

				if (msg.Name.EndsWith(".Yes"))
				{
					goto Step_PrintPackList;
				}
			}
			
            bool areAllPickListsForAllSSCCsPacked = m_sboProviderService.InvokeMethodWithDbConnection<bool>(false, false, null, null,
                delegate(PmxDbConnection conn, object[] parameters)
                {
                    PmxPickListProvider plProv = new PmxPickListProvider(conn);
                    return plProv.AreAllPickListsForAllSSCCsPacked(waveKey, null);
                });

            if (!areAllPickListsForAllSSCCsPacked)
            {
                goto Step_LogStopWave;
                //goto Step_SelectWave;
            }

            //m_subFlowAutomaticShipping = m_sboProviderService.InvokeMethodWithDbConnection<bool>(false, false, null, null,
            //    delegate(PmxDbConnection conn, object[] parameters)
            //    {
            //        PmxPickListProvider plProv = new PmxPickListProvider(conn);
            //        return plProv.IsPickListForAutomaticShipping(pickListDocEntry);
            //    });

            if (crossDockingCompleted)
            {
                m_subFlowAutomaticShipping = true;
            }
            else
            {

                m_subFlowAutomaticShipping = m_sboProviderService.InvokeMethodWithDbConnection<bool>(false, false, null, null,
                   delegate(PmxDbConnection conn, object[] parameters)
                   {
                       PmxPickListProvider plProv = new PmxPickListProvider(conn);
                       return plProv.IsWaveForAutomaticShipping(waveKey);
                   });
            }


            Collection<int> packedPickListDocEntries = m_sboProviderService.InvokeMethodWithDbConnection<Collection<int>>(false, false, null, null,
              delegate(PmxDbConnection conn, object[] parameters)
              {
                  PmxPickListProvider plProv = new PmxPickListProvider(conn);
                  return plProv.GetListOfPackedPickListsOfWave(waveKey, null);
              });


            listOfPickListsToBeAutoShipped = packedPickListDocEntries;


            Dictionary<string, object> groupedDeliveryData = new Dictionary<string, object>();

            //Clear the data
            m_subFlowDeliveryData = null;

            foreach (int currentPickListDocEntry in packedPickListDocEntries)
            {
                m_afterPickListPackedHookFlow.BindParameters(
                    this.DefaultCultureInfo,
                    this.PmxOseCompany,
                    new ReadOnlyBinder<int>("PickListDocEntry", currentPickListDocEntry),
                    new ReadWriteBinder<bool>("DoAutomaticShipping", this, "m_subFlowAutomaticShipping"),
                    new ReadWriteBinder<Dictionary<string, object>>("DeliveryData", this, "m_subFlowDeliveryData")
                );
                //m_subFlowDeliveryData = null;
                this.StartSubFlowAndWait(m_afterPickListPackedHookFlow, true);

                if (!m_subFlowAutomaticShipping)
                {
                    //save data because we will not use it this time on automatic shipping
                    if (m_subFlowDeliveryData != null)
                    {
                        m_sboProviderService.InvokeMethodWithDbConnection<object>(false, false, null, null,
                            delegate(PmxDbConnection conn, object[] parameters)
                            {
                                PmxPickListProvider plProv = new PmxPickListProvider(conn);
                                plProv.UpdateDeliveryData(currentPickListDocEntry, m_subFlowDeliveryData, true);

                                return null;
                            });
                    }
                }
                else
                {
                    if (m_subFlowDeliveryData != null)
                    {
                        foreach (string key in m_subFlowDeliveryData.Keys)
                        {
                            if (!groupedDeliveryData.ContainsKey(key))
                            {
                                groupedDeliveryData.Add(key, m_subFlowDeliveryData[key]);
                            }
                        }
                    }
                }

            }

            if (!m_subFlowAutomaticShipping)
            {
                goto Step_LogStopWave;
                //goto Step_SelectWave;
            }

            //set grouped delivery data to variable
            m_subFlowDeliveryData = groupedDeliveryData;
            if (m_subFlowDeliveryData != null &&
                m_subFlowDeliveryData.Keys.Count == 0)
            {
                m_subFlowDeliveryData = null;
            }

            session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
                this.DefaultCultureInfo, BuildParamCollection(
                "MessageKey", "MSG_ADDING_SALES_DELIVERIES"));

            goto Step_EnterDeliveryHeaderData;

            #endregion

        Step_EnterDeliveryHeaderData:

            #region ENTER DELIVERY DATA SUB FLOW

            doRetry = false;

            do
            {
                try
                {
                    m_backRequestedEnterSalesDeliveryDataSubFlow = false;

                    m_waveKeyEnterSalesDeliveryDataSubFlow = waveKey;
                    m_routeKeyEnterSalesDeliveryDataSubFlow = null;
                    m_picklistEntriesEnterSalesDeliveryDataSubFlow = listOfPickListsToBeAutoShipped;
                    m_deliveryHeaderDataEnterSalesDeliveryDataSubFlow = new Dictionary<string, object>();

                    m_enterSalesDeliveryDataSubFlow.BindParameters(
                       new ReadOnlyBinder<CultureInfo>("DefaultCultureInfo", this.DefaultCultureInfo),
                       new ReadOnlyBinder<PmxOseCompany>("PmxOseCompany", this.PmxOseCompany),
                       new ReadOnlyBinder<Collection<int>>("PickListEntries", m_picklistEntriesEnterSalesDeliveryDataSubFlow),
                       new ReadOnlyBinder<Collection<int>>("SalesDocEntries", null),
                       new ReadOnlyBinder<int?>("WaveKey", m_waveKeyEnterSalesDeliveryDataSubFlow),
                       new ReadOnlyBinder<int?>("RouteKey", m_routeKeyEnterSalesDeliveryDataSubFlow),
                       new ReadWriteBinder<bool>("BackRequested", this, "m_backRequestedEnterSalesDeliveryDataSubFlow"),
                       new ReadWriteBinder<Dictionary<string, object>>("DeliveryData", this, "m_deliveryHeaderDataEnterSalesDeliveryDataSubFlow")
                       );

                    this.StartSubFlowAndWait(m_enterSalesDeliveryDataSubFlow, true);

                    if (m_subFlowDeliveryData == null)
                    {
                        m_subFlowDeliveryData = new Dictionary<string, object>();
                    }
                    foreach (KeyValuePair<string, object> pair in m_deliveryHeaderDataEnterSalesDeliveryDataSubFlow)
                    {
                        m_subFlowDeliveryData.Add(pair.Key, pair.Value);
                    }

                    goBack = m_backRequestedEnterSalesDeliveryDataSubFlow;


                }
                catch (Exception ex)
                {
                    s_log.Error("", ex);
                    if (!(ex is ProdumexException) || ((ProdumexException)ex).Error == null)
                    {
                        // Inform the user with the error
                        session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
                            DefaultCultureInfo.Get(), BuildParamCollection(
                                "MessageKey", ExceptionHelper.Format(ExceptionHelper.Content.InnerExceptions, ex, DefaultCultureInfo.Get()),
                                "NoTranslationOfMessageKey", true,
                                "IsError", true,
                                "ShowButton", true));
                        WaitForMessage();


                    }
                    else
                    {
                        // Inform the user with the error
                        session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
                            DefaultCultureInfo.Get(), BuildParamCollection(
                                "PmxError", ((ProdumexException)ex).Error,
                                "ShowButton", true));
                        WaitForMessage();

                    }
                    //ask user if he wants to try again
                    session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IDecisionScreen),
                        DefaultCultureInfo.Get(), BuildParamCollection(
                            "TitleKey", "MSG_TITLE_RETRY_UPDATE",
                            "MessageKey", "MSG_QUESTION_AN_ERROR_OCCURED_WANT_RETRY"));
                    msg = WaitForMessage();

                    if (msg.Name.ToLower().EndsWith(".yes"))
                    {
                        doRetry = true;
                    }
                }
            }
            while (doRetry);


            #endregion


        Step_AddDeliveryAgain:

            #region ADD DELIVERY

            DocumentRef[] deliveryDocRefs = this.ExecuteMethodWithDbConnection<DocumentRef[]>(out pmxError, true, false,
                delegate(PmxDbConnection conn, object[] parameters)
                {
                    PmxSalesDeliveryNoteProvider sdnProv = new PmxSalesDeliveryNoteProvider(conn);
                    return sdnProv.ShipAllItemsForPickList(listOfPickListsToBeAutoShipped, m_subFlowDeliveryData, null);
                    //return sdnProv.ShipAllItemsForWave(waveKey, m_subFlowDeliveryData);
                });


            if (pmxError != null)
            {
                session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
                    this.DefaultCultureInfo, BuildParamCollection(
                    "PmxError", pmxError,
                    "ShowButton", true));
                msg = WaitForMessage();

                session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IDecisionScreen),
                       DefaultCultureInfo.Get(), BuildParamCollection(
                           "TitleKey", "MSG_TITLE_ERROR_TRY_AGAIN",
                           "MessageKey", "MSG_QUESTION_ERROR_TRY_AGAIN"));
                msg = WaitForMessage();

                if (msg.Name.EndsWith(".Yes"))
                {
                    goto Step_AddDeliveryAgain;
                }

                goto Step_LogStopWave;
                //goto Step_SelectWave;
            }

            //waveKey = 0;


            //delivery hook flow
            //sales delivery hook flow
            m_documentRefsSalesDeliveryCreatedHookFlow = deliveryDocRefs;

            try
            {
                //call flow
                this.StartSubFlowAndWait(m_salesDeliveryCreatedHookFlow, true);
            }
            catch (Exception ex)
            {
                s_log.Error("", ex);
                if (!(ex is ProdumexException) || ((ProdumexException)ex).Error == null)
                {
                    // Inform the user with the error
                    session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
                        DefaultCultureInfo.Get(), BuildParamCollection(
                            "MessageKey", ExceptionHelper.Format(ExceptionHelper.Content.InnerExceptions, ex, DefaultCultureInfo.Get()),
                            "NoTranslationOfMessageKey", true,
                            "IsError", true,
                            "ShowButton", true));
                    WaitForMessage();

                }
                else
                {
                    // Inform the user with the error
                    session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
                        DefaultCultureInfo.Get(), BuildParamCollection(
                            "PmxError", ((ProdumexException)ex).Error,
                            "ShowButton", true));
                    WaitForMessage();
                }
            }

            try
            {
                m_subFlowPickListEntry = pickListDocEntry;

                m_printDeliveryDocumentsSubFlow = this.Create("PrintSalesDeliveryDocumentsScript");
                m_printDeliveryDocumentsSubFlow.BindParameters(
                    this.DefaultCultureInfo,
                    this.PmxOseCompany,
                    new ReadOnlyBinder<DocumentRef[]>("ListOfDeliveries", deliveryDocRefs),
                    new ReadWriteBinder<int?>("PickListEntry", this, "m_subFlowPickListEntry"),
                    new ReadOnlyBinder<string>("LocationCode", dstLocationCode));
                this.StartSubFlowAndWait(m_printDeliveryDocumentsSubFlow, true);
            }
            catch (Exception ex)
            {
                s_log.Error("", ex);
                if (!(ex is ProdumexException) || ((ProdumexException)ex).Error == null)
                {
                    // Inform the user with the error
                    session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
                        DefaultCultureInfo.Get(), BuildParamCollection(
                            "MessageKey", ExceptionHelper.Format(ExceptionHelper.Content.InnerExceptions, ex, DefaultCultureInfo.Get()),
                            "NoTranslationOfMessageKey", true,
                            "IsError", true,
                            "ShowButton", true));
                    WaitForMessage();

                }
                else
                {
                    // Inform the user with the error
                    session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
                        DefaultCultureInfo.Get(), BuildParamCollection(
                            "PmxError", ((ProdumexException)ex).Error,
                            "ShowButton", true));
                    WaitForMessage();
                }
            }


            #endregion

        Step_AutoCreateInvoice:

            #region AUTO CREATE INVOICE

            //auto create invoice
            Collection<DocumentRef> listOfSalesInvoices = null;

            DocumentRef[] deliveryDocRefsAutomaticInvoicing = m_sboProviderService.InvokeMethodWithDbConnection<DocumentRef[]>(true, false, null, null,
            delegate(PmxDbConnection conn, object[] parameters)
            {
                PmxSalesDeliveryNoteProvider sdProv = new PmxSalesDeliveryNoteProvider(conn);
                return sdProv.GetDocumentRefsForAutomaticInvoice(deliveryDocRefs);
            });


            if (deliveryDocRefsAutomaticInvoicing != null &&
                deliveryDocRefsAutomaticInvoicing.Length > 0)
            {

                session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
                    this.DefaultCultureInfo, BuildParamCollection(
                    "MessageKey", "MSG_ADDING_SALES_INVOICE"));

                listOfSalesInvoices = this.ExecuteMethodWithDbConnection<Collection<DocumentRef>>(out pmxError, true, false,
                 delegate(PmxDbConnection conn, object[] parameters)
                 {
                     PmxSalesInvoiceProvider prov = new PmxSalesInvoiceProvider(conn);
                     return prov.CreateInvoiceFromDelivery(deliveryDocRefsAutomaticInvoicing);
                 });
            }

            if (pmxError != null)
            {
                session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
                    this.DefaultCultureInfo, BuildParamCollection(
                    "PmxError", pmxError,
                    "ShowButton", true));
                msg = WaitForMessage();

                session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IDecisionScreen),
                      DefaultCultureInfo.Get(), BuildParamCollection(
                          "TitleKey", "MSG_TITLE_ERROR_TRY_AGAIN",
                          "MessageKey", "MSG_QUESTION_ERROR_TRY_AGAIN"));
                msg = WaitForMessage();

                if (msg.Name.EndsWith(".Yes"))
                {
                    goto Step_AutoCreateInvoice;
                }
            }

            goto Step_PrintInvoiceDocuments;

            #endregion

        Step_PrintInvoiceDocuments:

            #region PRINT SALES INVOICE DOCUMENTS

            if (listOfSalesInvoices != null &&
                listOfSalesInvoices.Count > 0)
            {
                try
                {
                    m_printInvoiceDocumentsSubFlow = this.Create("PrintSalesInvoiceDocumentsScript");
                    m_printInvoiceDocumentsSubFlow.BindParameters(
                        this.DefaultCultureInfo,
                        this.PmxOseCompany,
                        new ReadOnlyBinder<Collection<DocumentRef>>("DocRefs", listOfSalesInvoices),
                        new ReadOnlyBinder<string>("LocationCode", dstLocationCode));
                    this.StartSubFlowAndWait(m_printInvoiceDocumentsSubFlow, true);
                }
                catch (Exception ex)
                {
                    s_log.Error("", ex);
                    if (!(ex is ProdumexException) || ((ProdumexException)ex).Error == null)
                    {
                        // Inform the user with the error
                        session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
                            DefaultCultureInfo.Get(), BuildParamCollection(
                                "MessageKey", ExceptionHelper.Format(ExceptionHelper.Content.InnerExceptions, ex, DefaultCultureInfo.Get()),
                                "NoTranslationOfMessageKey", true,
                                "IsError", true,
                                "ShowButton", true));
                        WaitForMessage();

                    }
                    else
                    {
                        // Inform the user with the error
                        session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
                            DefaultCultureInfo.Get(), BuildParamCollection(
                                "PmxError", ((ProdumexException)ex).Error,
                                "ShowButton", true));
                        WaitForMessage();
                    }
                }

                goto Step_SalesInvoiceCreatedHook;

            }

            goto Step_LogStopWave;
        //goto Step_SelectWave;

            #endregion

        Step_SalesInvoiceCreatedHook:

            #region SALES INVOICE CREATED HOOK

            //sales invoice hook flow
            m_documentRefsSalesInvoiceCreatedHookFlow = listOfSalesInvoices;

            try
            {
                //call flow
                this.StartSubFlowAndWait(m_salesInvoiceCreatedHookFlow, true);
            }
            catch (Exception ex)
            {
                s_log.Error("", ex);
                if (!(ex is ProdumexException) || ((ProdumexException)ex).Error == null)
                {
                    // Inform the user with the error
                    session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
                        DefaultCultureInfo.Get(), BuildParamCollection(
                            "MessageKey", ExceptionHelper.Format(ExceptionHelper.Content.InnerExceptions, ex, DefaultCultureInfo.Get()),
                            "NoTranslationOfMessageKey", true,
                            "IsError", true,
                            "ShowButton", true));
                    WaitForMessage();

                }
                else
                {
                    // Inform the user with the error
                    session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
                        DefaultCultureInfo.Get(), BuildParamCollection(
                            "PmxError", ((ProdumexException)ex).Error,
                            "ShowButton", true));
                    WaitForMessage();
                }
            }

            goto Step_LogStopWave;
        //goto Step_SelectWave;

            #endregion

        Step_Alternate_IdentifySSCCOrLocation:

            #region ALTERNATE: IDENTIFY SSCC OR LOCATION

            //alternatePickLocation = null;
            alternateSSCC = null;
            alternateLUID = null;
            //alternateLUIDLocationCode = null;
            alternateIsMonoLotPallet = false;
            //alternateItri = null;


            session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IIdentifyGS1DataScreen),
                DefaultCultureInfo.Get(), BuildParamCollection(
                "InitialErrorKey", initialErrorKey,
                "TitleKey", "MSG_TITLE_AD_HOC_ALTERNATE_PICKING_SCAN_SSCC_OR_LOCATION",
                "BarcodeController", barcodeController,
                "AllowedButtons", new string[1] { "MSG_BUTTON_SELECT_LOCATION" },
                "ForceGS1_128", false
                ));

            msg = WaitForMessage();

            //clear data
            initialErrorKey = null;

            if (msg.Name.EndsWith(".DataEntered"))
            {
                BarcodeContainer barcodeContainerLUID = ExtractParameter<BarcodeContainer>(msg.Parameters, "data");

                //check what has been scanned
                if (barcodeContainerLUID.GS1DataObject != null && !string.IsNullOrEmpty(barcodeContainerLUID.GS1DataObject.Sscc))
                {
                    //user scanned an SSCC, so we check if it is allowed
                    alternateSSCC = barcodeContainerLUID.GS1DataObject.Sscc;
                    goto Step_CheckAlternateSSCC;
                }

                string tempLocation = BarcodeHelper.GetProductIdentification(barcodeContainerLUID);

                if (!string.IsNullOrEmpty(tempLocation) &&
                    Regex.IsMatch(tempLocation, "^00[0-9]{18}$"))
                {
                    alternateSSCC = tempLocation.Substring(2, 18);
                    goto Step_CheckAlternateSSCC;
                }

                if (string.IsNullOrEmpty(tempLocation))
                {
                    initialErrorKey = "MSG_ERROR_BARCODE_DOES_NOT_CONTAIN_SSCC_OR_LOCATION";
                    goto Step_Alternate_IdentifySSCCOrLocation;
                }
                else
                {
                    enteredLocationCode = tempLocation;

                    goto Step_SelectLocationAlternate;
                }

            }
            else if (msg.Name.EndsWith(".ButtonClicked"))
            {
                //select location
                enteredLocationCode = null;
                goto Step_SelectLocationAlternate;
            }



            //go back
            alternateItemMode = false;
            goto Step_OtherTask;


            #endregion

        Step_CheckAlternateSSCC:

            #region CHECK ALTERNATE SSCC



            //get the SSCC info
            query = BuildQuery.GetProductsFromStockOrderByName( null, itemInfo.ItemCode, null, alternateSSCC, false, null, null, null, null, null, false, Produmex.Sbo.Logex.Data.SboCommands.PackagingUnit.Inventory, m_sboProviderService.GetDbTool() );
            dsItemsOnSSCC = m_sboProviderService.RunView(true, null, null, query);

            if (this.GetNrOfRows(dsItemsOnSSCC) == 0)
            {
                //Product not found on SSCC
                initialErrorKey = "MSG_ERROR_SSCC_AD_HOC_ALTERNATE_PICKING_NO_PRODUCT";
                goto Step_Alternate_IdentifySSCCOrLocation;
            }



            //Check supplier pallet number
            if (!string.IsNullOrEmpty(supplierPalletNumber))
            {
                query = BuildQuery.GetProductsFromStockOrderByName(null, itemInfo.ItemCode, null, alternateSSCC, false, null, null, null, null, null, false, Produmex.Sbo.Logex.Data.SboCommands.PackagingUnit.Inventory, false, supplierPalletNumber,m_sboProviderService.GetDbTool());
                if (this.GetNrOfRows(m_sboProviderService.RunView(true, null, null, query)) == 0)
                {
                    //Supplier pallet number does not correspond
                    initialErrorKey = "MSG_ERROR_SSCC_AD_HOC_ALTERNATE_PICKING_SUPPLIER_PALLET_NUMBER_INCORRECT";
                    goto Step_Alternate_IdentifySSCCOrLocation;
                }
            }


            //get all quantity on SSCC for item (for all quality statuses
            double fullQuantityOnPallet = 0;
            foreach (DataRow row in dsItemsOnSSCC.Tables[0].Rows)
            {
                fullQuantityOnPallet += Convert.ToDouble(row[PmxStockOverviewOnLocationDefinition.Columns.Quantity.NAME]);
            }


            query = BuildQuery.GetProductsFromStockOrderByName( null, itemInfo.ItemCode, null, alternateSSCC, false, null, null, alternateCopyBatchNumbers ? pickingItemInfo.BatchNumber : null, alternateCopyBatchNumbers ? pickingItemInfo.BatchNumber2 : null, alternateCopyBatchNumbers ? pickingItemInfo.BestBeforeDate : null, alternateCopyBatchNumbers, Produmex.Sbo.Logex.Data.SboCommands.PackagingUnit.Inventory, true, null, m_sboProviderService.GetDbTool() );
            dsItemsOnSSCC = m_sboProviderService.RunView(true, null, null, query);

            if (this.GetNrOfRows(dsItemsOnSSCC) == 0)
            {
                //Batch not found on location
                initialErrorKey = "MSG_ERROR_SSCC_AD_HOC_ALTERNATE_PICKING_WRONG_BATCH_QUALITY";
                goto Step_Alternate_IdentifySSCCOrLocation;
            }


            alternateLUID = Convert.ToInt32(dsItemsOnSSCC.Tables[0].Rows[0][PmxStockOverviewOnLocationDefinition.Columns.LogUnitIdentKey.NAME]);
            enteredLocationCode = Convert.ToString(dsItemsOnSSCC.Tables[0].Rows[0][PmxStockOverviewOnLocationDefinition.Columns.DefaultLocationCode.NAME]);

            //check if it is a mono pallet
            alternateIsMonoLotPallet = m_sboProviderService.InvokeMethodWithDbConnection<bool>(false, false, null, null, new DelegateWithPmxDbConnection<bool>(ExecuteIsMonoLotPallet), alternateLUID);

            if (alternateIsMonoLotPallet)
            {
                double quantityOnPallet = 0;
                foreach (DataRow row in dsItemsOnSSCC.Tables[0].Rows)
                {
                    quantityOnPallet += Convert.ToDouble(row[PmxStockOverviewOnLocationDefinition.Columns.Quantity.NAME]);
                }

                if (quantityOnPallet <= pickingItemInfo.Quantity &&
                    forceUserToPickFullPallet &&
                    fullQuantityOnPallet == quantityOnPallet)
                {
                    //this is a full pallet, so we can pick this full pallet
                    targetLuid = alternateLUID;


                    //set correct data
                    pickingItemInfo.IsFullPallet = true;




                    enteredItriKey = Convertor.ConvertToInt32Nullable(dsItemsOnSSCC.Tables[0].Rows[0][PmxStockOverviewOnLocationDefinition.Columns.ItemTransactionalInfoKey.NAME]);
                    enteredBatchNumber = Convert.ToString(dsItemsOnSSCC.Tables[0].Rows[0][PmxStockOverviewOnLocationDefinition.Columns.BatchNumber.NAME]);
                    enteredQualityStatusCode = Convert.ToString(dsItemsOnSSCC.Tables[0].Rows[0][PmxStockOverviewOnLocationDefinition.Columns.QualityStatusCode.NAME]);
                    enteredBestBeforeDate = Convertor.ConvertToDateTimeNullable(dsItemsOnSSCC.Tables[0].Rows[0][PmxStockOverviewOnLocationDefinition.Columns.BestBeforeDate.NAME]);

                    //check if change of batchnumber is allowed
                    if (alternateCopyBatchNumbers &&
                        (
                        enteredBatchNumber != pickingItemInfo.BatchNumber
                      || enteredBestBeforeDate != pickingItemInfo.BestBeforeDate
                      || enteredItriKey != pickingItemInfo.ItemTransactionalInfoKey
                      || enteredQualityStatusCode != pickingItemInfo.QualityStatusCode))
                    {
                        initialErrorKey = "MSG_ERROR_INCORRECT_PICK_BATCH_AND_BEST_BEFORE";
                        goto Step_Alternate_IdentifySSCCOrLocation;
                    }


                    goto Step_SetPickingItemInfo;
                }


            }


            goto Step_CheckLocation;

            #endregion

        Step_SelectLocationAlternate:

            #region SELECT LOCATION ALTERNATE

            query = BuildQuery.GetLocationsAlternates(enteredLocationCode, pmxWarehouseCode, pickingItemInfo.ItemCode);
            DataSet dsLocations = m_sboProviderService.RunView(false, null, null, query);


            if (!string.IsNullOrEmpty(enteredLocationCode) &&
                this.GetNrOfRows(dsLocations) == 1)
            {
                enteredLocationCode = dsLocations.Tables[0].Rows[0][LocationDataSet.Columns.CODE.Name] as string;

                //Check supplier pallet number
                if (!string.IsNullOrEmpty(supplierPalletNumber))
                {
                    query = BuildQuery.GetProductsFromStockOrderByName( null, itemInfo.ItemCode, enteredLocationCode, alternateSSCC, false, null, null, null, null, null, false, Produmex.Sbo.Logex.Data.SboCommands.PackagingUnit.Inventory, false, supplierPalletNumber, m_sboProviderService.GetDbTool() );
                    if (this.GetNrOfRows(m_sboProviderService.RunView(true, null, null, query)) == 0)
                    {
                        //Supplier pallet number does not correspond
                        initialErrorKey = "MSG_ERROR_SSCC_AD_HOC_ALTERNATE_PICKING_SUPPLIER_PALLET_NUMBER_INCORRECT";
                        goto Step_Alternate_IdentifySSCCOrLocation;
                    }
                }


                goto Step_IdentifyBaseComponent;
            }

            session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.ISelectLocationScreen),
                this.DefaultCultureInfo, BuildParamCollection(
                "InitialErrorKey", initialErrorKey,
                "TitleKey", "MSG_TITLE_SELECT_LOCATION",
                "LocationDS", dsLocations));
            msg = WaitForMessage();
            initialErrorKey = null;

            if (msg.Name == "Produmex.Foundation.SlimScreen.Interfaces.ISelectLocationScreenReply.LocationSelected")
            {
                enteredLocationCode = ExtractParameter<string>(msg.Parameters, "locationCode");

                //Check supplier pallet number
                if (!string.IsNullOrEmpty(supplierPalletNumber))
                {
                    query = BuildQuery.GetProductsFromStockOrderByName( null, itemInfo.ItemCode, enteredLocationCode, alternateSSCC, false, null, null, null, null, null, false, Produmex.Sbo.Logex.Data.SboCommands.PackagingUnit.Inventory, false, supplierPalletNumber, m_sboProviderService.GetDbTool() );
                    if (this.GetNrOfRows(m_sboProviderService.RunView(true, null, null, query)) == 0)
                    {
                        //Supplier pallet number does not correspond
                        initialErrorKey = "MSG_ERROR_SSCC_AD_HOC_ALTERNATE_PICKING_SUPPLIER_PALLET_NUMBER_INCORRECT";
                        goto Step_Alternate_IdentifySSCCOrLocation;
                    }
                }

                goto Step_IdentifyBaseComponent;
            }
            else if (msg.Name == "Produmex.Foundation.SlimScreen.Interfaces.ISelectLocationScreenReply.BackRequested")
            {
                goto Step_Alternate_IdentifySSCCOrLocation;
            }
            else if (msg.Name == "Produmex.Foundation.SlimScreen.Interfaces.ISelectLocationScreenReply.RefreshRequested")
            {
                goto Step_SelectLocationAlternate;
            }
            else
            {
                initialErrorKey = KnownMessageKeys.MSG_ERROR_OPERATION_NOT_SUPPORTED;
                goto Step_SelectLocationAlternate;
            }


            #endregion



        }

        #endregion

        #region HELPERS

        private DataSet BuildAlternateList(Collection<PickingItemInfo> pickingItemInfoList)
        {
            DataSet dsAlternateItems = new DataSet();
            DataTable tblAlternateItems = dsAlternateItems.Tables.Add("AlternateItem");
            tblAlternateItems.Columns.Add(AlternateItemDataSet.Columns.INTERNAL_KEY.Name, typeof(int));
            tblAlternateItems.Columns.Add(AlternateItemDataSet.Columns.STORAGE_LOCATION.Name);
            tblAlternateItems.Columns.Add(AlternateItemDataSet.Columns.QUANTITY.Name, typeof(double));
            tblAlternateItems.Columns.Add(AlternateItemDataSet.Columns.ITEMCODE.Name);
            tblAlternateItems.Columns.Add(AlternateItemDataSet.Columns.BESTBEFOREDATE.Name, typeof(DateTime));
            tblAlternateItems.Columns.Add(AlternateItemDataSet.Columns.BATCHNUMBER.Name);
            tblAlternateItems.Columns.Add(AlternateItemDataSet.Columns.UOM.Name);
            tblAlternateItems.Columns.Add(AlternateItemDataSet.Columns.ITEM_DESCRIPTION.Name);
            tblAlternateItems.Columns.Add(AlternateItemDataSet.Columns.ITEM_BARCODE.Name);
            tblAlternateItems.Columns.Add(AlternateItemDataSet.Columns.SSCC.Name);
            tblAlternateItems.Columns.Add(AlternateItemDataSet.Columns.UOM2.Name);
            tblAlternateItems.Columns.Add(AlternateItemDataSet.Columns.QUANTITY_UOM2.Name, typeof(double));
            tblAlternateItems.Columns.Add(AlternateItemDataSet.Columns.BATCHNUMBER2.Name);
            tblAlternateItems.Columns.Add(AlternateItemDataSet.Columns.MASTER_SSCC.Name);

            if (pickingItemInfoList != null && pickingItemInfoList.Count > 0)
            {
                for (int i = 0; i < pickingItemInfoList.Count; i++)
                {
                    DataRow row = tblAlternateItems.Rows.Add(i,
                        pickingItemInfoList[i].LocationCode,
                        pickingItemInfoList[i].Quantity,
                        pickingItemInfoList[i].ItemCode,
                        pickingItemInfoList[i].BestBeforeDate,
                        pickingItemInfoList[i].BatchNumber,
                        pickingItemInfoList[i].Uom,
                        pickingItemInfoList[i].CustomItemDescription,
                        pickingItemInfoList[i].ItemBarcode,
                        pickingItemInfoList[i].IsFullMasterSSCC ? pickingItemInfoList[i].MasterSSCC : pickingItemInfoList[i].IsFullPallet ? pickingItemInfoList[i].SSCC : "",
                        pickingItemInfoList[i].Uom2,
                        pickingItemInfoList[i].QuantityUom2,
                        pickingItemInfoList[i].BatchNumber2,
                        pickingItemInfoList[i].MasterSSCC
                    );
                }
            }

            return dsAlternateItems;
        }

        /// <summary>
        /// Executes the is mono lot pallet.
        /// </summary>
        /// <param name="conn">The conn.</param>
        /// <param name="parameters">The parameters.</param>
        /// <returns></returns>
        private static bool ExecuteIsMonoLotPallet(PmxDbConnection conn, object[] parameters)
        {
            PmxInventoryProvider prov = new PmxInventoryProvider(conn);
            return prov.IsMonoLotPallet((int)parameters[0]);
        }

        /// <summary>
        /// Gets the query for logistic units for customer.
        /// </summary>
        /// <param name="pickingItemInfoList">The picking item info list.</param>
        /// <param name="waveSameDelivery">if set to <c>true</c> [wave same delivery].</param>
        /// <param name="waveKey">The wave key.</param>
        /// <param name="allowMultipleCustomersOnSSCC">if set to <c>true</c> [allow multiple customers on SSCC].</param>
        /// <param name="luid">The luid.</param>
        /// <param name="pmxWarehouseCode">The PMX warehouse code.</param>
        /// <returns></returns>
        private string GetQueryForLogisticUnitsForCustomer(Collection<PickingItemInfo> pickingItemInfoList, bool waveSameDelivery, int waveKey, bool allowMultipleCustomersOnSSCC, int? luid, string pmxWarehouseCode)
        {
            string query = null;
            if (pickingItemInfoList != null && pickingItemInfoList.Count > 0)
            {
                if (!waveSameDelivery && allowMultipleCustomersOnSSCC)
                {
                    //This is a wave with multiple customers, and muliple customers are allowed, so we need to get all SSCC's for all customers/ship to addresses
                    string subQuery = BuildQuery.GetCustomerInfoForWave(waveKey);
                    DataSet ds = m_sboProviderService.RunView(false, null, null, subQuery);
                    if (GetNrOfRows(ds) > 0)
                    {
                        Collection<string> listOfCardCodes = new Collection<string>();
                        Collection<string> listOfShipToAddresses = new Collection<string>();

                        foreach (DataRow row in ds.Tables[0].Rows)
                        {
                            if (!listOfCardCodes.Contains(row[PmxPickListTable.Columns.CardCode.NAME] as string))
                            {
                                listOfCardCodes.Add(row[PmxPickListTable.Columns.CardCode.NAME] as string);
                            }
                            if (!listOfCardCodes.Contains(row[PmxPickListTable.Columns.ShipToAddress.NAME] as string))
                            {
                                listOfShipToAddresses.Add(row[PmxPickListTable.Columns.ShipToAddress.NAME] as string);
                            }
                        }

                        query = BuildQuery.GetLogisticUnitsForCustomer(listOfCardCodes, listOfShipToAddresses, pickingItemInfoList[0].RouteKey, luid, pmxWarehouseCode, null, false);


                    }
                    else
                    {
                        //this should not happen!
                        query = BuildQuery.GetLogisticUnitsForCustomer((string)null, (string)null, null, null, pmxWarehouseCode, null, false);
                    }
                }
                else
                {
                    query = BuildQuery.GetLogisticUnitsForCustomer(pickingItemInfoList[0].CustomerCode, pickingItemInfoList[0].ShipToAddress, pickingItemInfoList[0].RouteKey, luid, pmxWarehouseCode, pickingItemInfoList[0].MoveToWarehouse, true);
                }
            }
            else
            {
                query = BuildQuery.GetLogisticUnitsForCustomer((string)null, (string)null, null, null, pmxWarehouseCode, null, false);
            }

            return query;
        }

        private SortedList<int, MultiPickingPickListInfo> GetMultiPickingPickListInfo(string pickListBarcode, bool isProductionPicking, string pmxWarehouseCode, out DataSet dsPickLists, out bool pickingNeeded, PmxDbConnection dbConn)
        {
            return GetMultiPickingPickListInfo(pickListBarcode, isProductionPicking, new List<string> { pmxWarehouseCode }, out dsPickLists, out pickingNeeded, dbConn);
        }

        private SortedList<int, MultiPickingPickListInfo> GetMultiPickingPickListInfo(string pickListBarcode, bool isProductionPicking, List<string> pmxWarehouseCodes, out DataSet dsPickLists, out bool pickingNeeded, PmxDbConnection dbConn)
        {
            string query = BuildQuery.GetPickListMultiPickingInfo(pickListBarcode, isProductionPicking, dbConn.DbTool);
            using (ISboRecordset rs = SboRecordsetHelper.RunQuery(s_log, query, dbConn))
            {
                dsPickLists = SboRecordsetHelper.ConvertRecordsetToDataSet(s_log, rs);
            }

            pickingNeeded = false;
            SortedList<int, MultiPickingPickListInfo> multiPickingPickListInfos = new SortedList<int, MultiPickingPickListInfo>();
            foreach (DataRow row in dsPickLists.Tables[0].Rows)
            {
                if (!pickingNeeded)
                {
                    switch (Convert.ToString(row[PmxPickListTable.Columns.PickListStatus.NAME])[0])
                    {
                        case PmxPickListStatusColumnValues.NotReady.VALUE:
                        case PmxPickListStatusColumnValues.Ready.VALUE:
                        case PmxPickListStatusColumnValues.PartiallyReady.VALUE:
                        case PmxPickListStatusColumnValues.PartiallyPicked.VALUE:
                            pickingNeeded = pmxWarehouseCodes.Contains(Convert.ToString(row[PmxOseGeneralTable.Columns.PmxWarehouseCode.NAME]));
                            break;
                    }
                }
                int docEntry = Convert.ToInt32(row[PmxPickListTable.Columns.DocEntry.NAME]);
                MultiPickingPickListInfo info = new MultiPickingPickListInfo();
                info.DocEntry = docEntry;
                info.MovableLocation = Convert.ToString(row[PmxPickListTable.Columns.MultiPickingMovLoc.NAME]);
                multiPickingPickListInfos.Add(docEntry, info);
            }

            return multiPickingPickListInfos;
        }


        private string UpdateMultiPickingMovLoc(SortedList<int, MultiPickingPickListInfo> multiPickingPickListInfos, PmxDbConnection dbConn)
        {
            string query = BuildQuery.UpdateMultiPickingMovLoc(multiPickingPickListInfos, false, dbConn);
            if ( dbConn.DbTool.DbServerType == DbServerType.HANA )
            {
                // For HANA, this query only contains the UPDATE statement. Another query is used to select the number of affected rows.
                SboRecordsetHelper.RunQueryNoReturn( s_log, query, dbConn );
                query = BuildQuery.UpdateMultiPickingMovLoc( multiPickingPickListInfos, true, dbConn );
            }
            using (ISboRecordset rs = SboRecordsetHelper.RunQuery(s_log, query, dbConn))
            {
                if (rs.EoF || rs.GetTypedValue<int>(0) != multiPickingPickListInfos.Count)
                {
                    return "MSG_ERROR_PICKLISTS_CHANGED";
                }
            }

            return null;
        }

        private int GetNrOfRows(DataSet ds)
        {
            return (ds != null && ds.Tables.Count > 0) ? ds.Tables[0].Rows.Count : 0;
        }

        private T ExecuteMethodWithDbConnection<T>(out ProdumexError pmxError, bool useSboConnection, bool useUserNameOfConnectionString, DelegateWithPmxDbConnection<T> methodDelegate, params object[] parameters)
        {
            T result = default(T);
            pmxError = null;

            try
            {
                result = m_sboProviderService.InvokeMethodWithDbConnection<T>(useSboConnection, useUserNameOfConnectionString, null, null,
                    methodDelegate, parameters);
            }
            catch (TargetInvocationException ex)
            {
                MethodInfo methodInfo = methodDelegate.Method;
                if (ex.InnerException is ProdumexException)
                {
                    s_log.Warn("Execute -- The invoked method threw a produmex exception: ", ex.InnerException);
                    pmxError = ((ProdumexException)ex.InnerException).Error;
                }
                else if (ex.InnerException == null)
                {
                    s_log.Warn("Execute -- The invoked method threw an exception: ", ex);
                    pmxError = Produmex.Foundation.Wwf.Sbo.ProdumexErrors.ExceptionInvokingSboMethod(s_log, methodInfo, ex.Message);
                }
                else
                {
                    s_log.Warn("Execute -- The invoked method threw a TargetInvocationException with inner exception: ", ex.InnerException);
                    pmxError = Produmex.Foundation.Wwf.Sbo.ProdumexErrors.ExceptionInvokingSboMethod(s_log, methodInfo, ex.InnerException.Message);
                }
            }
            catch (TransactionAbortedException ex)
            {
                s_log.Error("", ex);
                MethodInfo methodInfo = methodDelegate.Method;
                pmxError = Produmex.Foundation.Wwf.Sbo.ProdumexErrors.ExceptionInvokingSboMethod(s_log, methodInfo, ex.InnerException == null ? ex.Message : ex.InnerException.Message);
            }
            catch (ProdumexException ex)
            {
                s_log.Error("", ex);
                if (ex.Error != null)
                {
                    pmxError = ex.Error;
                }
                else
                {
                    MethodInfo methodInfo = methodDelegate.Method;
                    pmxError = Produmex.Foundation.Wwf.Sbo.ProdumexErrors.ExceptionInvokingSboMethod(s_log, methodInfo, ex.Message);
                }
            }
            catch (Exception ex)
            {
                s_log.Error("", ex);
                MethodInfo methodInfo = methodDelegate.Method;
                pmxError = Produmex.Foundation.Wwf.Sbo.ProdumexErrors.ExceptionInvokingSboMethod(s_log, methodInfo, ex.Message);
            }

            return result;
        }

        private string ExecuteGetPmxWarehouseCode(PmxDbConnection conn, object[] parameters)
        {
            PmxOseWarehouseProvider prov = new PmxOseWarehouseProvider(conn);
            return prov.GetWarehouseOfElement((string)parameters[0]);
        }

        private bool ExecuteIsPickListForAutomaticShipping(PmxDbConnection conn, object[] parameters)
        {
            PmxPickListProvider prov = new PmxPickListProvider(conn);
            return prov.IsPickListForAutomaticShipping((int)parameters[0]);
        }

        private bool ExecuteIsPickListForAutomaticinvoicing(PmxDbConnection conn, object[] parameters)
        {
            PmxPickListProvider prov = new PmxPickListProvider(conn);
            return prov.IsPickListForAutomaticInvoicing((int)parameters[0]);
        }

        #endregion

        private static class BuildQuery
        {

			/// <summary>
			/// Gets the pick lists for wave.
			/// </summary>
			/// <param name="waveKey">The wave key.</param>
			/// <param name="culture">The culture.</param>
			public static string GetPickListsForWave(int waveKey, CultureInfo culture)
			{
				return Produmex.Sbo.Logex.Data.SboCommands.GetQueryGetPickListsForWave(waveKey, culture);
			}

            /// <summary>
            /// Gets the name of the products from stock order by.
            /// </summary>
            /// <param name="gtin">The gtin.</param>
            /// <param name="itemCode">The item code.</param>
            /// <param name="locationCode">The location code.</param>
            /// <param name="sscc">The SSCC.</param>
            /// <param name="forceSpecificSSCC">if set to <c>true</c> [force specific SSCC].</param>
            /// <param name="luid">The luid.</param>
            /// <param name="isLogisticCarrier">The is logistic carrier.</param>
            /// <param name="batchNumber">The batch number.</param>
            /// <param name="secondBatchNumber">The second batch number.</param>
            /// <param name="bestBeforeDate">The best before date.</param>
            /// <param name="forceSpecificITRI">if set to <c>true</c> [force specific ITRI].</param>
            /// <param name="packagingUnit">The packaging unit.</param>
            /// <param name="dbTool">The db tool.</param>
            /// <returns></returns>
            public static string GetProductsFromStockOrderByName(string gtin, string itemCode, string locationCode, string sscc, bool forceSpecificSSCC, int? luid, bool? isLogisticCarrier, string batchNumber, string secondBatchNumber, DateTime? bestBeforeDate, bool forceSpecificITRI, Produmex.Sbo.Logex.Data.SboCommands.PackagingUnit packagingUnit, DbTool dbTool)
            {
                return GetProductsFromStockOrderByName(gtin, itemCode, locationCode, sscc, forceSpecificSSCC, luid, isLogisticCarrier, batchNumber, secondBatchNumber, bestBeforeDate, forceSpecificITRI, packagingUnit, false, null, dbTool);
            }


            /// <summary>
            /// Gets the name of the products from stock order by.
            /// </summary>
            /// <param name="gtin">The gtin.</param>
            /// <param name="itemCode">The item code.</param>
            /// <param name="locationCode">The location code.</param>
            /// <param name="sscc">The SSCC.</param>
            /// <param name="forceSpecificSSCC">if set to <c>true</c> [force specific SSCC].</param>
            /// <param name="luid">The luid.</param>
            /// <param name="isLogisticCarrier">The is logistic carrier.</param>
            /// <param name="batchNumber">The batch number.</param>
            /// <param name="secondBatchNumber">The second batch number.</param>
            /// <param name="bestBeforeDate">The best before date.</param>
            /// <param name="forceSpecificITRI">if set to <c>true</c> [force specific ITRI].</param>
            /// <param name="packagingUnit">The packaging unit.</param>
            /// <param name="onlyPickableQualityStatuses">if set to <c>true</c> [only pickable quality statuses].</param>
            /// <param name="supplierPalletNumber">The supplier pallet number.</param>
            /// <param name="dbTool">The db tool.</param>
            /// <returns></returns>
            public static string GetProductsFromStockOrderByName(string gtin, string itemCode, string locationCode, string sscc, bool forceSpecificSSCC, int? luid, bool? isLogisticCarrier, string batchNumber, string secondBatchNumber, DateTime? bestBeforeDate, bool forceSpecificITRI, Produmex.Sbo.Logex.Data.SboCommands.PackagingUnit packagingUnit, bool onlyPickableQualityStatuses, string supplierPalletNumber, DbTool dbTool)
            {
                DbQueryBuilder strBuilder = dbTool.GetQueryBuilder();
                strBuilder.Append("SELECT ");

                SqlCommandHelper.AddTableColumnNameWithAs( strBuilder, SboItemTable.NAME, SboItemTable.Columns.ItemCode.NAME, PmxStockOverviewOnLocationDefinition.Columns.ProductCode.NAME );

                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnNameWithAs( strBuilder, SboItemTable.NAME, SboItemTable.Columns.ItemName.NAME, PmxStockOverviewOnLocationDefinition.Columns.ProductDescription.NAME );

                strBuilder.Append(", ");
                Produmex.Sbo.Logex.Data.SboCommands.AddBarcodeUsingPackagingUnit(strBuilder, packagingUnit, PmxStockOverviewOnLocationDefinition.Columns.GTIN.NAME);

                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnNameWithAs( strBuilder, SboItemTable.NAME, SboItemTable.Columns.InventoryUom.NAME, PmxStockOverviewOnLocationDefinition.Columns.InventoryUom.NAME );

                strBuilder.Append(", ");
                strBuilder.Append(SboRecordsetHelper.GetSelectPartsColumnAndIsColumnNull(PmxItemTransactionalInfoTable.NAME, PmxItemTransactionalInfoTable.Columns.BatchNumber1.NAME, PmxStockOverviewOnLocationDefinition.Columns.BatchNumber.NAME));
                strBuilder.Append(", ");
                strBuilder.Append(SboRecordsetHelper.GetSelectPartsColumnAndIsColumnNull(PmxItemTransactionalInfoTable.NAME, PmxItemTransactionalInfoTable.Columns.BatchNumber2.NAME, PmxStockOverviewOnLocationDefinition.Columns.BatchNumber2.NAME));
                strBuilder.Append(", ");
                strBuilder.Append(SboRecordsetHelper.GetSelectPartsColumnAndIsColumnNull(PmxItemTransactionalInfoTable.NAME, PmxItemTransactionalInfoTable.Columns.BestBeforeDate.NAME, PmxStockOverviewOnLocationDefinition.Columns.BestBeforeDate.NAME));
                strBuilder.Append(", ");
                strBuilder.Append(SboRecordsetHelper.GetSelectPartsColumnAndIsColumnNull(PmxInventoryTotalTable.NAME, PmxInventoryTotalTable.Columns.QualityStatusCode.NAME, PmxStockOverviewOnLocationDefinition.Columns.QualityStatusCode.NAME));
                strBuilder.Append(", ");
                strBuilder.Append(SboRecordsetHelper.GetSelectPartsColumnAndIsColumnNull(PmxInventoryTotalTable.NAME, PmxInventoryTotalTable.Columns.SSCC.NAME, PmxStockOverviewOnLocationDefinition.Columns.SSCC.NAME));
                strBuilder.Append(", ");
                strBuilder.Append(SboRecordsetHelper.GetSelectPartsColumnAndIsColumnNull(PmxInventoryTotalTable.NAME, PmxInventoryTotalTable.Columns.LogUnitIdentKey.NAME, PmxStockOverviewOnLocationDefinition.Columns.LogUnitIdentKey.NAME));
                strBuilder.Append(", ");
                strBuilder.Append(SboRecordsetHelper.GetSelectPartsColumnAndIsColumnNull(PmxInventoryTotalTable.NAME, PmxInventoryTotalTable.Columns.InternalKey.NAME, PmxStockOverviewOnLocationDefinition.Columns.InternalKey.NAME));
                strBuilder.Append(", ");
                strBuilder.Append(SboRecordsetHelper.GetSelectPartsColumnAndIsColumnNull(PmxInventoryTotalTable.NAME, PmxInventoryTotalTable.Columns.StorLocCode.NAME, PmxStockOverviewOnLocationDefinition.Columns.DefaultLocationCode.NAME));
                strBuilder.Append(", ");
                strBuilder.Append(SboRecordsetHelper.GetSelectPartsColumnAndIsColumnNull(PmxInventoryTotalTable.NAME, PmxInventoryTotalTable.Columns.ItemTransactionalInfoKey.NAME, PmxStockOverviewOnLocationDefinition.Columns.ItemTransactionalInfoKey.NAME));
                strBuilder.Append(", SUM(",strBuilder.AsIdentifier(PmxInventoryTotalTable.NAME, PmxInventoryTotalTable.Columns.Quantity.NAME));
                strBuilder.Append(") AS ");
                strBuilder.Append(strBuilder.AsIdentifier(PmxStockOverviewOnLocationDefinition.Columns.Quantity.NAME));
                strBuilder.Append(", SUM(",strBuilder.AsIdentifier(PmxInventoryTotalTable.NAME, PmxInventoryTotalTable.Columns.QuantityUom2.NAME));
                strBuilder.Append(") AS ");
                strBuilder.Append(strBuilder.AsIdentifier(PmxStockOverviewOnLocationDefinition.Columns.QuantityUom2.NAME));

                strBuilder.Append(" FROM ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(SboItemTable.NAME));

                //inner join Product - inventory
                strBuilder.Append(" INNER JOIN ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxInventoryTotalTable.NAME));
                strBuilder.Append(" ON ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxInventoryTotalTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxInventoryTotalTable.Columns.ItemCode.NAME));
                strBuilder.Append(" = ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(SboItemTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(SboItemTable.Columns.ItemCode.NAME));

                if (onlyPickableQualityStatuses)
                {
                    //inner join inventory - quality status
                    strBuilder.Append(" INNER JOIN ");
                    strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxQualityStatusTable.NAME));
                    strBuilder.Append(" ON ");
                    strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxInventoryTotalTable.NAME));
                    strBuilder.Append(".");
                    strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxInventoryTotalTable.Columns.QualityStatusCode.NAME));
                    strBuilder.Append(" = ");
                    strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxQualityStatusTable.NAME));
                    strBuilder.Append(".");
                    strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxQualityStatusTable.Columns.Code.NAME));

                    strBuilder.Append(" AND ");
                    strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxQualityStatusTable.NAME));
                    strBuilder.Append(".");
                    strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxQualityStatusTable.Columns.CanBeShipped.NAME));
                    strBuilder.Append(" = ");
                    strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(SboBooleanColumnValues.Y.VALUE, CultureInfo.InvariantCulture));

                }


                //left join for transaction info
                strBuilder.Append(" LEFT JOIN ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxItemTransactionalInfoTable.NAME));
                strBuilder.Append(" ON ",strBuilder.AsIdentifier(PmxInventoryTotalTable.NAME, PmxInventoryTotalTable.Columns.ItemTransactionalInfoKey.NAME));
                strBuilder.Append(" = ",strBuilder.AsIdentifier(PmxItemTransactionalInfoTable.NAME, PmxItemTransactionalInfoTable.Columns.InternalKey.NAME));


                //left join for LUID
                strBuilder.Append(" LEFT JOIN ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxLogisticUnitIDTable.NAME));
                strBuilder.Append(" ON ",strBuilder.AsIdentifier(PmxInventoryTotalTable.NAME, PmxInventoryTotalTable.Columns.LogUnitIdentKey.NAME));
                strBuilder.Append(" = ",strBuilder.AsIdentifier(PmxLogisticUnitIDTable.NAME, PmxLogisticUnitIDTable.Columns.InternalKey.NAME));


                string operand = " WHERE ";
                if (!string.IsNullOrEmpty(gtin))
                {
                    strBuilder.Append(operand);
                    Produmex.Sbo.Logex.Data.SboCommands.AddWhereClauseSearchForBarcode(strBuilder, gtin, false);
                    operand = " AND ";
                }

                if (!string.IsNullOrEmpty(itemCode))
                {
                    strBuilder.Append(operand);
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(SboItemTable.NAME));
                    strBuilder.Append(".");
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(SboItemTable.Columns.ItemCode.NAME));
                    strBuilder.Append(" = ");
                    strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(itemCode, CultureInfo.InvariantCulture));
                    operand = " AND ";
                }


                if (!string.IsNullOrEmpty(supplierPalletNumber))
                {
                    strBuilder.Append(operand);
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxLogisticUnitIDTable.NAME));
                    strBuilder.Append(".");
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxLogisticUnitIDTable.Columns.SupplierLogisticUnitID.NAME));
                    strBuilder.Append(" = ");
                    strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(supplierPalletNumber, CultureInfo.InvariantCulture));
                    operand = " AND ";
                }

                if (!string.IsNullOrEmpty(sscc))
                {
                    strBuilder.Append(operand);
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxInventoryTotalTable.NAME));
                    strBuilder.Append(".");
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxInventoryTotalTable.Columns.SSCC.NAME));
                    strBuilder.Append(" = ");
                    strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(sscc, CultureInfo.InvariantCulture));
                    operand = " AND ";
                }
                else if (forceSpecificSSCC)
                {
                    strBuilder.Append(operand);
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxInventoryTotalTable.NAME));
                    strBuilder.Append(".");
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxInventoryTotalTable.Columns.SSCC.NAME));
                    strBuilder.Append(" IS NULL ");
                    operand = " AND ";

                }

                if (!string.IsNullOrEmpty(batchNumber))
                {
                    strBuilder.Append(operand);
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxItemTransactionalInfoTable.NAME));
                    strBuilder.Append(".");
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxItemTransactionalInfoTable.Columns.BatchNumber1.NAME));
                    strBuilder.Append(" = ");
                    strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(batchNumber, CultureInfo.InvariantCulture));
                    operand = " AND ";
                }
                else if (forceSpecificITRI)
                {
                    strBuilder.Append(operand);
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxItemTransactionalInfoTable.NAME));
                    strBuilder.Append(".");
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxItemTransactionalInfoTable.Columns.BatchNumber1.NAME));
                    strBuilder.Append(" IS NULL ");
                    operand = " AND ";

                }
                if (!string.IsNullOrEmpty(secondBatchNumber))
                {
                    strBuilder.Append(operand);
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxItemTransactionalInfoTable.NAME));
                    strBuilder.Append(".");
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxItemTransactionalInfoTable.Columns.BatchNumber2.NAME));
                    strBuilder.Append(" = ");
                    strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(secondBatchNumber, CultureInfo.InvariantCulture));
                    operand = " AND ";
                }
                else if (forceSpecificITRI)
                {
                    strBuilder.Append(operand);
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxItemTransactionalInfoTable.NAME));
                    strBuilder.Append(".");
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxItemTransactionalInfoTable.Columns.BatchNumber2.NAME));
                    strBuilder.Append(" IS NULL ");
                    operand = " AND ";

                }
                if (bestBeforeDate.HasValue)
                {
                    strBuilder.Append(operand);
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxItemTransactionalInfoTable.NAME));
                    strBuilder.Append(".");
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxItemTransactionalInfoTable.Columns.BestBeforeDate.NAME));
                    strBuilder.Append(" = ");
                    strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(bestBeforeDate.Value, CultureInfo.InvariantCulture));
                    operand = " AND ";
                }
                else if (forceSpecificITRI)
                {
                    strBuilder.Append(operand);
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxItemTransactionalInfoTable.NAME));
                    strBuilder.Append(".");
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxItemTransactionalInfoTable.Columns.BestBeforeDate.NAME));
                    strBuilder.Append(" IS NULL ");
                    operand = " AND ";

                }

                if (!string.IsNullOrEmpty(locationCode))
                {
                    strBuilder.Append(operand);
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxInventoryTotalTable.NAME));
                    strBuilder.Append(".");
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxInventoryTotalTable.Columns.StorLocCode.NAME));
                    strBuilder.Append(" = ");
                    strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(locationCode, CultureInfo.InvariantCulture));
                    operand = " AND ";
                }

                if (luid.HasValue)
                {
                    strBuilder.AppendFormat(" {0} {1}.{2}  = {3} ", operand, SqlCommandHelper.EscapeTableName(PmxInventoryTotalTable.NAME), SqlCommandHelper.EscapeColumnName(PmxInventoryTotalTable.Columns.LogUnitIdentKey.NAME), SqlCommandHelper.FormatAndEscapeColumnValue(luid.Value, CultureInfo.InvariantCulture));
                    operand = " AND ";
                }

                if (isLogisticCarrier.HasValue)
                {
                    strBuilder.Append(operand);
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(SboItemTable.NAME));
                    strBuilder.Append(".");
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxItemTable.Columns.IsLogisticCarrier.NAME));
                    strBuilder.Append(" = ");
                    if (isLogisticCarrier.Value)
                    {
                        strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(SboBooleanColumnValues.Y.VALUE, CultureInfo.InvariantCulture));
                    }
                    else
                    {
                        strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(SboBooleanColumnValues.N.VALUE, CultureInfo.InvariantCulture));
                    }
                    operand = " AND ";
                }

                strBuilder.Append(" GROUP BY ");

                strBuilder.Append(SqlCommandHelper.EscapeColumnName(SboItemTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(SboItemTable.Columns.ItemCode.NAME));

                strBuilder.Append(", ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(SboItemTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(SboItemTable.Columns.ItemName.NAME));

                strBuilder.Append(", ",strBuilder.AsIdentifier(SboItemTable.NAME, SboItemTable.Columns.NumberOfItemsPerPurchaseUnit.NAME));
                strBuilder.Append(", ",strBuilder.AsIdentifier(SboItemTable.NAME, SboItemTable.Columns.CodeBars.NAME));
                strBuilder.Append(", ",strBuilder.AsIdentifier(PmxItemTable.NAME, PmxItemTable.Columns.PurchaseBarcode.NAME));
                strBuilder.Append(", ",strBuilder.AsIdentifier(SboItemTable.NAME, SboItemTable.Columns.NumberOfItemsPerSalesUnit.NAME));
                strBuilder.Append(", ",strBuilder.AsIdentifier(PmxItemTable.NAME, PmxItemTable.Columns.SalesBarcode.NAME));

                strBuilder.Append(", ",strBuilder.AsIdentifier(SboItemTable.NAME, SboItemTable.Columns.InventoryUom.NAME));

                strBuilder.Append(", ",strBuilder.AsIdentifier(PmxItemTransactionalInfoTable.NAME, PmxItemTransactionalInfoTable.Columns.BatchNumber1.NAME));
                strBuilder.Append(", ",strBuilder.AsIdentifier(PmxItemTransactionalInfoTable.NAME, PmxItemTransactionalInfoTable.Columns.BatchNumber2.NAME));
                strBuilder.Append(", ",strBuilder.AsIdentifier(PmxItemTransactionalInfoTable.NAME, PmxItemTransactionalInfoTable.Columns.BestBeforeDate.NAME));
                strBuilder.Append(", ",strBuilder.AsIdentifier(PmxInventoryTotalTable.NAME, PmxInventoryTotalTable.Columns.QualityStatusCode.NAME));
                strBuilder.Append(", ",strBuilder.AsIdentifier(PmxInventoryTotalTable.NAME, PmxInventoryTotalTable.Columns.SSCC.NAME));
                strBuilder.Append(", ",strBuilder.AsIdentifier(PmxInventoryTotalTable.NAME, PmxInventoryTotalTable.Columns.LogUnitIdentKey.NAME));
                strBuilder.Append(", ",strBuilder.AsIdentifier(PmxInventoryTotalTable.NAME, PmxInventoryTotalTable.Columns.InternalKey.NAME));
                strBuilder.Append(", ",strBuilder.AsIdentifier(PmxInventoryTotalTable.NAME, PmxInventoryTotalTable.Columns.StorLocCode.NAME));
                strBuilder.Append(", ",strBuilder.AsIdentifier(PmxInventoryTotalTable.NAME, PmxInventoryTotalTable.Columns.ItemTransactionalInfoKey.NAME));


                strBuilder.Append(" ORDER BY ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(SboItemTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(SboItemTable.Columns.ItemName.NAME));
                return strBuilder.ToString();
            }



            /// <summary>
            /// Command for getting all products order by name
            /// </summary>
            /// <param name="itemCode">The item code.</param>
            /// <param name="gtin">The gtin.</param>
            /// <param name="packagingUnit">The packaging unit.</param>
            /// <param name="filterOutFrozenItems">if set to <c>true</c> [filter out frozen items].</param>
            /// <returns>
            /// Command for getting all products order by name
            /// </returns>
            public static string GetCodeNameItemsOrderByName(string itemCode, string gtin, Produmex.Sbo.Logex.Data.SboCommands.PackagingUnit packagingUnit, bool filterOutFrozenItems, DbTool dbTool)
            {
                DbQueryBuilder strBuilder = dbTool.GetQueryBuilder();
                strBuilder.Append("SELECT ");

                SqlCommandHelper.AddTableColumnNameWithAs( strBuilder, SboItemTable.NAME, SboItemTable.Columns.ItemCode.NAME, ProductDataSet.Columns.PRODUCT_CODE.Name );

                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnNameWithAs( strBuilder, SboItemTable.NAME, PmxItemTable.Columns.CustomItemDescr.NAME, ProductDataSet.Columns.PRODUCT_DESCRIPTION.Name );

                strBuilder.Append(", ");
                Produmex.Sbo.Logex.Data.SboCommands.AddBarcodeUsingPackagingUnit(strBuilder, packagingUnit, ProductDataSet.Columns.GTIN.Name);

                strBuilder.Append(", ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(SboItemTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxItemTable.Columns.HasBestBeforeDate.NAME));

                strBuilder.Append(", ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxItemTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxItemTable.Columns.PrintItemLabelReception.NAME));


                strBuilder.Append(", ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(SboItemTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(SboItemTable.Columns.ManBtchNum.NAME));

                strBuilder.Append(", ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(SboItemTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxItemTable.Columns.IsLogisticUnit.NAME));

                strBuilder.Append(", ",strBuilder.AsIdentifier(SboItemTable.NAME, SboItemTable.Columns.NumberOfItemsPerPurchaseUnit.NAME));

                strBuilder.Append(", ",strBuilder.AsIdentifier(SboItemTable.NAME, SboItemTable.Columns.PurchaseUom.NAME));

                strBuilder.Append(", ",strBuilder.AsIdentifier(SboItemTable.NAME, SboItemTable.Columns.InventoryUom.NAME));

                strBuilder.Append(", ");
                Produmex.Sbo.Logex.Data.SboCommands.AddBarcodeUsingPackagingUnit(strBuilder, packagingUnit, SboItemTable.Columns.CodeBars.NAME);
                //SqlCommandHelper.AddTableColumnName(strBuilder, PmxItemTable.NAME, PmxItemTable.Columns.PurchaseBarcode.NAME);

               

                strBuilder.Append(" FROM ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(SboItemTable.NAME));

              
                string operand = " WHERE ";
                
                if (!string.IsNullOrEmpty(gtin))
                {
                    strBuilder.Append(operand);
                    Produmex.Sbo.Logex.Data.SboCommands.AddWhereClauseSearchForBarcode(strBuilder, gtin, false);
                    operand = " AND ";
                }

                if (!string.IsNullOrEmpty(itemCode))
                {
                    strBuilder.Append(operand);
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(SboItemTable.NAME));
                    strBuilder.Append(".");
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(SboItemTable.Columns.ItemCode.NAME));
                    strBuilder.Append(" = ");
                    strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(itemCode, CultureInfo.InvariantCulture));
                    operand = " AND ";
                }

              
                strBuilder.Append(operand);
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(SboItemTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(SboItemTable.Columns.InventoryItem.NAME));
                strBuilder.Append(" = ");
                strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(SboBooleanColumnValues.Y.VALUE, CultureInfo.InvariantCulture));
                operand = " AND ";

                switch (packagingUnit)
                {
                    case Produmex.Sbo.Logex.Data.SboCommands.PackagingUnit.Purchasing:
                        strBuilder.Append(operand,strBuilder.AsIdentifier(SboItemTable.NAME, SboItemTable.Columns.PurchaseItem.NAME));
                        strBuilder.Append(" = ");
                        strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(SboBooleanColumnValues.Y.VALUE, CultureInfo.InvariantCulture));
                        operand = " AND ";
                        break;
                    case Produmex.Sbo.Logex.Data.SboCommands.PackagingUnit.Sales:
                        strBuilder.Append(operand,strBuilder.AsIdentifier(SboItemTable.NAME, SboItemTable.Columns.SalesItem.NAME));
                        strBuilder.Append(" = ");
                        strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(SboBooleanColumnValues.Y.VALUE, CultureInfo.InvariantCulture));
                        operand = " AND ";
                        break;
                    default:
                        strBuilder.Append(operand,strBuilder.AsIdentifier(SboItemTable.NAME, SboItemTable.Columns.InventoryItem.NAME));
                        strBuilder.Append(" = ");
                        strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(SboBooleanColumnValues.Y.VALUE, CultureInfo.InvariantCulture));
                        operand = " AND ";
                        break;
                }

                if (filterOutFrozenItems)
                {
                    Produmex.Sbo.Logex.Data.SboCommands.GetFrozenItemFilterSQLWherePart(strBuilder, " AND ");
                }

                strBuilder.Append(" ORDER BY ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(SboItemTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxItemTable.Columns.CustomItemDescr.NAME));
                return strBuilder.ToString();
            }



            /// <summary>
            /// Gets all recursive zones for wave.
            /// </summary>
            /// <param name="waveKey">The wave key.</param>
            /// <param name="pmxWhsCode">The PMX WHS code.</param>
            /// <returns></returns>
            public static string GetAllRecursiveZonesForWaveAndNoLocationsSet(int waveKey, string pmxWhsCode)
            {
                /*WITH Locations ( [ParentCode], [Code], [Name], [Level] ) AS (
    SELECT [ParentCode], [Code], [Name], 0 AS Level
    FROM PMX_OSEL
    WHERE [Code] in 
(
            SELECT DISTINCT [PMX_OSEL]."PmxZoneCode" AS "Code" FROM [PMX_OSSL] 
                 INNER JOIN [PMX_OSEL] ON [PMX_OSEL]."Code" = [PMX_OSSL]."Code" 
                 INNER JOIN [PMX_OSWH] ON [PMX_OSEL]."PmxWhsCode" = [PMX_OSWH]."Code" 
                 INNER JOIN [PMX_OSZO] ON [PMX_OSEL]."PmxZoneCode" = [PMX_OSZO]."Code" 
                 WHERE "PMX_OSSL"."IsActive" = N'Y' 
)
    UNION ALL
    SELECT child.ParentCode, child.Code, child.Name, [Level] + 1
    FROM Locations parent
    INNER JOIN PMX_OSEL child ON parent.Parentcode = child.Code
)


SELECT Distinct Locations.code
,[PMX_OSZO]."Name" AS "Description", [PMX_OSEL]."PmxWhsCode" AS "WarehouseCode", [PMX_OSWH]."Name" AS "WarehouseName" 
FROM Locations
INNER JOIN PMX_OSZO ON PMX_OSZO.Code = Locations.Code
INNER JOIN PMX_OSEL ON PMX_OSEL.Code = Locations.Code
INNER JOIN [PMX_OSWH] ON [PMX_OSEL]."PmxWhsCode" = [PMX_OSWH]."Code"
ORDER BY PMX_OSZO.Name*/

                StringBuilder strBuilder = new StringBuilder();

                strBuilder.Append(Produmex.Sbo.Logex.Data.SboCommands.GetRecursiveLocationsLowToHighLevel(null, GetAllZones(waveKey), CultureInfo.InvariantCulture));

                strBuilder.Append("SELECT DISTINCT ");

                SqlCommandHelper.AddTableColumnNameWithAs(strBuilder, PmxRecursiveLocationViewDefinition.NAME, PmxRecursiveLocationViewDefinition.Columns.Code.NAME, LocationDataSet.Columns.CODE.Name);

                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnNameWithAs(strBuilder, PmxOseZoneTable.NAME, PmxOseZoneTable.Columns.Name.NAME, LocationDataSet.Columns.DESCRIPTION.Name);

                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnNameWithAs(strBuilder, PmxOseGeneralTable.NAME, PmxOseGeneralTable.Columns.PmxWarehouseCode.NAME, LocationDataSet.Columns.WAREHOUSECODE.Name);

                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnNameWithAs(strBuilder, PmxOseWarehouseTable.NAME, PmxOseWarehouseTable.Columns.Name.NAME, LocationDataSet.Columns.WAREHOUSENAME.Name);

                strBuilder.Append(" FROM ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxRecursiveLocationViewDefinition.NAME));
                strBuilder.Append(" INNER JOIN ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxOseGeneralTable.NAME));
                strBuilder.Append(" ON ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseGeneralTable.NAME, PmxOseGeneralTable.Columns.Code.NAME);
                strBuilder.Append(" = ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxRecursiveLocationViewDefinition.NAME, PmxRecursiveLocationViewDefinition.Columns.Code.NAME);

                strBuilder.Append(" INNER JOIN ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxOseWarehouseTable.NAME));
                strBuilder.Append(" ON ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseGeneralTable.NAME, PmxOseGeneralTable.Columns.PmxWarehouseCode.NAME);
                strBuilder.Append(" = ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseWarehouseTable.NAME, PmxOseWarehouseTable.Columns.Code.NAME);


                strBuilder.Append(" INNER JOIN ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxOseZoneTable.NAME));
                strBuilder.Append(" ON ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseGeneralTable.NAME, PmxOseGeneralTable.Columns.Code.NAME);
                strBuilder.Append(" = ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseZoneTable.NAME, PmxOseZoneTable.Columns.Code.NAME);

                strBuilder.Append(" WHERE ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseZoneTable.NAME, PmxOseZoneTable.Columns.UseForZonePicking.NAME);
                strBuilder.Append(" = ");
                strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(SboBooleanColumnValues.Y.VALUE, CultureInfo.InvariantCulture));

                if (!string.IsNullOrEmpty(pmxWhsCode))
                {
                    strBuilder.Append(" AND ");
                    SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseGeneralTable.NAME, PmxOseGeneralTable.Columns.PmxWarehouseCode.NAME);
                    strBuilder.Append(" = ");
                    strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(pmxWhsCode, CultureInfo.InvariantCulture));
                }


                strBuilder.Append(" ORDER BY ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxOseZoneTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxOseZoneTable.Columns.Name.NAME));


                return strBuilder.ToString();

            }


            /// <summary>
            /// Gets all recursive zones for wave.
            /// </summary>
            /// <param name="waveKey">The wave key.</param>
            /// <param name="userCode">The user code.</param>
            /// <param name="pmxWhsCode">The PMX WHS code.</param>
            /// <returns></returns>
            public static string GetAllRecursiveZonesForWave( int waveKey, string userCode, string pmxWhsCode, out string preQueryForHana, DbTool dbTool )
            {
                /*WITH Locations ( [ParentCode], [Code], [Name], [Level] ) AS (
    SELECT [ParentCode], [Code], [Name], 0 AS Level
    FROM PMX_OSEL
    WHERE [Code] in 
(
            SELECT DISTINCT [PMX_OSEL]."PmxZoneCode" AS "Code" FROM [PMX_OSSL] 
                 INNER JOIN [PMX_OSEL] ON [PMX_OSEL]."Code" = [PMX_OSSL]."Code" 
                 INNER JOIN [PMX_OSWH] ON [PMX_OSEL]."PmxWhsCode" = [PMX_OSWH]."Code" 
                 INNER JOIN [PMX_OSZO] ON [PMX_OSEL]."PmxZoneCode" = [PMX_OSZO]."Code" 
                 INNER JOIN [PMX_PLLI] ON [PMX_PLLI]."StorLocCode" = [PMX_OSEL]."Code" 
                 INNER JOIN [PMX_PLHE] ON [PMX_PLHE]."DocEntry" = [PMX_PLLI]."DocEntry" 
                 LEFT JOIN [PMX_WALO] ON [PMX_PLLI]."ItemCode" = [PMX_WALO]."ItemCode" 
                 AND [PMX_PLLI]."StorLocCode" = [PMX_WALO]."LocationCode" 
                 AND [PMX_PLLI]."QualityStatusCode" = [PMX_WALO]."QualityStatusCode" 
                 AND ISNULL([PMX_PLLI].[LogUnitIdentKey], 0) = ISNULL( [PMX_WALO].[LUID], 0)  
                 AND ISNULL([PMX_PLLI].[ItemTransactionalInfoKey], 0) = ISNULL( [PMX_WALO].[Itri], 0)  
                 LEFT JOIN [OUSR] ON [PMX_WALO]."UserSign" = [OUSR]."INTERNAL_K" 
                 WHERE "PMX_OSSL"."IsActive" = N'Y' AND "PMX_PLHE"."WaveKey" = 191 
                 AND ([PMX_WALO]."UserSign" IS NULL  OR [OUSR]."USER_CODE" = N'manager') 
)
    UNION ALL
    SELECT child.ParentCode, child.Code, child.Name, [Level] + 1
    FROM Locations parent
    INNER JOIN PMX_OSEL child ON parent.Parentcode = child.Code
)


SELECT Distinct Locations.code
,[PMX_OSZO]."Name" AS "Description", [PMX_OSEL]."PmxWhsCode" AS "WarehouseCode", [PMX_OSWH]."Name" AS "WarehouseName" 
FROM Locations
INNER JOIN PMX_OSZO ON PMX_OSZO.Code = Locations.Code
INNER JOIN PMX_OSEL ON PMX_OSEL.Code = Locations.Code
INNER JOIN [PMX_OSWH] ON [PMX_OSEL]."PmxWhsCode" = [PMX_OSWH]."Code"

LEFT JOIN PMX_WALO ON PMX_WALO.WaveKey = 191
AND ISNULL(PMX_WALO.ZoneCode, '') = PMX_OSEL.PmxZoneCode
LEFT JOIN [OUSR] AS OUSR ON PMX_WALO.usersign = [OUSR].[INTERNAL_K]
 
  
ORDER BY PMX_OSZO.Name*/

                DbQueryBuilder strBuilder = dbTool.GetQueryBuilder();

                bool isHana = dbTool.DbServerType == DbServerType.HANA;
                string extraDb = null;
                preQueryForHana = null;

                if ( isHana )
                {
                    extraDb = ExtraDatabaseHelper.GetDatabaseSchemaExtraDb( (PmxDbConnection)dbTool.DbConnection ).SchemaName;

                    strBuilder.Append( "CALL ", strBuilder.AsIdentifier( extraDb, GetLocationsForRecursiveZonesInTmpHanaProcedure.NAME ), "(" + SqlCommandHelper.FormatAndEscapeColumnValue( waveKey, CultureInfo.InvariantCulture ) + "," + SqlCommandHelper.FormatAndEscapeColumnValue( userCode, CultureInfo.InvariantCulture ) + ")" );

                    preQueryForHana = strBuilder.ToString();
                    strBuilder = dbTool.GetQueryBuilder();
                }
                else
                {
                    strBuilder.Append( Produmex.Sbo.Logex.Data.SboCommands.GetRecursiveLocationsLowToHighLevel( null, GetAllZonesForWave( waveKey, userCode, dbTool ), CultureInfo.InvariantCulture ) );
                }

                strBuilder.Append("SELECT DISTINCT ");

                string locationsTable = isHana ?
                    strBuilder.AsIdentifierWithAlias( extraDb, GetLocationsForRecursiveZonesInTmpHanaProcedure.TMP_TABLE, PmxRecursiveLocationViewDefinition.NAME ) :
                    strBuilder.AsIdentifier( PmxRecursiveLocationViewDefinition.NAME );

                SqlCommandHelper.AddTableColumnNameWithAs(strBuilder, PmxRecursiveLocationViewDefinition.NAME, PmxRecursiveLocationViewDefinition.Columns.Code.NAME, LocationDataSet.Columns.CODE.Name);

                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnNameWithAs(strBuilder, PmxOseZoneTable.NAME, PmxOseZoneTable.Columns.Name.NAME, LocationDataSet.Columns.DESCRIPTION.Name);

                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnNameWithAs(strBuilder, PmxOseGeneralTable.NAME, PmxOseGeneralTable.Columns.PmxWarehouseCode.NAME, LocationDataSet.Columns.WAREHOUSECODE.Name);

                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnNameWithAs(strBuilder, PmxOseWarehouseTable.NAME, PmxOseWarehouseTable.Columns.Name.NAME, LocationDataSet.Columns.WAREHOUSENAME.Name);

                strBuilder.Append(" FROM ");
                strBuilder.Append(locationsTable);
                strBuilder.Append(" INNER JOIN ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxOseGeneralTable.NAME));
                strBuilder.Append(" ON ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseGeneralTable.NAME, PmxOseGeneralTable.Columns.Code.NAME);
                strBuilder.Append(" = ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxRecursiveLocationViewDefinition.NAME, PmxRecursiveLocationViewDefinition.Columns.Code.NAME);

                strBuilder.Append(" INNER JOIN ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxOseWarehouseTable.NAME));
                strBuilder.Append(" ON ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseGeneralTable.NAME, PmxOseGeneralTable.Columns.PmxWarehouseCode.NAME);
                strBuilder.Append(" = ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseWarehouseTable.NAME, PmxOseWarehouseTable.Columns.Code.NAME);


                strBuilder.Append(" INNER JOIN ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxOseZoneTable.NAME));
                strBuilder.Append(" ON ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseGeneralTable.NAME, PmxOseGeneralTable.Columns.Code.NAME);
                strBuilder.Append(" = ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseZoneTable.NAME, PmxOseZoneTable.Columns.Code.NAME);


                strBuilder.Append(" LEFT JOIN ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxWaveLockTable.NAME));
                strBuilder.Append(" ON ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxWaveLockTable.NAME, PmxWaveLockTable.Columns.WaveKey.NAME);
                strBuilder.Append(" = ");
                strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(waveKey, CultureInfo.InvariantCulture));
                strBuilder.Append(" AND ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxWaveLockTable.NAME, PmxWaveLockTable.Columns.ZoneCode.NAME);
                strBuilder.Append(" = ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseZoneTable.NAME, PmxOseZoneTable.Columns.Code.NAME);

                strBuilder.Append(" LEFT JOIN ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxUserTable.NAME));
                strBuilder.Append(" ON ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxWaveLockTable.NAME, PmxWaveLockTable.Columns.UserSign.NAME);
                strBuilder.Append(" = ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxUserTable.NAME, PmxUserTable.Columns.InternalKey.NAME);


                strBuilder.Append(" WHERE ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseZoneTable.NAME, PmxOseZoneTable.Columns.UseForZonePicking.NAME);
                strBuilder.Append(" = ");
                strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(SboBooleanColumnValues.Y.VALUE, CultureInfo.InvariantCulture));

                if (!string.IsNullOrEmpty(pmxWhsCode))
                {
                    strBuilder.Append(" AND ");
                    SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseGeneralTable.NAME, PmxOseGeneralTable.Columns.PmxWarehouseCode.NAME);
                    strBuilder.Append(" = ");
                    strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(pmxWhsCode, CultureInfo.InvariantCulture));
                }


                strBuilder.Append(" AND (");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxWaveLockTable.NAME, PmxWaveLockTable.Columns.UserSign.NAME);
                strBuilder.Append(" IS NULL ");
                strBuilder.Append(" OR ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxUserTable.NAME, PmxUserTable.Columns.UserCode.NAME);
                strBuilder.Append(" = ");
                strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(userCode, CultureInfo.InvariantCulture));
                strBuilder.Append(") ");


                strBuilder.Append(" ORDER BY ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxOseZoneTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxOseZoneTable.Columns.Name.NAME));


                return strBuilder.ToString();

            }

            /// <summary>
            /// Gets all zones for wave.
            /// </summary>
            /// <param name="waveKey">The wave key.</param>
            /// <param name="userCode">The user code.</param>
            /// <param name="dbTool">The db tool.</param>
            /// <returns></returns>
            public static string GetAllZonesForWave( int waveKey, string userCode, DbTool dbTool )
            {
                DbQueryBuilder strBuilder = dbTool.GetQueryBuilder();

                strBuilder.Append( "SELECT DISTINCT " );
                SqlCommandHelper.AddTableColumnNameWithAs(strBuilder, PmxOseGeneralTable.NAME, PmxOseGeneralTable.Columns.PmxZoneCode.NAME, LocationDataSet.Columns.CODE.Name);
                strBuilder.Append(" FROM ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxOseStorageLocationGeneralTable.NAME));
                strBuilder.Append(" INNER JOIN ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxOseGeneralTable.NAME));
                strBuilder.Append(" ON ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseGeneralTable.NAME, PmxOseGeneralTable.Columns.Code.NAME);
                strBuilder.Append(" = ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseStorageLocationGeneralTable.NAME, PmxOseStorageLocationGeneralTable.Columns.Code.NAME);

                strBuilder.Append(" INNER JOIN ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxOseWarehouseTable.NAME));
                strBuilder.Append(" ON ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseGeneralTable.NAME, PmxOseGeneralTable.Columns.PmxWarehouseCode.NAME);
                strBuilder.Append(" = ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseWarehouseTable.NAME, PmxOseWarehouseTable.Columns.Code.NAME);


                strBuilder.Append(" INNER JOIN ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxOseZoneTable.NAME));
                strBuilder.Append(" ON ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseGeneralTable.NAME, PmxOseGeneralTable.Columns.PmxZoneCode.NAME);
                strBuilder.Append(" = ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseZoneTable.NAME, PmxOseZoneTable.Columns.Code.NAME);

                strBuilder.Append(" INNER JOIN ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxPickListLineTable.NAME));
                strBuilder.Append(" ON ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxPickListLineTable.NAME, PmxPickListLineTable.Columns.StorLocCode.NAME);
                strBuilder.Append(" = ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseGeneralTable.NAME, PmxOseGeneralTable.Columns.Code.NAME);

                strBuilder.Append(" INNER JOIN ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxPickListTable.NAME));
                strBuilder.Append(" ON ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxPickListTable.NAME, PmxPickListTable.Columns.DocEntry.NAME);
                strBuilder.Append(" = ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxPickListLineTable.NAME, PmxPickListLineTable.Columns.DocEntry.NAME);

                //Locking on wave item detail

                strBuilder.Append(" LEFT JOIN ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxWaveLockTable.NAME));
                strBuilder.Append(" ON ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxPickListLineTable.NAME, PmxPickListLineTable.Columns.ItemCode.NAME);
                strBuilder.Append(" = ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxWaveLockTable.NAME, PmxWaveLockTable.Columns.ItemCode.NAME);

                strBuilder.Append(" AND ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxPickListTable.NAME, PmxPickListTable.Columns.WaveKey.NAME);
                strBuilder.Append(" = ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxWaveLockTable.NAME, PmxWaveLockTable.Columns.WaveKey.NAME);


                strBuilder.AppendFormat(" AND " + strBuilder.IFNULL + "({0}, '') = " + strBuilder.IFNULL + "( {1}, '') ",
                    SqlCommandHelper.EscapeTableColumnName(PmxPickListLineTable.NAME, PmxPickListLineTable.Columns.StorLocCode.NAME),
                    SqlCommandHelper.EscapeTableColumnName(PmxWaveLockTable.NAME, PmxWaveLockTable.Columns.LocationCode.NAME));

                strBuilder.Append(" AND ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxPickListLineTable.NAME, PmxPickListLineTable.Columns.QualityStatusCode.NAME);
                strBuilder.Append(" = ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxWaveLockTable.NAME, PmxWaveLockTable.Columns.QualityStatusCode.NAME);

                strBuilder.AppendFormat(" AND " + strBuilder.IFNULL + "({0}, 0) = " + strBuilder.IFNULL + "( {1}, 0) ",
                SqlCommandHelper.EscapeTableColumnName(PmxPickListLineTable.NAME, PmxPickListLineTable.Columns.LogUnitIdentKey.NAME),
                SqlCommandHelper.EscapeTableColumnName(PmxWaveLockTable.NAME, PmxWaveLockTable.Columns.LUID.NAME));

                strBuilder.AppendFormat(" AND " + strBuilder.IFNULL + "({0}, 0) = " + strBuilder.IFNULL + "( {1}, 0) ",
                SqlCommandHelper.EscapeTableColumnName(PmxPickListLineTable.NAME, PmxPickListLineTable.Columns.ItemTransactionalInfoKey.NAME),
                SqlCommandHelper.EscapeTableColumnName(PmxWaveLockTable.NAME, PmxWaveLockTable.Columns.Itri.NAME));


                strBuilder.Append(" LEFT JOIN ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxUserTable.NAME));
                strBuilder.Append(" ON ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxWaveLockTable.NAME, PmxWaveLockTable.Columns.UserSign.NAME);
                strBuilder.Append(" = ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxUserTable.NAME, PmxUserTable.Columns.InternalKey.NAME);

                //Filter out WA locations
                strBuilder.Append(" LEFT JOIN ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxOseWarehouseAutomationTable.NAME));
                strBuilder.Append(" ON ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxPickListLineTable.NAME, PmxPickListLineTable.Columns.StorLocCode.NAME);
                strBuilder.Append(" = ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseWarehouseAutomationTable.NAME, PmxOseWarehouseAutomationTable.Columns.Code.NAME);


                strBuilder.Append(" WHERE ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxOseStorageLocationGeneralTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxOseStorageLocationGeneralTable.Columns.IsActive.NAME));
                strBuilder.Append(" = ");
                strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(SboBooleanColumnValues.Y.VALUE, CultureInfo.InvariantCulture));


                strBuilder.Append(" AND ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxPickListTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxPickListTable.Columns.WaveKey.NAME));
                strBuilder.Append(" = ");
                strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(waveKey, CultureInfo.InvariantCulture));

                strBuilder.Append(" AND ");
                strBuilder.Append(SqlCommandHelper.EscapeTableColumnName(PmxPickListLineTable.NAME, PmxPickListLineTable.Columns.PickListLineStatus.NAME));
                strBuilder.Append(" IN (");
                strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(PmxPickListStatusColumnValues.NotReady.VALUE, CultureInfo.InvariantCulture));
                strBuilder.Append(" , ");
                strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(PmxPickListStatusColumnValues.Ready.VALUE, CultureInfo.InvariantCulture));
                strBuilder.Append(" )");




                strBuilder.Append(" AND (");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxWaveLockTable.NAME, PmxWaveLockTable.Columns.UserSign.NAME);
                strBuilder.Append(" IS NULL ");
                strBuilder.Append(" OR ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxUserTable.NAME, PmxUserTable.Columns.UserCode.NAME);
                strBuilder.Append(" = ");
                strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(userCode, CultureInfo.InvariantCulture));
                strBuilder.Append(") ");

                //Filter out WA locations
                strBuilder.Append(" AND ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseWarehouseAutomationTable.NAME, PmxOseWarehouseAutomationTable.Columns.Code.NAME);
                strBuilder.Append(" IS NULL ");


                return strBuilder.ToString();
            }

            /// <summary>
            /// Gets all zones for wave.
            /// </summary>
            /// <param name="userCode">The user code.</param>
            /// <param name="dbTool">The db tool.</param>
            /// <returns></returns>
            public static string GetAllZonesForWaveJoinPart( string userCode, DbTool dbTool )
            {
                DbQueryBuilder strBuilder = dbTool.GetQueryBuilder();

                string joinTableName = "PickingZones";

                strBuilder.Append(" \r\n INNER JOIN  ( \r\n SELECT DISTINCT ");

                SqlCommandHelper.AddTableColumnNameWithAs(strBuilder, PmxOseGeneralTable.NAME, PmxOseGeneralTable.Columns.PmxZoneCode.NAME, LocationDataSet.Columns.CODE.Name);
                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnNameWithAs(strBuilder, PmxPickListTable.NAME, PmxPickListTable.Columns.WaveKey.NAME, PmxPickListTable.Columns.WaveKey.NAME);
                strBuilder.Append(" \r\n FROM ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxOseStorageLocationGeneralTable.NAME));
                strBuilder.Append(" \r\n INNER JOIN ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxOseGeneralTable.NAME));
                strBuilder.Append(" ON ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseGeneralTable.NAME, PmxOseGeneralTable.Columns.Code.NAME);
                strBuilder.Append(" = ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseStorageLocationGeneralTable.NAME, PmxOseStorageLocationGeneralTable.Columns.Code.NAME);

                strBuilder.Append(" \r\n INNER JOIN ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxOseWarehouseTable.NAME));
                strBuilder.Append(" ON ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseGeneralTable.NAME, PmxOseGeneralTable.Columns.PmxWarehouseCode.NAME);
                strBuilder.Append(" = ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseWarehouseTable.NAME, PmxOseWarehouseTable.Columns.Code.NAME);


                strBuilder.Append(" \r\n INNER JOIN ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxOseZoneTable.NAME));
                strBuilder.Append(" ON ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseGeneralTable.NAME, PmxOseGeneralTable.Columns.PmxZoneCode.NAME);
                strBuilder.Append(" = ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseZoneTable.NAME, PmxOseZoneTable.Columns.Code.NAME);

                strBuilder.Append(" INNER JOIN ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxPickListLineTable.NAME));
                strBuilder.Append(" ON ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxPickListLineTable.NAME, PmxPickListLineTable.Columns.StorLocCode.NAME);
                strBuilder.Append(" = ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseGeneralTable.NAME, PmxOseGeneralTable.Columns.Code.NAME);

                strBuilder.Append(" \r\n INNER JOIN ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxPickListTable.NAME));
                strBuilder.Append(" ON ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxPickListTable.NAME, PmxPickListTable.Columns.DocEntry.NAME);
                strBuilder.Append(" = ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxPickListLineTable.NAME, PmxPickListLineTable.Columns.DocEntry.NAME);

                strBuilder.Append(" \r\n LEFT JOIN ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxWaveLockTable.NAME));
                strBuilder.Append(" ON ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxPickListLineTable.NAME, PmxPickListLineTable.Columns.ItemCode.NAME);
                strBuilder.Append(" = ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxWaveLockTable.NAME, PmxWaveLockTable.Columns.ItemCode.NAME);

                strBuilder.Append(" AND ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxPickListLineTable.NAME, PmxPickListLineTable.Columns.StorLocCode.NAME);
                strBuilder.Append(" = ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxWaveLockTable.NAME, PmxWaveLockTable.Columns.LocationCode.NAME);

                strBuilder.Append(" AND ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxPickListLineTable.NAME, PmxPickListLineTable.Columns.QualityStatusCode.NAME);
                strBuilder.Append(" = ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxWaveLockTable.NAME, PmxWaveLockTable.Columns.QualityStatusCode.NAME);

                strBuilder.AppendFormat(" AND " + strBuilder.IFNULL + "({0}, 0) = " + strBuilder.IFNULL + "( {1}, 0) ",
                SqlCommandHelper.EscapeTableColumnName(PmxPickListLineTable.NAME, PmxPickListLineTable.Columns.LogUnitIdentKey.NAME),
                SqlCommandHelper.EscapeTableColumnName(PmxWaveLockTable.NAME, PmxWaveLockTable.Columns.LUID.NAME));

                strBuilder.AppendFormat(" AND " + strBuilder.IFNULL + "({0}, 0) = " + strBuilder.IFNULL + "( {1}, 0) ",
                SqlCommandHelper.EscapeTableColumnName(PmxPickListLineTable.NAME, PmxPickListLineTable.Columns.ItemTransactionalInfoKey.NAME),
                SqlCommandHelper.EscapeTableColumnName(PmxWaveLockTable.NAME, PmxWaveLockTable.Columns.Itri.NAME));


                strBuilder.Append(" \r\n LEFT JOIN ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxUserTable.NAME));
                strBuilder.Append(" ON ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxWaveLockTable.NAME, PmxWaveLockTable.Columns.UserSign.NAME);
                strBuilder.Append(" = ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxUserTable.NAME, PmxUserTable.Columns.InternalKey.NAME);

                strBuilder.Append(" \r\n WHERE ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxOseStorageLocationGeneralTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxOseStorageLocationGeneralTable.Columns.IsActive.NAME));
                strBuilder.Append(" = ");
                strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(SboBooleanColumnValues.Y.VALUE, CultureInfo.InvariantCulture));






                strBuilder.Append(" AND (");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxWaveLockTable.NAME, PmxWaveLockTable.Columns.UserSign.NAME);
                strBuilder.Append(" IS NULL ");
                strBuilder.Append(" OR ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxUserTable.NAME, PmxUserTable.Columns.UserCode.NAME);
                strBuilder.Append(" = ");
                strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(userCode, CultureInfo.InvariantCulture));
                strBuilder.Append(") ");

                strBuilder.Append(" AND ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxPickListTable.Columns.PickListStatus.NAME));
                strBuilder.Append(" IN (");
                strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(PmxPickListStatusColumnValues.NotReady.VALUE, CultureInfo.InvariantCulture));
                strBuilder.Append(" , ");
                strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(PmxPickListStatusColumnValues.Ready.VALUE, CultureInfo.InvariantCulture));
                strBuilder.Append(" , ");
                strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(PmxPickListStatusColumnValues.PartiallyReady.VALUE, CultureInfo.InvariantCulture));
                strBuilder.Append(" , ");
                strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(PmxPickListStatusColumnValues.PartiallyPicked.VALUE, CultureInfo.InvariantCulture));
                strBuilder.Append(" )");

                strBuilder.Append(" AND ");
                strBuilder.Append(SqlCommandHelper.EscapeTableColumnName(PmxPickListLineTable.NAME, PmxPickListLineTable.Columns.PickListLineStatus.NAME));
                strBuilder.Append(" IN (");
                strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(PmxPickListStatusColumnValues.NotReady.VALUE, CultureInfo.InvariantCulture));
                strBuilder.Append(" , ");
                strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(PmxPickListStatusColumnValues.Ready.VALUE, CultureInfo.InvariantCulture));
                strBuilder.Append(" )");

                strBuilder.Append(") AS ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(joinTableName));

                strBuilder.Append(" \r\n ON ");
                SqlCommandHelper.AddTableColumnName(strBuilder, joinTableName, LocationDataSet.Columns.CODE.Name);
                strBuilder.Append(" = ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseGeneralTable.NAME, PmxOseGeneralTable.Columns.Code.NAME);


                return strBuilder.ToString();
            }

            /// <summary>
            /// Gets all zones.
            /// </summary>
            /// <param name="waveKey">The wave key.</param>
            /// <returns></returns>
            public static string GetAllZones(int waveKey)
            {
                StringBuilder strBuilder = new StringBuilder();
                strBuilder.Append("SELECT DISTINCT ");

                SqlCommandHelper.AddTableColumnNameWithAs(strBuilder, PmxOseGeneralTable.NAME, PmxOseGeneralTable.Columns.PmxZoneCode.NAME, LocationDataSet.Columns.CODE.Name);
                strBuilder.Append(" FROM ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxOseStorageLocationGeneralTable.NAME));
                strBuilder.Append(" INNER JOIN ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxOseGeneralTable.NAME));
                strBuilder.Append(" ON ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseGeneralTable.NAME, PmxOseGeneralTable.Columns.Code.NAME);
                strBuilder.Append(" = ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseStorageLocationGeneralTable.NAME, PmxOseStorageLocationGeneralTable.Columns.Code.NAME);

                strBuilder.Append(" INNER JOIN ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxOseWarehouseTable.NAME));
                strBuilder.Append(" ON ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseGeneralTable.NAME, PmxOseGeneralTable.Columns.PmxWarehouseCode.NAME);
                strBuilder.Append(" = ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseWarehouseTable.NAME, PmxOseWarehouseTable.Columns.Code.NAME);


                strBuilder.Append(" INNER JOIN ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxOseZoneTable.NAME));
                strBuilder.Append(" ON ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseGeneralTable.NAME, PmxOseGeneralTable.Columns.PmxZoneCode.NAME);
                strBuilder.Append(" = ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseZoneTable.NAME, PmxOseZoneTable.Columns.Code.NAME);


                strBuilder.Append(" WHERE ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxOseStorageLocationGeneralTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxOseStorageLocationGeneralTable.Columns.IsActive.NAME));
                strBuilder.Append(" = ");
                strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(SboBooleanColumnValues.Y.VALUE, CultureInfo.InvariantCulture));





                return strBuilder.ToString();
            }



            /// <summary>
            /// Gets the pick list for picking by wave.
            /// </summary>
            /// <param name="userName">Name of the user.</param>
            /// <param name="pmxWarehouseCode">The PMX warehouse code.</param>
            /// <param name="deviceID">The Device ID</param>
            /// <param name="customDescrFields">The custom descr fields.</param>
            /// <param name="customDescrJoin">The custom descr join.</param>
            /// <param name="availablePickListTypes">The available pick list types.</param>
            /// <param name="availableItemPickType1">The available item pick type1.</param>
            /// <param name="availableItemPickType2">The available item pick type2.</param>
            /// <param name="orderBy">The order by.</param>
            /// <param name="isZonePicking">if set to <c>true</c> [is zone picking].</param>
            /// <param name="onlyCompleteUnlockedWaves">if set to <c>true</c> [only complete unlocked waves].</param>
            /// <param name="isProductionPicking">if set to <c>true</c> [is production picking].</param>
            /// <returns></returns>
            public static string GetPickListForPickingByWave(string userName, string pmxWarehouseCode, string deviceID, string[] customDescrFields, string customDescrJoin, List<string> availablePickListTypes, List<string> availableItemPickType1, List<string> availableItemPickType2, string orderBy, bool isZonePicking, bool onlyCompleteUnlockedWaves, bool isProductionPicking, DbTool dbTool)
            {
                DbQueryBuilder strBuilder = dbTool.GetQueryBuilder();

                //string tableNameJoin = "ZoneLocking";

                string tableNameUserForWALO = PmxUserTable.NAME + "_WALO";



                if (isZonePicking)
                {
                    //We need to add recursive zone selection
                    //strBuilder.Append(Produmex.Sbo.Logex.Data.SboCommands.GetRecursiveLocationsLowToHighLevelWithExtraColumnsAndJoin(null, GetAllZonesForWaveJoinPart(userName), null, new Collection<string>() { PmxPickListTable.Columns.WaveKey.NAME }, CultureInfo.InvariantCulture));

                }



                strBuilder.Append(" \r\n SELECT ");

                SqlCommandHelper.AddTableColumnNameWithAs(strBuilder, PmxWaveTable.NAME, PmxWaveTable.Columns.InternalKey.NAME, WaveDataSet.Columns.INTERNAL_KEY.Name);

                // AsString because DataSet filtering using a string containing non-numeric chars will give error on numeric columns
                strBuilder.Append(", CAST(");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxWaveTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxWaveTable.Columns.InternalKey.NAME));
                strBuilder.Append(" AS NVARCHAR) AS ");
                strBuilder.Append(strBuilder.AsIdentifier(WaveDataSet.Columns.INTERNAL_KEY_AS_STRING.Name));

                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnNameWithAs(strBuilder, PmxWaveTable.NAME, PmxWaveTable.Columns.DueDate.NAME, WaveDataSet.Columns.DUE_DATE.Name);
                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnNameWithAs(strBuilder, PmxWaveTable.NAME, PmxWaveTable.Columns.Description.NAME, WaveDataSet.Columns.DESCRIPTION.Name);
                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxWaveTable.NAME, PmxWaveTable.Columns.Priority.NAME);
                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxWaveTable.NAME, PmxWaveTable.Columns.SameDelivery.NAME);
                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseGeneralTable.NAME, PmxOseGeneralTable.Columns.PmxWarehouseCode.NAME);

                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxWaveTable.NAME, PmxWaveTable.Columns.ItemLabelsPrinted.NAME);

                bool includeORDR = false;
                if (customDescrFields != null && customDescrFields.Length > 0 &&
                    !isProductionPicking)
                {
                    strBuilder.Append(", NULLIF( ");

                    strBuilder.Flags( QBF.Push | QBF.Translate );

                    string concat = "";
                    foreach (string fldName in customDescrFields)
                    {
                        strBuilder.Append(concat);
						if (dbTool.DbServerType == DbServerType.HANA)
						{
							strBuilder.Append("IFNULL( NULLIF( TO_VARCHAR( ", fldName, " ), '' ), '.' )");
						}
						else
						{
							strBuilder.Append("IFNULL( NULLIF( CONVERT( NVARCHAR, ", fldName, " ), '' ), '.' )");
						}
                        concat = " || ' - ' || ";

                        includeORDR |= fldName.ToUpper().Contains(PmxSalesOrderTable.NAME.ToUpper());
                    }

                    strBuilder.Flags( QBF.Pop );

                    concat = ", '";
                    foreach (string fldName in customDescrFields)
                    {
                        strBuilder.Append(concat);
                        strBuilder.Append(".");
                        concat = " - ";
                    }
                    strBuilder.Append("' ) AS ");
                    strBuilder.Append(strBuilder.AsIdentifier(WaveDataSet.Columns.CUSTOM_DESCR.Name));
                }

                strBuilder.Append("\r\n FROM ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxWaveTable.NAME));
                strBuilder.Append("\r\n INNER JOIN ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxPickListTable.NAME));
                strBuilder.Append(" ON ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxWaveTable.NAME, PmxWaveTable.Columns.InternalKey.NAME);
                strBuilder.Append(" = ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxPickListTable.NAME, PmxPickListTable.Columns.WaveKey.NAME);

                strBuilder.Append("\r\n INNER JOIN ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxOseGeneralTable.NAME));
                strBuilder.Append(" ON ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseGeneralTable.NAME, PmxOseGeneralTable.Columns.Code.NAME);
                strBuilder.Append(" = ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxPickListTable.NAME, PmxPickListTable.Columns.DestStorLocCode.NAME);

                strBuilder.Append("\r\n LEFT JOIN ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxUserTable.NAME));
                strBuilder.Append(" ON ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxWaveTable.NAME, PmxWaveTable.Columns.LockedBy.NAME);
                strBuilder.Append(" = ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxUserTable.NAME, PmxUserTable.Columns.InternalKey.NAME);


                strBuilder.AppendFormat("\r\n LEFT JOIN (  {0}",
                    SqlCommandHelper.EscapeTableName(PmxPickListLineTable.NAME));
                {
                    strBuilder.AppendFormat("\r\n    INNER JOIN {0}",
                        SqlCommandHelper.EscapeTableName(PmxPickListProposalLineTable.NAME));
                    strBuilder.AppendFormat(" ON {0} = {1}",
                        SqlCommandHelper.EscapeTableColumnName(PmxPickListProposalLineTable.NAME, PmxPickListProposalLineTable.Columns.DocEntry.NAME),
                        SqlCommandHelper.EscapeTableColumnName(PmxPickListLineTable.NAME, PmxPickListLineTable.Columns.BaseEntry.NAME));
                    strBuilder.AppendFormat(" AND {0} = {1}",
                        SqlCommandHelper.EscapeTableColumnName(PmxPickListProposalLineTable.NAME, PmxPickListProposalLineTable.Columns.LineNum.NAME),
                        SqlCommandHelper.EscapeTableColumnName(PmxPickListLineTable.NAME, PmxPickListLineTable.Columns.BaseLine.NAME));
                    strBuilder.AppendFormat(" AND {0} = {1}",
                        SqlCommandHelper.EscapeTableColumnName(PmxPickListLineTable.NAME, PmxPickListLineTable.Columns.BaseType.NAME),
                        SqlCommandHelper.FormatAndEscapeColumnValue(Produmex.Sbo.Logex.Data.BusinessObjects.Definitions.PmxObjectTypes.PmxPickListProposal, CultureInfo.InvariantCulture));

                    if (includeORDR || (customDescrJoin != null && customDescrJoin.ToUpper().Contains(PmxSalesOrderTable.NAME) && !isProductionPicking))
                    {
                        strBuilder.AppendFormat("\r\n    INNER JOIN {0}",
                            SqlCommandHelper.EscapeTableName(PmxSalesOrderTable.NAME));
                        strBuilder.AppendFormat(" ON {0} = {1}",
                            SqlCommandHelper.EscapeTableColumnName(PmxSalesOrderTable.NAME, PmxSalesOrderTable.Columns.DocEntry.NAME),
                            SqlCommandHelper.EscapeTableColumnName(PmxPickListProposalLineTable.NAME, PmxPickListProposalLineTable.Columns.BaseEntry.NAME));
                        strBuilder.AppendFormat(" AND {0} = {1}",
                            SqlCommandHelper.EscapeTableColumnName(PmxPickListProposalLineTable.NAME, PmxPickListProposalLineTable.Columns.BaseType.NAME),
                            SqlCommandHelper.FormatAndEscapeColumnValue(((int)SboObjectType.SalesOrder).ToString(), CultureInfo.InvariantCulture));
                    }

                    strBuilder.AppendFormat("\r\n    LEFT JOIN {0}",
                        SqlCommandHelper.EscapeTableName(PmxOseWarehouseAutomationTable.NAME));
                    strBuilder.AppendFormat(" ON {0} = {1}",
                        SqlCommandHelper.EscapeTableColumnName(PmxOseWarehouseAutomationTable.NAME, PmxOseWarehouseAutomationTable.Columns.Code.NAME),
                        SqlCommandHelper.EscapeTableColumnName(PmxPickListLineTable.NAME, PmxPickListLineTable.Columns.StorLocCode.NAME));
                }
                strBuilder.AppendFormat("\r\n ) ON {0} = {1}",
                    SqlCommandHelper.EscapeTableColumnName(PmxPickListLineTable.NAME, PmxPickListLineTable.Columns.DocEntry.NAME),
                    SqlCommandHelper.EscapeTableColumnName(PmxPickListTable.NAME, PmxPickListTable.Columns.DocEntry.NAME));
                strBuilder.AppendFormat(" AND {0} IS NULL",
                    SqlCommandHelper.EscapeTableColumnName(PmxOseWarehouseAutomationTable.NAME, PmxOseWarehouseAutomationTable.Columns.Code.NAME));
                strBuilder.AppendFormat("\r\n  AND {0} IN ({1},{2})",
                    SqlCommandHelper.EscapeTableColumnName(PmxPickListLineTable.NAME, PmxPickListLineTable.Columns.PickListLineStatus.NAME),
                    SqlCommandHelper.FormatAndEscapeColumnValue(PmxPickListStatusColumnValues.NotReady.VALUE, CultureInfo.InvariantCulture),
                    SqlCommandHelper.FormatAndEscapeColumnValue(PmxPickListStatusColumnValues.Ready.VALUE, CultureInfo.InvariantCulture));

                if (!String.IsNullOrEmpty(customDescrJoin) && !isProductionPicking)
                {
                    strBuilder.Append("\r\n ");
                    strBuilder.Append(customDescrJoin);
                }



                if (isZonePicking)
                {
                    //add join part from recursive query
                    //strBuilder.Append("\r\n LEFT JOIN ( \r\n SELECT DISTINCT ");

                    //SqlCommandHelper.AddTableColumnNameWithAs(strBuilder, PmxRecursiveLocationViewDefinition.NAME, PmxPickListTable.Columns.WaveKey.NAME, PmxPickListTable.Columns.WaveKey.NAME);



                    //strBuilder.Append(" \r\n FROM ");
                    //strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxRecursiveLocationViewDefinition.NAME));
                    //strBuilder.Append(" \r\n INNER JOIN ");
                    //strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxOseGeneralTable.NAME));
                    //strBuilder.Append(" ON ");
                    //SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseGeneralTable.NAME, PmxOseGeneralTable.Columns.Code.NAME);
                    //strBuilder.Append(" = ");
                    //SqlCommandHelper.AddTableColumnName(strBuilder, PmxRecursiveLocationViewDefinition.NAME, PmxRecursiveLocationViewDefinition.Columns.Code.NAME);

                    //strBuilder.Append(" \r\n INNER JOIN ");
                    //strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxOseWarehouseTable.NAME));
                    //strBuilder.Append(" ON ");
                    //SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseGeneralTable.NAME, PmxOseGeneralTable.Columns.PmxWarehouseCode.NAME);
                    //strBuilder.Append(" = ");
                    //SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseWarehouseTable.NAME, PmxOseWarehouseTable.Columns.Code.NAME);


                    //strBuilder.Append(" \r\n INNER JOIN ");
                    //strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxOseZoneTable.NAME));
                    //strBuilder.Append(" ON ");
                    //SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseGeneralTable.NAME, PmxOseGeneralTable.Columns.Code.NAME);
                    //strBuilder.Append(" = ");
                    //SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseZoneTable.NAME, PmxOseZoneTable.Columns.Code.NAME);


                    //strBuilder.Append(" \r\n LEFT JOIN ");
                    //strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxWaveLockTable.NAME));
                    //strBuilder.Append(" ON ");
                    //SqlCommandHelper.AddTableColumnName(strBuilder, PmxRecursiveLocationViewDefinition.NAME, PmxWaveLockTable.Columns.WaveKey.NAME);
                    //strBuilder.Append(" = ");
                    //SqlCommandHelper.AddTableColumnName(strBuilder, PmxWaveLockTable.NAME, PmxPickListTable.Columns.WaveKey.NAME);
                    //strBuilder.Append(" AND ");
                    //SqlCommandHelper.AddTableColumnName(strBuilder, PmxWaveLockTable.NAME, PmxWaveLockTable.Columns.ZoneCode.NAME);
                    //strBuilder.Append(" = ");
                    //SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseZoneTable.NAME, PmxOseZoneTable.Columns.Code.NAME);


                    //strBuilder.Append(" \r\n LEFT JOIN ");
                    //strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxUserTable.NAME));
                    //strBuilder.Append(" ON ");
                    //SqlCommandHelper.AddTableColumnName(strBuilder, PmxWaveLockTable.NAME, PmxWaveLockTable.Columns.UserSign.NAME);
                    //strBuilder.Append(" = ");
                    //SqlCommandHelper.AddTableColumnName(strBuilder, PmxUserTable.NAME, PmxUserTable.Columns.InternalKey.NAME);


                    //strBuilder.Append(" \r\n WHERE ");
                    //SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseZoneTable.NAME, PmxOseZoneTable.Columns.UseForZonePicking.NAME);
                    //strBuilder.Append(" = ");
                    //strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(SboBooleanColumnValues.Y.VALUE, CultureInfo.InvariantCulture));



                    //strBuilder.Append(" AND (");
                    //{
                    //    SqlCommandHelper.AddTableColumnName(strBuilder, PmxWaveLockTable.NAME, PmxWaveLockTable.Columns.UserSign.NAME);
                    //    strBuilder.Append(" IS NULL ");
                    //    strBuilder.Append(" OR ");
                    //    SqlCommandHelper.AddTableColumnName(strBuilder, PmxUserTable.NAME, PmxUserTable.Columns.UserCode.NAME);
                    //    strBuilder.Append(" = ");
                    //    strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(userName, CultureInfo.InvariantCulture));
                    //}
                    //strBuilder.Append(") ");

                    //strBuilder.Append(" )  ");

                    //strBuilder.Append(" AS  ");
                    //strBuilder.Append(SqlCommandHelper.EscapeTableName(tableNameJoin));
                    //strBuilder.Append(" \r\n ON ");
                    //SqlCommandHelper.AddTableColumnName(strBuilder, tableNameJoin, PmxPickListTable.Columns.WaveKey.NAME);
                    //strBuilder.Append(" = ");
                    //SqlCommandHelper.AddTableColumnName(strBuilder, PmxWaveTable.NAME, PmxWaveTable.Columns.InternalKey.NAME);



                    //Join on PMX_WALO for locking on item detail level


                    strBuilder.Append(" LEFT JOIN ");
                    strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxWaveLockTable.NAME));
                    strBuilder.Append(" ON ");
                    SqlCommandHelper.AddTableColumnName(strBuilder, PmxPickListLineTable.NAME, PmxPickListLineTable.Columns.ItemCode.NAME);
                    strBuilder.Append(" = ");
                    SqlCommandHelper.AddTableColumnName(strBuilder, PmxWaveLockTable.NAME, PmxWaveLockTable.Columns.ItemCode.NAME);

                    strBuilder.Append(" AND ");
                    SqlCommandHelper.AddTableColumnName(strBuilder, PmxWaveTable.NAME, PmxWaveTable.Columns.InternalKey.NAME);
                    strBuilder.Append(" = ");
                    SqlCommandHelper.AddTableColumnName(strBuilder, PmxWaveLockTable.NAME, PmxWaveLockTable.Columns.WaveKey.NAME);


                    strBuilder.AppendFormat( " AND " + strBuilder.IFNULL + "({0}, '') = " + strBuilder.IFNULL + "( {1}, '') ",
                        SqlCommandHelper.EscapeTableColumnName(PmxPickListLineTable.NAME, PmxPickListLineTable.Columns.StorLocCode.NAME),
                        SqlCommandHelper.EscapeTableColumnName(PmxWaveLockTable.NAME, PmxWaveLockTable.Columns.LocationCode.NAME));


                    strBuilder.Append(" AND ");
                    SqlCommandHelper.AddTableColumnName(strBuilder, PmxPickListLineTable.NAME, PmxPickListLineTable.Columns.QualityStatusCode.NAME);
                    strBuilder.Append(" = ");
                    SqlCommandHelper.AddTableColumnName(strBuilder, PmxWaveLockTable.NAME, PmxWaveLockTable.Columns.QualityStatusCode.NAME);

                    strBuilder.AppendFormat( " AND " + strBuilder.IFNULL + "({0}, 0) = " + strBuilder.IFNULL + "( {1}, 0) ",
                    SqlCommandHelper.EscapeTableColumnName(PmxPickListLineTable.NAME, PmxPickListLineTable.Columns.LogUnitIdentKey.NAME),
                    SqlCommandHelper.EscapeTableColumnName(PmxWaveLockTable.NAME, PmxWaveLockTable.Columns.LUID.NAME));

                    strBuilder.AppendFormat( " AND " + strBuilder.IFNULL + "({0}, 0) = " + strBuilder.IFNULL + "( {1}, 0) ",
                    SqlCommandHelper.EscapeTableColumnName(PmxPickListLineTable.NAME, PmxPickListLineTable.Columns.ItemTransactionalInfoKey.NAME),
                    SqlCommandHelper.EscapeTableColumnName(PmxWaveLockTable.NAME, PmxWaveLockTable.Columns.Itri.NAME));


                    strBuilder.Append(" LEFT JOIN ");
                    strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxUserTable.NAME));
                    strBuilder.AppendFormat(" AS {0} ",
                        SqlCommandHelper.EscapeTableName(tableNameUserForWALO));
                    strBuilder.Append(" ON ");
                    SqlCommandHelper.AddTableColumnName(strBuilder, PmxWaveLockTable.NAME, PmxWaveLockTable.Columns.UserSign.NAME);
                    strBuilder.Append(" = ");
                    SqlCommandHelper.AddTableColumnName(strBuilder, tableNameUserForWALO, PmxUserTable.Columns.InternalKey.NAME);


                }



                strBuilder.Append("\r\n \r\n  WHERE ");
                strBuilder.Append(SqlCommandHelper.EscapeTableColumnName(PmxWaveTable.NAME, PmxWaveTable.Columns.InternalKey.NAME));
                {
                    strBuilder.Append(" IN  ( SELECT DISTINCT ");
                    SqlCommandHelper.AddTableColumnName(strBuilder, PmxPickListTable.NAME, PmxPickListTable.Columns.WaveKey.NAME);
                    strBuilder.Append(" FROM ");
                    strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxPickListTable.NAME));


                    strBuilder.Append(" WHERE ");
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxPickListTable.Columns.PickListStatus.NAME));
                    strBuilder.Append(" IN (");
                    strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(PmxPickListStatusColumnValues.NotReady.VALUE, CultureInfo.InvariantCulture));
                    strBuilder.Append(" , ");
                    strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(PmxPickListStatusColumnValues.Ready.VALUE, CultureInfo.InvariantCulture));
                    strBuilder.Append(" , ");
                    strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(PmxPickListStatusColumnValues.PartiallyReady.VALUE, CultureInfo.InvariantCulture));
                    strBuilder.Append(" , ");
                    strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(PmxPickListStatusColumnValues.PartiallyPicked.VALUE, CultureInfo.InvariantCulture));
                    strBuilder.Append(" ))");
                }

                if (isZonePicking && onlyCompleteUnlockedWaves)
                {
                    strBuilder.Append("\r\n\r\n AND ");
                    strBuilder.Append(SqlCommandHelper.EscapeTableColumnName(PmxWaveTable.NAME, PmxWaveTable.Columns.InternalKey.NAME));
                    {

                        strBuilder.Append(" NOT IN  ( \r\n SELECT ");
                        SqlCommandHelper.AddTableColumnName(strBuilder, PmxWaveLockTable.NAME, PmxWaveLockTable.Columns.WaveKey.NAME);
                        strBuilder.Append(" FROM ");
                        strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxWaveLockTable.NAME));

                        strBuilder.Append("\r\n  INNER JOIN ");
                        strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxWaveTable.NAME));
                        strBuilder.Append(" ON ");
                        SqlCommandHelper.AddTableColumnName(strBuilder, PmxWaveTable.NAME, PmxWaveTable.Columns.InternalKey.NAME);
                        strBuilder.Append(" = ");
                        SqlCommandHelper.AddTableColumnName(strBuilder, PmxWaveLockTable.NAME, PmxWaveLockTable.Columns.WaveKey.NAME);


                        strBuilder.Append("\r\n  INNER JOIN ");
                        strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxUserTable.NAME));
                        strBuilder.Append(" ON ");
                        SqlCommandHelper.AddTableColumnName(strBuilder, PmxWaveLockTable.NAME, PmxWaveLockTable.Columns.UserSign.NAME);
                        strBuilder.Append(" = ");
                        SqlCommandHelper.AddTableColumnName(strBuilder, PmxUserTable.NAME, PmxUserTable.Columns.InternalKey.NAME);


                        strBuilder.Append("\r\n  WHERE ");
                        strBuilder.Append(SqlCommandHelper.EscapeTableColumnName(PmxUserTable.NAME, PmxUserTable.Columns.UserCode.NAME));
                        strBuilder.Append(" <> ");
                        strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(userName, CultureInfo.InvariantCulture));
                        strBuilder.Append(" ) ");
                    }
                }



                strBuilder.Append(" AND (");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxWaveTable.NAME, PmxWaveTable.Columns.LockedBy.NAME);
                strBuilder.Append(" IS NULL ");
                strBuilder.Append(" OR ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxUserTable.NAME, PmxUserTable.Columns.UserCode.NAME);
                strBuilder.Append(" = ");
                strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(userName, CultureInfo.InvariantCulture));
                strBuilder.Append(") ");

                if (isZonePicking)
                {
                    strBuilder.Append(" AND (");
                    SqlCommandHelper.AddTableColumnName(strBuilder, PmxWaveLockTable.NAME, PmxWaveLockTable.Columns.UserSign.NAME);
                    strBuilder.Append(" IS NULL ");
                    strBuilder.Append(" OR ");
                    SqlCommandHelper.AddTableColumnName(strBuilder, tableNameUserForWALO, PmxUserTable.Columns.UserCode.NAME);
                    strBuilder.Append(" = ");
                    strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(userName, CultureInfo.InvariantCulture));
                    strBuilder.Append(") ");
                }


                strBuilder.Append("\r\n\r\n AND (");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxOseGeneralTable.Columns.PmxWarehouseCode.NAME));
                strBuilder.Append(" = ");
                strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(pmxWarehouseCode, CultureInfo.InvariantCulture));
                strBuilder.Append(" OR ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxOseGeneralTable.Columns.PmxWarehouseCode.NAME));
                strBuilder.Append(" IN ( ");
                {
                    strBuilder.AppendFormat("SELECT {0} FROM {1} WHERE {2} = {3}",
                        SqlCommandHelper.EscapeTableColumnName(PmxThinClientWarehouseTable.NAME, PmxThinClientWarehouseTable.Columns.PmxWarehouseCode.NAME),
                        SqlCommandHelper.EscapeTableName(PmxThinClientWarehouseTable.NAME),
                        SqlCommandHelper.EscapeTableColumnName(PmxThinClientWarehouseTable.NAME, PmxThinClientWarehouseTable.Columns.ParentCode.NAME),
                        SqlCommandHelper.FormatAndEscapeColumnValue(deviceID, CultureInfo.InvariantCulture));
                }
                strBuilder.Append(") ) ");

                strBuilder.Append(" AND ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxPickListTable.Columns.IsCustomerCollect.NAME));
                strBuilder.Append(" = ");
                strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(SboBooleanColumnValues.N.VALUE, CultureInfo.InvariantCulture));

                strBuilder.Append(" AND ");
                strBuilder.Append(SqlCommandHelper.EscapeTableColumnName(PmxPickListLineTable.NAME, PmxPickListLineTable.Columns.InvLockLevel.NAME));
                strBuilder.Append(" <> ");
                strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(PmxInventoryLockingLevelColumnValues.ItemNoLocking.VALUE, CultureInfo.InvariantCulture));

                if (availablePickListTypes.Count > 0)
                {
                    strBuilder.AppendFormat(" AND {0} IN ( {1} )",
                        SqlCommandHelper.EscapeColumnName(PmxPickListTable.Columns.PickListType.NAME),
                        SqlCommandHelper.FormatAndEscapeColumnValue(availablePickListTypes, CultureInfo.InvariantCulture));
                }

                if (availableItemPickType1 != null &&
                    availableItemPickType1.Count > 0)
                {
                    strBuilder.AppendFormat("\r\n\r\n AND {0} IN ( SELECT  {1} FROM {2} ",
                        SqlCommandHelper.EscapeTableColumnName(PmxPickListTable.NAME, PmxPickListTable.Columns.DocEntry.NAME),
                        SqlCommandHelper.EscapeTableColumnName(PmxPickListLineTable.NAME, PmxPickListLineTable.Columns.DocEntry.NAME),
                        SqlCommandHelper.EscapeTableName(PmxPickListLineTable.NAME));

                    strBuilder.AppendFormat("\r\n    INNER JOIN {0}",
                        SqlCommandHelper.EscapeTableName(PmxItemTable.NAME));
                    strBuilder.AppendFormat(" ON {0} = {1}",
                        SqlCommandHelper.EscapeTableColumnName(PmxItemTable.NAME, PmxItemTable.Columns.ItemCode.NAME),
                        SqlCommandHelper.EscapeTableColumnName(PmxPickListLineTable.NAME, PmxPickListLineTable.Columns.ItemCode.NAME));

                    strBuilder.AppendFormat( "\r\n    WHERE " + strBuilder.IFNULL + "({0}, '') = '' OR {0} IN ( {1} ))",
                        SqlCommandHelper.EscapeTableColumnName(PmxItemTable.NAME, PmxItemTable.Columns.PickType.NAME),
                        SqlCommandHelper.FormatAndEscapeColumnValue(availableItemPickType1, CultureInfo.InvariantCulture));
                }

                if (availableItemPickType2 != null &&
                    availableItemPickType2.Count > 0)
                {
                    strBuilder.AppendFormat("\r\n\r\n AND {0} IN ( SELECT  {1} FROM {2} ",
                        SqlCommandHelper.EscapeTableColumnName(PmxPickListTable.NAME, PmxPickListTable.Columns.DocEntry.NAME),
                        SqlCommandHelper.EscapeTableColumnName(PmxPickListLineTable.NAME, PmxPickListLineTable.Columns.DocEntry.NAME),
                        SqlCommandHelper.EscapeTableName(PmxPickListLineTable.NAME));

                    strBuilder.AppendFormat("\r\n    INNER JOIN {0}",
                        SqlCommandHelper.EscapeTableName(PmxItemTable.NAME));
                    strBuilder.AppendFormat(" ON {0} = {1}",
                        SqlCommandHelper.EscapeTableColumnName(PmxItemTable.NAME, PmxItemTable.Columns.ItemCode.NAME),
                        SqlCommandHelper.EscapeTableColumnName(PmxPickListLineTable.NAME, PmxPickListLineTable.Columns.ItemCode.NAME));

                    strBuilder.AppendFormat( "\r\n    WHERE " + strBuilder.IFNULL + "({0}, '') = '' OR {0} IN ( {1} ))",
                        SqlCommandHelper.EscapeTableColumnName(PmxItemTable.NAME, PmxItemTable.Columns.PickType2.NAME),
                        SqlCommandHelper.FormatAndEscapeColumnValue(availableItemPickType2, CultureInfo.InvariantCulture));
                }

                List<string> allowedPickObjTypes = new List<string>();
                if ( isProductionPicking )
                {
                    allowedPickObjTypes.Add( PmxPickObjectTypeColumnValues.Production.VALUE );
                    allowedPickObjTypes.Add( PmxPickObjectTypeColumnValues.WhsTransferProd.VALUE );
                }
                else
                {
                    allowedPickObjTypes.Add( PmxPickObjectTypeColumnValues.Sales.VALUE );
                    allowedPickObjTypes.Add( PmxPickObjectTypeColumnValues.WhsTransfer.VALUE );
                }
                strBuilder.AppendFormat( "\r\n AND {0} IN ({1}) ",
                   SqlCommandHelper.EscapeTableColumnName( PmxPickListTable.NAME, PmxPickListTable.Columns.PickObjType.NAME ),
                   SqlCommandHelper.FormatAndEscapeColumnValue( allowedPickObjTypes, CultureInfo.InvariantCulture ) );




                if (isZonePicking)
                {
                    //Only show wave if a zone can be found, or if no location is set yet ...
                    //strBuilder.Append(" AND (");
                    //strBuilder.Append(SqlCommandHelper.EscapeTableColumnName(tableNameJoin, PmxPickListTable.Columns.WaveKey.NAME));
                    //strBuilder.Append(" IS NOT NULL OR ");
                    //strBuilder.Append(SqlCommandHelper.EscapeTableColumnName(PmxPickListLineTable.NAME, PmxPickListLineTable.Columns.StorLocCode.NAME));
                    //strBuilder.Append(" IS NULL  ");

                    //strBuilder.Append(" )");
                }

                strBuilder.Append("\r\n GROUP BY ");
                strBuilder.Append(SqlCommandHelper.EscapeTableColumnName(PmxWaveTable.NAME, PmxWaveTable.Columns.LockedBy.NAME));
                strBuilder.Append(", ");
                strBuilder.Append(SqlCommandHelper.EscapeTableColumnName(PmxWaveTable.NAME, PmxWaveTable.Columns.Priority.NAME));
                strBuilder.Append(", ");
                strBuilder.Append(SqlCommandHelper.EscapeTableColumnName(PmxWaveTable.NAME, PmxWaveTable.Columns.DueDate.NAME));
                strBuilder.Append(", ");
                strBuilder.Append(SqlCommandHelper.EscapeTableColumnName(PmxWaveTable.NAME, PmxWaveTable.Columns.InternalKey.NAME));
                strBuilder.Append(", ");
                strBuilder.Append(SqlCommandHelper.EscapeTableColumnName(PmxWaveTable.NAME, PmxWaveTable.Columns.Description.NAME));
                strBuilder.Append(", ");
                strBuilder.Append(SqlCommandHelper.EscapeTableColumnName(PmxWaveTable.NAME, PmxWaveTable.Columns.SameDelivery.NAME));
                strBuilder.Append(", ");
                strBuilder.Append(SqlCommandHelper.EscapeTableColumnName(PmxWaveTable.NAME, PmxWaveTable.Columns.ItemLabelsPrinted.NAME));
                strBuilder.Append(", ");
                strBuilder.Append(SqlCommandHelper.EscapeTableColumnName(PmxOseGeneralTable.NAME, PmxOseGeneralTable.Columns.PmxWarehouseCode.NAME));

                if (customDescrFields != null && !isProductionPicking)
                {
                    foreach (string fldName in customDescrFields)
                    {
                        strBuilder.Append(", ");
                        strBuilder.Append(fldName);
                    }
                }

                strBuilder.Append("\r\n ORDER BY ");
                strBuilder.Append(SqlCommandHelper.EscapeTableColumnName(PmxWaveTable.NAME, PmxWaveTable.Columns.LockedBy.NAME));
                strBuilder.Append(" DESC, ");

                if (isZonePicking)
                {
                    strBuilder.Append(" MAX ( ");
                    strBuilder.Append(SqlCommandHelper.EscapeTableColumnName(PmxWaveLockTable.NAME, PmxWaveLockTable.Columns.UserSign.NAME));
                    strBuilder.Append(" ) DESC, ");
                }

                switch (orderBy)
                {
                    case PmxWaveOrderByCodes.DUE_PRIORITY_KEY:
                        strBuilder.Append(SqlCommandHelper.EscapeTableColumnName(PmxWaveTable.NAME, PmxWaveTable.Columns.DueDate.NAME));
                        strBuilder.Append(", ");
                        strBuilder.Append(SqlCommandHelper.EscapeTableColumnName(PmxWaveTable.NAME, PmxWaveTable.Columns.Priority.NAME));
                        strBuilder.Append(", ");
                        strBuilder.Append(SqlCommandHelper.EscapeTableColumnName(PmxWaveTable.NAME, PmxWaveTable.Columns.InternalKey.NAME));
                        // Next column adds no value as sorting on its primary key has already been done
                        //strBuilder.Append(", ");
                        //strBuilder.Append(SqlCommandHelper.EscapeTableColumnName(PmxWaveTable.NAME, PmxWaveTable.Columns.Description.NAME));

                        break;
                    default:
                        strBuilder.Append(SqlCommandHelper.EscapeTableColumnName(PmxWaveTable.NAME, PmxWaveTable.Columns.Priority.NAME));
                        strBuilder.Append(", ");
                        strBuilder.Append(SqlCommandHelper.EscapeTableColumnName(PmxWaveTable.NAME, PmxWaveTable.Columns.DueDate.NAME));
                        strBuilder.Append(", ");
                        strBuilder.Append(SqlCommandHelper.EscapeTableColumnName(PmxWaveTable.NAME, PmxWaveTable.Columns.InternalKey.NAME));
                        // Next column adds no value as sorting on its primary key has already been done
                        //strBuilder.Append(", ");
                        //strBuilder.Append(SqlCommandHelper.EscapeTableColumnName(PmxWaveTable.NAME, PmxWaveTable.Columns.Description.NAME));

                        break;
                }

                return strBuilder.ToString();
            }


            /// <summary>
            /// Gets all the movable locations.
            /// </summary>
            /// <param name="moveableLocationCode">The moveable location code.</param>
            /// <param name="pmxWarehouseCode">The PMX warehouse code.</param>
            /// <returns>
            /// The sql statement to get all the movable locations
            /// </returns>
            public static string GetMovableLocations(string moveableLocationCode, string pmxWarehouseCode)
            {
                StringBuilder strBuilder = new StringBuilder();
                strBuilder.Append("SELECT ");


                SqlCommandHelper.AddTableColumnNameWithAs(strBuilder, PmxOseStorageLocationGeneralTable.NAME, PmxOseStorageLocationGeneralTable.Columns.Code.NAME, LocationDataSet.Columns.CODE.Name);

                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnNameWithAs(strBuilder, PmxOseStorageLocationGeneralTable.NAME, PmxOseStorageLocationGeneralTable.Columns.Name.NAME, LocationDataSet.Columns.DESCRIPTION.Name);

                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnNameWithAs(strBuilder, PmxOseGeneralTable.NAME, PmxOseGeneralTable.Columns.PmxWarehouseCode.NAME, LocationDataSet.Columns.WAREHOUSECODE.Name);

                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnNameWithAs(strBuilder, PmxOseWarehouseTable.NAME, PmxOseWarehouseTable.Columns.Name.NAME, LocationDataSet.Columns.WAREHOUSENAME.Name);



                strBuilder.Append(" FROM ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxOseStorageLocationGeneralTable.NAME));

                strBuilder.Append(" INNER JOIN ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxOseGeneralTable.NAME));
                strBuilder.Append(" ON ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseGeneralTable.NAME, PmxOseGeneralTable.Columns.Code.NAME);
                strBuilder.Append(" = ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseStorageLocationGeneralTable.NAME, PmxOseStorageLocationGeneralTable.Columns.Code.NAME);

                strBuilder.Append(" INNER JOIN ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxOseWarehouseTable.NAME));
                strBuilder.Append(" ON ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseGeneralTable.NAME, PmxOseGeneralTable.Columns.PmxWarehouseCode.NAME);
                strBuilder.Append(" = ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseWarehouseTable.NAME, PmxOseWarehouseTable.Columns.Code.NAME);


                strBuilder.Append(" INNER JOIN ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxOseMovableLocationTable.NAME));
                strBuilder.Append(" ON ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseGeneralTable.NAME, PmxOseGeneralTable.Columns.Code.NAME);
                strBuilder.Append(" = ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseMovableLocationTable.NAME, PmxOseMovableLocationTable.Columns.Code.NAME);


                strBuilder.Append(" WHERE ");

                SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseStorageLocationGeneralTable.NAME, PmxOseStorageLocationGeneralTable.Columns.IsActive.NAME);
                strBuilder.Append(" = ");
                strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(SboBooleanColumnValues.Y.VALUE, CultureInfo.InvariantCulture));


                if (pmxWarehouseCode != null &&
                    pmxWarehouseCode.Trim() != "")
                {
                    strBuilder.Append(" AND ");
                    SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseGeneralTable.NAME, PmxOseGeneralTable.Columns.PmxWarehouseCode.NAME);
                    strBuilder.Append(" = ");
                    strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(pmxWarehouseCode, CultureInfo.InvariantCulture));

                }

                if (moveableLocationCode != null &&
                    moveableLocationCode.Trim() != "")
                {
                    strBuilder.Append(" AND ");
                    SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseGeneralTable.NAME, PmxOseGeneralTable.Columns.Code.NAME);
                    strBuilder.Append(" = ");
                    strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(moveableLocationCode, CultureInfo.InvariantCulture));

                }

                strBuilder.Append(" ORDER BY ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseStorageLocationGeneralTable.NAME, PmxOseStorageLocationGeneralTable.Columns.Name.NAME);


                return strBuilder.ToString();
            }


            /// <summary>
            /// Gets the logistic units for customer.
            /// </summary>
            /// <param name="cardCode">The card code.</param>
            /// <param name="shipToAddress">The ship to address.</param>
            /// <param name="routeKey">The route key.</param>
            /// <param name="luid">The luid.</param>
            /// <param name="pmxWarehouseCode">The PMX warehouse code.</param>
            /// <param name="moveToPmxWarehouse">The move automatic PMX warehouse.</param>
            /// <param name="forceSpecificMoveToWarehouse">if set to <c>true</c> [force specific move automatic warehouse].</param>
            /// <returns></returns>
            public static string GetLogisticUnitsForCustomer(string cardCode, string shipToAddress, int? routeKey, int? luid, string pmxWarehouseCode, string moveToPmxWarehouse, bool forceSpecificMoveToWarehouse)
            {
                Collection<string> cardCodes = new Collection<string>();
                cardCodes.Add(cardCode);
                Collection<string> shipToAddresses = new Collection<string>();
                shipToAddresses.Add(shipToAddress);
                return GetLogisticUnitsForCustomer(cardCodes, shipToAddresses, routeKey, luid, pmxWarehouseCode, moveToPmxWarehouse, forceSpecificMoveToWarehouse);
            }



            /// <summary>
            /// Gets the logistic units for customer.
            /// </summary>
            /// <param name="cardCodes">The card codes.</param>
            /// <param name="shipToAddresses">The ship to addresses.</param>
            /// <param name="routeKey">The route key.</param>
            /// <param name="luid">The luid.</param>
            /// <param name="pmxWarehouseCode">The PMX warehouse code.</param>
            /// <param name="moveToPmxWarehouse">The move automatic PMX warehouse.</param>
            /// <param name="forceSpecificMoveToWarehouse">if set to <c>true</c> [force specific move automatic warehouse].</param>
            /// <returns></returns>
            public static string GetLogisticUnitsForCustomer(Collection<string> cardCodes, Collection<string> shipToAddresses, int? routeKey, int? luid, string pmxWarehouseCode, string moveToPmxWarehouse, bool forceSpecificMoveToWarehouse)
            {
                StringBuilder strBuilder = new StringBuilder();
                strBuilder.Append("SELECT DISTINCT ");
              
                SqlCommandHelper.AddTableColumnNameWithAs(strBuilder, PmxPickingLUIDByCustomerViewDefinition.NAME, PmxPickingLUIDByCustomerViewDefinition.Columns.LUID.NAME, LogisticCarrierDataSet.Columns.LUID.Name);
                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnNameWithAs(strBuilder, PmxPickingLUIDByCustomerViewDefinition.NAME, PmxPickingLUIDByCustomerViewDefinition.Columns.SSCC.NAME, LogisticCarrierDataSet.Columns.SSCC.Name);
                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnNameWithAs(strBuilder, PmxPickingLUIDByCustomerViewDefinition.NAME, PmxPickingLUIDByCustomerViewDefinition.Columns.StorLocCode.NAME, LogisticCarrierDataSet.Columns.STORAGE_LOCATION_CODE.Name);
                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnNameWithAs(strBuilder, PmxPickingLUIDByCustomerViewDefinition.NAME, PmxPickingLUIDByCustomerViewDefinition.Columns.MasterLUID.NAME, LogisticCarrierDataSet.Columns.MASTER_LUID.Name);
                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnNameWithAs(strBuilder, PmxPickingLUIDByCustomerViewDefinition.NAME, PmxPickingLUIDByCustomerViewDefinition.Columns.MasterSSCC.NAME, LogisticCarrierDataSet.Columns.MASTER_SSCC.Name);


                strBuilder.Append(" FROM ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxPickingLUIDByCustomerViewDefinition.NAME));

                strBuilder.Append(" WHERE ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxPickingLUIDByCustomerViewDefinition.Columns.CardCode.NAME));
                strBuilder.Append(" IN ( ");

                string comma = "";
                {
                    foreach (string cardCode in cardCodes)
                    {
                        strBuilder.Append(comma);
                        strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(cardCode, CultureInfo.InvariantCulture));
                        comma = ", ";
                    }
                }
                strBuilder.Append(" ) ");

                strBuilder.Append(" AND ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxPickingLUIDByCustomerViewDefinition.Columns.ShipToAddress.NAME));
                strBuilder.Append(" IN ( ");

                comma = "";
                {
                    foreach (string shipToAddress in shipToAddresses)
                    {
                        strBuilder.Append(comma);
                        strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(shipToAddress, CultureInfo.InvariantCulture));
                        comma = ", ";
                    }
                }
                strBuilder.Append(" ) ");

                if (routeKey.HasValue)
                {
                    strBuilder.Append(" AND ");
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxPickingLUIDByCustomerViewDefinition.Columns.RouteKey.NAME));
                    strBuilder.Append(" = ");
                    strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(routeKey.Value, CultureInfo.InvariantCulture));
                }
                else
                {
                    strBuilder.Append(" AND ");
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxPickingLUIDByCustomerViewDefinition.Columns.RouteKey.NAME));
                    strBuilder.Append(" IS NULL ");
                }

                if (luid.HasValue)
                {
                    strBuilder.Append(" AND ");
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxPickingLUIDByCustomerViewDefinition.Columns.LUID.NAME));
                    strBuilder.Append(" = ");
                    strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(luid.Value, CultureInfo.InvariantCulture));
                }

                strBuilder.Append(" AND ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxPickingLUIDByCustomerViewDefinition.Columns.SSCC.NAME));
                strBuilder.Append(" IS NOT NULL ");

                if (!string.IsNullOrEmpty(pmxWarehouseCode))
                {
                    strBuilder.Append(" AND ");
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxPickingLUIDByCustomerViewDefinition.Columns.PmxWhsCode.NAME));
                    strBuilder.Append(" = ");
                    strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(pmxWarehouseCode, CultureInfo.InvariantCulture));
                }

                if (!string.IsNullOrEmpty(moveToPmxWarehouse))
                {
                    strBuilder.Append(" AND ");
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxPickingLUIDByCustomerViewDefinition.Columns.MoveToWarehouse.NAME));
                    strBuilder.Append(" = ");
                    strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(moveToPmxWarehouse, CultureInfo.InvariantCulture));
                }
                else if(forceSpecificMoveToWarehouse)
                {
                    strBuilder.Append(" AND ");
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxPickingLUIDByCustomerViewDefinition.Columns.MoveToWarehouse.NAME));
                    strBuilder.Append(" IS NULL ");

                }

                return strBuilder.ToString();
            }



            /// <summary>
            /// Gets the customer info for wave.
            /// </summary>
            /// <param name="waveKey">The wave key.</param>
            /// <returns></returns>
            public static string GetCustomerInfoForWave(int waveKey)
            {
                StringBuilder strBuilder = new StringBuilder();
                strBuilder.Append("SELECT DISTINCT ");

                SqlCommandHelper.AddTableColumnNameWithAs(strBuilder, PmxPickListTable.NAME, PmxPickListTable.Columns.CardCode.NAME, PmxPickListTable.Columns.CardCode.NAME);
                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnNameWithAs(strBuilder, PmxPickListTable.NAME, PmxPickListTable.Columns.ShipToAddress.NAME, PmxPickListTable.Columns.ShipToAddress.NAME);

                strBuilder.Append(" FROM ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxPickListTable.NAME));

                strBuilder.Append(" WHERE ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxPickListTable.Columns.WaveKey.NAME));
                strBuilder.Append(" = ");
                strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(waveKey, CultureInfo.InvariantCulture));


                return strBuilder.ToString();
            }



            /// <summary>
            /// Command for getting all logistic carriers order by name
            /// </summary>
            /// <param name="onlyInStock">if set to <c>true</c> the system will get the logistic carriers that are in stock and have no products on it (LUID IS NULL)</param>
            /// <param name="filterOutFrozenItems">if set to <c>true</c> [filter out frozen items].</param>
            /// <param name="dbTool">The db tool.</param>
            /// <returns>
            /// Command for getting all logistic carriers order by name
            /// </returns>
            public static string GetLogisticCarriersOrderByName(bool onlyInStock, bool filterOutFrozenItems, DbTool dbTool)
            {
                DbQueryBuilder strBuilder = dbTool.GetQueryBuilder();

                strBuilder.Append("SELECT DISTINCT ");
                SqlCommandHelper.AddTableColumnNameWithAs( strBuilder, SboItemTable.NAME, SboItemTable.Columns.ItemCode.NAME, ProductDataSet.Columns.PRODUCT_CODE.Name );

                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnNameWithAs( strBuilder, SboItemTable.NAME, PmxItemTable.Columns.CustomItemDescr.NAME, ProductDataSet.Columns.PRODUCT_DESCRIPTION.Name );

                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnNameWithAs( strBuilder, SboItemTable.NAME, SboItemTable.Columns.CodeBars.NAME, ProductDataSet.Columns.GTIN.Name );

                strBuilder.Append(", ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(SboItemTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxItemTable.Columns.IsLogisticCarrier.NAME));

                strBuilder.Append("\r\n FROM ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(SboItemTable.NAME));

                if (onlyInStock)
                {
                    strBuilder.Append("\r\n INNER JOIN ");
                    strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxInventoryTotalTable.NAME));
                    strBuilder.Append(" ON ", strBuilder.AsIdentifier(PmxInventoryTotalTable.NAME, PmxInventoryTotalTable.Columns.ItemCode.NAME));
                    strBuilder.Append(" = ", strBuilder.AsIdentifier(SboItemTable.NAME, SboItemTable.Columns.ItemCode.NAME));
                    strBuilder.Append(" AND ", strBuilder.AsIdentifier(PmxInventoryTotalTable.NAME, PmxInventoryTotalTable.Columns.LogUnitIdentKey.NAME));
                    strBuilder.Append(" IS NULL ");
                }

                strBuilder.Append("\r\n WHERE ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(SboItemTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(SboItemTable.Columns.InventoryItem.NAME));
                strBuilder.Append(" = ");
                strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(SboBooleanColumnValues.Y.VALUE, CultureInfo.InvariantCulture));

                strBuilder.Append(" AND ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(SboItemTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxItemTable.Columns.IsLogisticCarrier.NAME));
                strBuilder.Append(" = ");
                strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(SboBooleanColumnValues.Y.VALUE, CultureInfo.InvariantCulture));

                if (filterOutFrozenItems)
                {
                    Produmex.Sbo.Logex.Data.SboCommands.GetFrozenItemFilterSQLWherePart(strBuilder, " AND ");
                }

                strBuilder.Append("\r\n ORDER BY ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(SboItemTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxItemTable.Columns.CustomItemDescr.NAME));

                return strBuilder.ToString();
            }


            /// <summary>
            /// Gets all the pick location.
            /// </summary>
            /// <param name="warehouseCode">The warehouse code.</param>
            /// <returns>The sql statement to get all pick locations</returns>
            public static string GetPickLocation(string warehouseCode)
            {
                StringBuilder strBuilder = new StringBuilder();
                strBuilder.Append("SELECT ");

                SqlCommandHelper.AddTableColumnNameWithAs( strBuilder, PmxOseStorageLocationGeneralTable.NAME, PmxOseStorageLocationGeneralTable.Columns.Code.NAME, LocationDataSet.Columns.CODE.Name );

                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnNameWithAs( strBuilder, PmxOseStorageLocationGeneralTable.NAME, PmxOseStorageLocationGeneralTable.Columns.Name.NAME, LocationDataSet.Columns.DESCRIPTION.Name );

                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnNameWithAs( strBuilder, PmxOseGeneralTable.NAME, PmxOseGeneralTable.Columns.PmxWarehouseCode.NAME, LocationDataSet.Columns.WAREHOUSECODE.Name );

                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnNameWithAs(strBuilder, PmxOseWarehouseTable.NAME, PmxOseWarehouseTable.Columns.Name.NAME, LocationDataSet.Columns.WAREHOUSENAME.Name);


                strBuilder.Append(" FROM ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxOseStorageLocationGeneralTable.NAME));
                strBuilder.Append(" INNER JOIN ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxOseGeneralTable.NAME));
                strBuilder.Append(" ON ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseGeneralTable.NAME, PmxOseGeneralTable.Columns.Code.NAME);
                strBuilder.Append(" = ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseStorageLocationGeneralTable.NAME, PmxOseStorageLocationGeneralTable.Columns.Code.NAME);

                strBuilder.Append(" INNER JOIN ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxOseWarehouseTable.NAME));
                strBuilder.Append(" ON ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseGeneralTable.NAME, PmxOseGeneralTable.Columns.PmxWarehouseCode.NAME);
                strBuilder.Append(" = ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseWarehouseTable.NAME, PmxOseWarehouseTable.Columns.Code.NAME);

                strBuilder.Append(" LEFT JOIN ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxOseBinTable.NAME));
                strBuilder.Append(" ON ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseBinTable.NAME, PmxOseBinTable.Columns.Code.NAME);
                strBuilder.Append(" = ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseStorageLocationGeneralTable.NAME, PmxOseStorageLocationGeneralTable.Columns.Code.NAME);

                strBuilder.Append(" WHERE (");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxOseStorageLocationGeneralTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxOseStorageLocationGeneralTable.Columns.IsPickLoc.NAME));
                strBuilder.Append(" = ");
                strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(SboBooleanColumnValues.Y.VALUE, CultureInfo.InvariantCulture));

                strBuilder.Append(" OR ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxOseBinTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxOseBinTable.Columns.Code.NAME));
                strBuilder.Append(" IS NOT NULL ");

                strBuilder.Append(") AND ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxOseStorageLocationGeneralTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxOseStorageLocationGeneralTable.Columns.IsActive.NAME));
                strBuilder.Append(" = ");
                strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(SboBooleanColumnValues.Y.VALUE, CultureInfo.InvariantCulture));


                if (warehouseCode != null &&
                    warehouseCode.Trim() != "")
                {
                    strBuilder.Append(" AND ");
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxOseGeneralTable.NAME));
                    strBuilder.Append(".");
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxOseGeneralTable.Columns.PmxWarehouseCode.NAME));
                    strBuilder.Append(" = ");
                    strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(warehouseCode, CultureInfo.InvariantCulture));

                }

                strBuilder.Append(" ORDER BY ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxOseStorageLocationGeneralTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxOseStorageLocationGeneralTable.Columns.Name.NAME));


                return strBuilder.ToString();
            }




            /// <summary>
            /// Gets the locations alternates.
            /// </summary>
            /// <param name="locationCode">The location code.</param>
            /// <param name="warehouseCode">The warehouse code.</param>
            /// <param name="itemCode">The item code.</param>
            /// <returns></returns>
            public static string GetLocationsAlternates(string locationCode, string warehouseCode, string itemCode)
            {
                StringBuilder strBuilder = new StringBuilder();
                strBuilder.Append("SELECT DISTINCT ");

                SqlCommandHelper.AddTableColumnNameWithAs( strBuilder, PmxOseStorageLocationGeneralTable.NAME, PmxOseStorageLocationGeneralTable.Columns.Code.NAME, LocationDataSet.Columns.CODE.Name );

                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnNameWithAs( strBuilder, PmxOseStorageLocationGeneralTable.NAME, PmxOseStorageLocationGeneralTable.Columns.Name.NAME, LocationDataSet.Columns.DESCRIPTION.Name );

                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnNameWithAs( strBuilder, PmxOseGeneralTable.NAME, PmxOseGeneralTable.Columns.PmxWarehouseCode.NAME, LocationDataSet.Columns.WAREHOUSECODE.Name );

                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnNameWithAs(strBuilder, PmxOseWarehouseTable.NAME, PmxOseWarehouseTable.Columns.Name.NAME, LocationDataSet.Columns.WAREHOUSENAME.Name);


                strBuilder.Append(" FROM ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxOseStorageLocationGeneralTable.NAME));
                strBuilder.Append(" INNER JOIN ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxOseGeneralTable.NAME));
                strBuilder.Append(" ON ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseGeneralTable.NAME, PmxOseGeneralTable.Columns.Code.NAME);
                strBuilder.Append(" = ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseStorageLocationGeneralTable.NAME, PmxOseStorageLocationGeneralTable.Columns.Code.NAME);

                strBuilder.Append(" INNER JOIN ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxOseWarehouseTable.NAME));
                strBuilder.Append(" ON ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseGeneralTable.NAME, PmxOseGeneralTable.Columns.PmxWarehouseCode.NAME);
                strBuilder.Append(" = ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseWarehouseTable.NAME, PmxOseWarehouseTable.Columns.Code.NAME);


                strBuilder.Append(" INNER JOIN ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxInventoryTotalTable.NAME));
                strBuilder.Append(" ON ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseGeneralTable.NAME, PmxOseGeneralTable.Columns.Code.NAME);
                strBuilder.Append(" = ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxInventoryTotalTable.NAME, PmxInventoryTotalTable.Columns.StorLocCode.NAME);


                strBuilder.Append(" WHERE ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxOseStorageLocationGeneralTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxOseStorageLocationGeneralTable.Columns.IsPickLoc.NAME));
                strBuilder.Append(" = ");
                strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(SboBooleanColumnValues.Y.VALUE, CultureInfo.InvariantCulture));

                strBuilder.Append(" AND ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxOseStorageLocationGeneralTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxOseStorageLocationGeneralTable.Columns.IsActive.NAME));
                strBuilder.Append(" = ");
                strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(SboBooleanColumnValues.Y.VALUE, CultureInfo.InvariantCulture));


                if (warehouseCode != null &&
                    warehouseCode.Trim() != "")
                {
                    strBuilder.Append(" AND ");
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxOseGeneralTable.NAME));
                    strBuilder.Append(".");
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxOseGeneralTable.Columns.PmxWarehouseCode.NAME));
                    strBuilder.Append(" = ");
                    strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(warehouseCode, CultureInfo.InvariantCulture));

                }

                if (locationCode != null &&
                    locationCode.Trim() != "")
                {
                    strBuilder.Append(" AND ");
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxOseGeneralTable.NAME));
                    strBuilder.Append(".");
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxOseGeneralTable.Columns.Code.NAME));
                    strBuilder.Append(" = ");
                    strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(locationCode, CultureInfo.InvariantCulture));

                }


                if (itemCode != null &&
                    itemCode.Trim() != "")
                {
                    strBuilder.Append(" AND ");
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxInventoryTotalTable.NAME));
                    strBuilder.Append(".");
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxInventoryTotalTable.Columns.ItemCode.NAME));
                    strBuilder.Append(" = ");
                    strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(itemCode, CultureInfo.InvariantCulture));

                }

                strBuilder.Append(" ORDER BY ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxOseStorageLocationGeneralTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxOseStorageLocationGeneralTable.Columns.Name.NAME));


                return strBuilder.ToString();
            }


            /// <summary>
            /// Command for getting all products order by name
            /// </summary>
            /// <param name="gtin">The gtin.</param>
            /// <param name="itemCode">The item code.</param>
            /// <param name="locationCode">The location code.</param>
            /// <param name="sscc">The SSCC.</param>
            /// <param name="forceSpecific">if set to <c>true</c>, and the sscc is NULL, we only get inventory for items where the sscc is NULL.</param>
            /// <param name="packagingUnit">The packaging unit.</param>
            /// <param name="filterOutFrozenItems">if set to <c>true</c> [filter out frozen items].</param>
            /// <param name="dbTool">The db tool.</param>
            /// <returns>
            /// Command for getting all products order by name
            /// </returns>
            public static string GetProductsFromStockOrderByName(string gtin, string itemCode, string locationCode, string sscc, bool forceSpecific, Produmex.Sbo.Logex.Data.SboCommands.PackagingUnit packagingUnit, bool filterOutFrozenItems, DbTool dbTool)
            {
                return GetProductsFromStockOrderByName(gtin, itemCode, locationCode, sscc, forceSpecific, packagingUnit, null, filterOutFrozenItems, dbTool);
            }


            /// <summary>
            /// Gets the name of the products from stock order by.
            /// </summary>
            /// <param name="gtin">The gtin.</param>
            /// <param name="itemCode">The item code.</param>
            /// <param name="locationCode">The location code.</param>
            /// <param name="sscc">The SSCC.</param>
            /// <param name="forceSpecific">if set to <c>true</c> [force specific].</param>
            /// <param name="packagingUnit">The packaging unit.</param>
            /// <param name="logisticCarrierLocationCode">The logistic carrier location code.</param>
            /// <param name="filterOutFrozenItems">if set to <c>true</c> [filter out frozen items].</param>
            /// <param name="dbTool">The db tool.</param>
            /// <returns></returns>
            public static string GetProductsFromStockOrderByName(string gtin, string itemCode, string locationCode, string sscc, bool forceSpecific, Produmex.Sbo.Logex.Data.SboCommands.PackagingUnit packagingUnit, string logisticCarrierLocationCode, bool filterOutFrozenItems, DbTool dbTool)
            {
                DbQueryBuilder strBuilder = dbTool.GetQueryBuilder();
                strBuilder.Append("SELECT DISTINCT ");

                SqlCommandHelper.AddTableColumnNameWithAs( strBuilder, SboItemTable.NAME, SboItemTable.Columns.ItemCode.NAME, ProductDataSet.Columns.PRODUCT_CODE.Name );

                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnNameWithAs( strBuilder, SboItemTable.NAME, PmxItemTable.Columns.CustomItemDescr.NAME, ProductDataSet.Columns.PRODUCT_DESCRIPTION.Name );

                strBuilder.Append(", ");
                strBuilder.Append(SboRecordsetHelper.GetSelectPartsColumnAndIsColumnNull(PmxItemTable.NAME, PmxItemTable.Columns.DefaultStorageLocationOrZoneCode.NAME, ProductDataSet.Columns.DEFAULT_LOCATION_CODE.Name));

                strBuilder.Append(", ");
                Produmex.Sbo.Logex.Data.SboCommands.AddBarcodeUsingPackagingUnit(strBuilder, packagingUnit, ProductDataSet.Columns.GTIN.Name);

                strBuilder.Append(", ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(SboItemTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxItemTable.Columns.HasBestBeforeDate.NAME));

                strBuilder.Append(", ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(SboItemTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(SboItemTable.Columns.ManBtchNum.NAME));

                strBuilder.Append(", ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(SboItemTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxItemTable.Columns.IsLogisticUnit.NAME));

                strBuilder.Append(", ",strBuilder.AsIdentifier(SboItemTable.NAME, SboItemTable.Columns.InventoryUom.NAME));

                strBuilder.Append(" FROM ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(SboItemTable.NAME));

                //inner join Product - inventory
                strBuilder.Append(" INNER JOIN ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxInventoryTotalTable.NAME));
                strBuilder.Append(" ON ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxInventoryTotalTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxInventoryTotalTable.Columns.ItemCode.NAME));
                strBuilder.Append(" = ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(SboItemTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(SboItemTable.Columns.ItemCode.NAME));

                strBuilder.Append(" LEFT JOIN ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxItemPackagingTypeTable.NAME));
                strBuilder.Append(" ON ",strBuilder.AsIdentifier(PmxItemPackagingTypeTable.NAME, PmxItemPackagingTypeTable.Columns.ItemCode.NAME));
                strBuilder.Append(" = ",strBuilder.AsIdentifier(SboItemTable.NAME, SboItemTable.Columns.ItemCode.NAME));


                string operand = " WHERE ";
                if (!string.IsNullOrEmpty(gtin))
                {
                    strBuilder.Append(operand);
                    Produmex.Sbo.Logex.Data.SboCommands.AddWhereClauseSearchForBarcode(strBuilder, gtin, true);
                    operand = " AND ";
                }

                if (!string.IsNullOrEmpty(itemCode))
                {
                    strBuilder.Append(operand);
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(SboItemTable.NAME));
                    strBuilder.Append(".");
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(SboItemTable.Columns.ItemCode.NAME));
                    strBuilder.Append(" = ");
                    strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(itemCode, CultureInfo.InvariantCulture));
                    operand = " AND ";
                }

                if (!string.IsNullOrEmpty(sscc))
                {
                    strBuilder.Append(operand);
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxInventoryTotalTable.NAME));
                    strBuilder.Append(".");
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxInventoryTotalTable.Columns.SSCC.NAME));
                    strBuilder.Append(" = ");
                    strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(sscc, CultureInfo.InvariantCulture));
                    operand = " AND ";
                }
                else if (forceSpecific)
                {
                    strBuilder.Append(operand);
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxInventoryTotalTable.NAME));
                    strBuilder.Append(".");
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxInventoryTotalTable.Columns.SSCC.NAME));
                    strBuilder.Append(" IS NULL ");
                    operand = " AND ";

                }

                if (!string.IsNullOrEmpty(locationCode) &&
                    !string.IsNullOrEmpty(logisticCarrierLocationCode))
                {
                    strBuilder.Append(operand);
                    strBuilder.Append("(");


                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxInventoryTotalTable.NAME));
                    strBuilder.Append(".");
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxInventoryTotalTable.Columns.StorLocCode.NAME));
                    strBuilder.Append(" = ");
                    strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(locationCode, CultureInfo.InvariantCulture));

                    strBuilder.Append(" OR ");
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxInventoryTotalTable.NAME));
                    strBuilder.Append(".");
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxInventoryTotalTable.Columns.StorLocCode.NAME));
                    strBuilder.Append(" = ");
                    strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(logisticCarrierLocationCode, CultureInfo.InvariantCulture));

                    strBuilder.Append(")");

                    operand = " AND ";
                }
                else
                {

                    if (!string.IsNullOrEmpty(locationCode))
                    {
                        strBuilder.Append(operand);
                        strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxInventoryTotalTable.NAME));
                        strBuilder.Append(".");
                        strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxInventoryTotalTable.Columns.StorLocCode.NAME));
                        strBuilder.Append(" = ");
                        strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(locationCode, CultureInfo.InvariantCulture));
                        operand = " AND ";
                    }

                    if (!string.IsNullOrEmpty(logisticCarrierLocationCode))
                    {
                        strBuilder.Append(operand);
                        strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxInventoryTotalTable.NAME));
                        strBuilder.Append(".");
                        strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxInventoryTotalTable.Columns.StorLocCode.NAME));
                        strBuilder.Append(" = ");
                        strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(logisticCarrierLocationCode, CultureInfo.InvariantCulture));
                        operand = " AND ";
                    }
                }


                strBuilder.Append(operand);
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(SboItemTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(SboItemTable.Columns.InventoryItem.NAME));
                strBuilder.Append(" = ");
                strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(SboBooleanColumnValues.Y.VALUE, CultureInfo.InvariantCulture));
                operand = " AND ";

                if (filterOutFrozenItems)
                {
                    Produmex.Sbo.Logex.Data.SboCommands.GetFrozenItemFilterSQLWherePart(strBuilder, operand);
                }


                strBuilder.Append(" ORDER BY ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(SboItemTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxItemTable.Columns.CustomItemDescr.NAME));
                return strBuilder.ToString();
            }


            /// <summary>
            /// Gets the batch numbers.
            /// </summary>
            /// <param name="sscc">The SSCC.</param>
            /// <param name="location">The location.</param>
            /// <param name="itemCode">The item code.</param>
            /// <param name="batchNumber">The batch number.</param>
            /// <param name="bestBeforeDate">The best before date.</param>
            /// <returns></returns>
            public static string GetBatchNumbers(String sscc, string location, string itemCode, string batchNumber, DateTime? bestBeforeDate)
            {
                StringBuilder strBuilder = new StringBuilder();
                strBuilder.Append("SELECT DISTINCT ");

                //strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxInventoryTotalTable.NAME));
                //strBuilder.Append(".");
                //strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxInventoryTotalTable.Columns.BatchNumber.NAME));
                //strBuilder.Append(" AS ");
                //strBuilder.Append(SqlCommandHelper.EscapeColumnName(BatchNumberDataSet.Columns.BATCH_NUMBER.Name));

                //strBuilder.Append(", ");
                //strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxInventoryTotalTable.NAME));
                //strBuilder.Append(".");
                //strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxInventoryTotalTable.Columns.BestBeforeDate.NAME));
                //strBuilder.Append(" AS ");
                //strBuilder.Append(SqlCommandHelper.EscapeColumnName(BatchNumberDataSet.Columns.BEST_BEFORE_DATE.Name));

                strBuilder.Append(SboRecordsetHelper.GetSelectPartsColumnAndIsColumnNull(PmxItemTransactionalInfoTable.NAME, PmxItemTransactionalInfoTable.Columns.BatchNumber1.NAME, BatchNumberDataSet.Columns.BATCH_NUMBER.Name));
                strBuilder.Append(", ");
                strBuilder.Append(SboRecordsetHelper.GetSelectPartsColumnAndIsColumnNull(PmxItemTransactionalInfoTable.NAME, PmxItemTransactionalInfoTable.Columns.BatchNumber2.NAME, BatchNumberDataSet.Columns.BATCH_NUMBER2.Name));
                strBuilder.Append(", ");
                strBuilder.Append(SboRecordsetHelper.GetSelectPartsColumnAndIsColumnNull(PmxItemTransactionalInfoTable.NAME, PmxItemTransactionalInfoTable.Columns.BestBeforeDate.NAME, BatchNumberDataSet.Columns.BEST_BEFORE_DATE.Name));
                strBuilder.Append(", ");
                strBuilder.Append(SboRecordsetHelper.GetSelectPartsColumnAndIsColumnNull(PmxInventoryTotalTable.NAME, PmxInventoryTotalTable.Columns.QualityStatusCode.NAME, BatchNumberDataSet.Columns.QUALITY_STATUS_CODE.Name));
                strBuilder.Append(", ");
                strBuilder.Append(SboRecordsetHelper.GetSelectPartsColumnAndIsColumnNull(PmxInventoryTotalTable.NAME, PmxInventoryTotalTable.Columns.ItemTransactionalInfoKey.NAME, BatchNumberDataSet.Columns.ITEMTRANSACTIONALINFOKEY.Name));
                strBuilder.Append(", ");
                SqlCommandHelper.AddDoubleZeroColumnWithName(strBuilder, BatchNumberDataSet.Columns.QUANTITY.Name);


                strBuilder.Append(" FROM ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxInventoryTotalTable.NAME));

                strBuilder.Append(" LEFT JOIN ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxItemTransactionalInfoTable.NAME));
                strBuilder.Append(" ON ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxInventoryTotalTable.NAME, PmxInventoryTotalTable.Columns.ItemTransactionalInfoKey.NAME);
                strBuilder.Append(" = ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxItemTransactionalInfoTable.NAME, PmxItemTransactionalInfoTable.Columns.InternalKey.NAME);



                string operand = " WHERE ";

                if (!string.IsNullOrEmpty(sscc))
                {
                    strBuilder.Append(operand);
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxInventoryTotalTable.NAME));
                    strBuilder.Append(".");
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxInventoryTotalTable.Columns.SSCC.NAME));
                    strBuilder.Append(" = ");
                    strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(sscc, CultureInfo.InvariantCulture));
                    operand = " AND ";
                }

                if (!string.IsNullOrEmpty(location))
                {
                    strBuilder.Append(operand);
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxInventoryTotalTable.NAME));
                    strBuilder.Append(".");
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxInventoryTotalTable.Columns.StorLocCode.NAME));
                    strBuilder.Append(" = ");
                    strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(location, CultureInfo.InvariantCulture));
                    operand = " AND ";
                }

                if (!string.IsNullOrEmpty(itemCode))
                {
                    strBuilder.Append(operand);
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxInventoryTotalTable.NAME));
                    strBuilder.Append(".");
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxInventoryTotalTable.Columns.ItemCode.NAME));
                    strBuilder.Append(" = ");
                    strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(itemCode, CultureInfo.InvariantCulture));
                    operand = " AND ";
                }

                if (!string.IsNullOrEmpty(batchNumber))
                {
                    strBuilder.Append(operand);
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxItemTransactionalInfoTable.NAME));
                    strBuilder.Append(".");
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxItemTransactionalInfoTable.Columns.BatchNumber1.NAME));
                    strBuilder.Append(" = ");
                    strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(batchNumber, CultureInfo.InvariantCulture));
                    operand = " AND ";
                }

                if (bestBeforeDate.HasValue)
                {
                    strBuilder.Append(operand);
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxItemTransactionalInfoTable.NAME));
                    strBuilder.Append(".");
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxItemTransactionalInfoTable.Columns.BestBeforeDate.NAME));
                    strBuilder.Append(" = ");
                    strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(bestBeforeDate.Value, CultureInfo.InvariantCulture));
                    operand = " AND ";
                }


                strBuilder.Append(" ORDER BY ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxItemTransactionalInfoTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxItemTransactionalInfoTable.Columns.BestBeforeDate.NAME));



                return strBuilder.ToString();
            }

            /// <summary>
            /// Gets the SSCC info.
            /// </summary>
            /// <param name="luid">The luid.</param>
            /// <param name="masterLuid">The master luid.</param>
            /// <param name="dbTool">The database tool.</param>
            /// <returns></returns>
            public static string GetSSCCInfo(int? luid, int? masterLuid, DbTool dbTool )
            {
                DbQueryBuilder strBuilder = dbTool.GetQueryBuilder();
                // SELECT DISTINCT 
                //                      dbo.PMX_LUID.SSCC, dbo.PMX_LUID.InternalKey AS LUID, dbo.PMX_OSSL.Name AS ShippingDock, dbo.PMX_PLHE.CardName, 
                //                      dbo.PMX_PLHE.CardCode
                //FROM         dbo.PMX_LPLM INNER JOIN
                //                      dbo.PMX_PLHE ON dbo.PMX_LPLM.PickListDocEntry = dbo.PMX_PLHE.DocEntry INNER JOIN
                //                      dbo.PMX_LUID INNER JOIN
                //                      dbo.PMX_MVLI ON dbo.PMX_LUID.InternalKey = dbo.PMX_MVLI.DestLogUnitIdentKey ON 
                //                      dbo.PMX_LPLM.MoveDocEntry = dbo.PMX_MVLI.DocEntry AND dbo.PMX_LPLM.MoveLineNum = dbo.PMX_MVLI.LineNum INNER JOIN
                //                      dbo.PMX_OSSL ON dbo.PMX_MVLI.DestStorLocCode = dbo.PMX_OSSL.Code


                string masterLUIDTable = "masterLUID";

                strBuilder.Append("SELECT DISTINCT ");

                if (masterLuid.HasValue)
                {
                    SqlCommandHelper.AddTableColumnNameWithAs(strBuilder, masterLUIDTable, PmxLogisticUnitIDTable.Columns.SSCC.NAME, SSCCPackingDataSet.Columns.SSCC.Name);
                }
                else
                {
                    SqlCommandHelper.AddTableColumnNameWithAs(strBuilder, PmxLogisticUnitIDTable.NAME, PmxLogisticUnitIDTable.Columns.SSCC.NAME, SSCCPackingDataSet.Columns.SSCC.Name);
                }

                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnNameWithAs(strBuilder, PmxLogisticUnitIDTable.NAME, PmxLogisticUnitIDTable.Columns.InternalKey.NAME, SSCCPackingDataSet.Columns.LUID.Name);

                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnNameWithAs(strBuilder, PmxPickListTable.NAME, PmxPickListTable.Columns.CardCode.NAME, SSCCPackingDataSet.Columns.CARD_CODE.Name);

                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnNameWithAs(strBuilder, PmxPickListTable.NAME, PmxPickListTable.Columns.CardName.NAME, SSCCPackingDataSet.Columns.CARD_NAME.Name);

                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnNameWithAs(strBuilder, PmxPickListTable.NAME, PmxPickListTable.Columns.ShipToAddress.NAME, SSCCPackingDataSet.Columns.SHIP_TO_ADDRESS.Name);

                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnNameWithAs(strBuilder, PmxPickListTable.NAME, PmxPickListTable.Columns.ShipToCode.NAME, SSCCPackingDataSet.Columns.SHIP_TO_ADDRESS_CODE.Name);

                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnNameWithAs(strBuilder, PmxOseGeneralTable.NAME, PmxOseGeneralTable.Columns.Name.NAME, SSCCPackingDataSet.Columns.LOCATION_CODE.Name);

                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnNameWithAs(strBuilder, PmxPickListTable.NAME, PmxPickListTable.Columns.MoveToWarehouse.NAME, SSCCPackingDataSet.Columns.MOVE_TO_WAREHOUSE.Name);




                strBuilder.Append(" FROM ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxMoveLineTable.NAME));

                strBuilder.Append(" INNER JOIN ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxLinkPickListLastMoveTable.NAME));
                strBuilder.Append(" ON ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxLinkPickListLastMoveTable.NAME, PmxLinkPickListLastMoveTable.Columns.MoveDocEntry.NAME);
                strBuilder.Append(" = ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxMoveLineTable.NAME, PmxMoveLineTable.Columns.DocEntry.NAME);
                strBuilder.Append(" AND ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxLinkPickListLastMoveTable.NAME, PmxLinkPickListLastMoveTable.Columns.MoveLineNum.NAME);
                strBuilder.Append(" = ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxMoveLineTable.NAME, PmxMoveLineTable.Columns.LineNum.NAME);

                strBuilder.Append(" INNER JOIN ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxPickListTable.NAME));
                strBuilder.Append(" ON ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxPickListTable.NAME, PmxPickListTable.Columns.DocEntry.NAME);
                strBuilder.Append(" = ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxLinkPickListLastMoveTable.NAME, PmxLinkPickListLastMoveTable.Columns.PickListDocEntry.NAME);

                strBuilder.Append(" INNER JOIN ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxOseGeneralTable.NAME));
                strBuilder.Append(" ON ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxMoveLineTable.NAME, PmxMoveLineTable.Columns.DestinationStorageLocationCode.NAME);
                strBuilder.Append(" = ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseGeneralTable.NAME, PmxOseGeneralTable.Columns.Code.NAME);

                strBuilder.Append(" LEFT OUTER JOIN ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxLogisticUnitIDTable.NAME));
                strBuilder.Append(" ON ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxLogisticUnitIDTable.NAME, PmxLogisticUnitIDTable.Columns.InternalKey.NAME);
                strBuilder.Append(" = ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxMoveLineTable.NAME, PmxMoveLineTable.Columns.DestinationLogisticUnitIdentificationKey.NAME);

                if (masterLuid.HasValue)
                {
                    Produmex.Sbo.Logex.Data.SboCommands.GetLinkedMasterLUIDSubSelect("\r\n LEFT JOIN ", PmxLinkMasterLUIDTable.NAME, strBuilder);
                    //strBuilder.Append("\r\n LEFT JOIN ");
                    //strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxLinkMasterLUIDTable.NAME));
                    strBuilder.Append(" ON ");
                    SqlCommandHelper.AddTableColumnName(strBuilder, PmxMoveLineTable.NAME, PmxMoveLineTable.Columns.DestinationLogisticUnitIdentificationKey.NAME);
                    strBuilder.Append(" = ");
                    SqlCommandHelper.AddTableColumnName(strBuilder, PmxLinkMasterLUIDTable.NAME, PmxLinkMasterLUIDTable.Columns.LUID.NAME);

                    strBuilder.Append(" LEFT JOIN ");
                    strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxLogisticUnitIDTable.NAME));
                    strBuilder.Append(" AS ");
                    strBuilder.Append(SqlCommandHelper.EscapeTableName(masterLUIDTable));
                    strBuilder.Append(" ON ");
                    SqlCommandHelper.AddTableColumnName(strBuilder, masterLUIDTable, PmxLogisticUnitIDTable.Columns.InternalKey.NAME);
                    strBuilder.Append(" = ");
                    SqlCommandHelper.AddTableColumnName(strBuilder, PmxLinkMasterLUIDTable.NAME, PmxLinkMasterLUIDTable.Columns.MasterLUID.NAME);
                }

                strBuilder.Append(" WHERE ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxPickListTable.NAME, PmxPickListTable.Columns.IsCustomerCollect.NAME);
                strBuilder.Append(" = ");
                strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(SboBooleanColumnValues.N.VALUE, CultureInfo.InvariantCulture));

                if (masterLuid.HasValue)
                {
                    strBuilder.Append(" AND ");
                    SqlCommandHelper.AddTableColumnName(strBuilder, PmxLinkMasterLUIDTable.NAME, PmxLinkMasterLUIDTable.Columns.MasterLUID.NAME);
                    strBuilder.Append(" = ");
                    strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(masterLuid.Value, CultureInfo.InvariantCulture));

                }
                else if (luid.HasValue)
                {
                    strBuilder.Append(" AND ");
                    SqlCommandHelper.AddTableColumnName(strBuilder, PmxMoveLineTable.NAME, PmxMoveLineTable.Columns.DestinationLogisticUnitIdentificationKey.NAME);
                    strBuilder.Append(" = ");
                    strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(luid.Value, CultureInfo.InvariantCulture));
                }




                return strBuilder.ToString();
            }


            /// <summary>
            /// Gets the pick reasons.
            /// </summary>
            /// <returns></returns>
            public static string GetPickReasons()
            {
                StringBuilder strBuilder = new StringBuilder();
                strBuilder.Append("SELECT  ");

                SqlCommandHelper.AddTableColumnNameWithAs(strBuilder, PmxReasonTable.NAME, PmxReasonTable.Columns.Code.NAME, ReasonDataSet.Columns.REASON_CODE.Name);
                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnNameWithAs(strBuilder, PmxReasonTable.NAME, PmxReasonTable.Columns.Name.NAME, ReasonDataSet.Columns.REASON_NAME.Name);
                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnNameWithAs(strBuilder, PmxReasonTable.NAME, PmxReasonTable.Columns.RequiresFreeText.NAME, ReasonDataSet.Columns.REASON_REQUIRES_FREE_TEXT.Name);

                strBuilder.Append(" FROM ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxReasonTable.NAME));

                strBuilder.Append(" WHERE ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxReasonTable.Columns.UseForPicking.NAME));
                strBuilder.Append(" = ");
                strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(SboBooleanColumnValues.Y.VALUE, CultureInfo.InvariantCulture));

                strBuilder.AppendFormat(
                   " ORDER BY CASE WHEN {0} IS NULL THEN 1 ELSE 0 END, {0} ",
                    SqlCommandHelper.EscapeTableColumnName(PmxReasonTable.NAME, PmxReasonTable.Columns.SequenceNumber.NAME));


                return strBuilder.ToString();
            }



            ///// <summary>
            ///// Gets the shipping type for pick list.
            ///// </summary>
            ///// <param name="pickListDocEntry">The pick list doc entry.</param>
            ///// <returns>
            ///// The sql statement to get all pick locations
            ///// </returns>
            //public static string GetShippingTypeForPickList( int pickListDocEntry )
            //{

            //    //SELECT     dbo.OSHP.U_PMX_AUSH
            //    //FROM         dbo.PMX_PLHE INNER JOIN
            //    //          dbo.PMX_PLPH ON dbo.PMX_PLHE.PickListProposalEntry = dbo.PMX_PLPH.DocEntry INNER JOIN
            //    //          dbo.PMX_PLPL ON dbo.PMX_PLPH.DocEntry = dbo.PMX_PLPL.DocEntry INNER JOIN
            //    //          dbo.ORDR ON dbo.PMX_PLPL.BaseEntry = dbo.ORDR.DocEntry AND dbo.PMX_PLPL.BaseType = dbo.ORDR.ObjType LEFT OUTER JOIN
            //    //          dbo.OSHP ON dbo.ORDR.TrnspCode = dbo.OSHP.TrnspCode

            //    StringBuilder strBuilder = new StringBuilder();
            //    strBuilder.Append( "SELECT " );

            //    SqlCommandHelper.AddTableColumnName( strBuilder, PmxShippingTypeTable.NAME, PmxShippingTypeTable.Columns.AutomaticShipping.NAME );

            //    strBuilder.Append( " FROM " );
            //    strBuilder.Append( SqlCommandHelper.EscapeTableName( PmxPickListTable.NAME ) );

            //    strBuilder.Append( " INNER JOIN " );
            //    strBuilder.Append( SqlCommandHelper.EscapeTableName( PmxPickListProposalTable.NAME ) );
            //    strBuilder.Append( " ON " );
            //    SqlCommandHelper.AddTableColumnName( strBuilder, PmxPickListProposalTable.NAME, PmxPickListProposalTable.Columns.DocEntry.NAME );
            //    strBuilder.Append( " = " );
            //    SqlCommandHelper.AddTableColumnName( strBuilder, PmxPickListTable.NAME, PmxPickListTable.Columns.PickListProposalEntry.NAME );

            //    strBuilder.Append( " INNER JOIN " );
            //    strBuilder.Append( SqlCommandHelper.EscapeTableName( PmxPickListProposalLineTable.NAME ) );
            //    strBuilder.Append( " ON " );
            //    SqlCommandHelper.AddTableColumnName( strBuilder, PmxPickListProposalLineTable.NAME, PmxPickListProposalLineTable.Columns.DocEntry.NAME );
            //    strBuilder.Append( " = " );
            //    SqlCommandHelper.AddTableColumnName( strBuilder, PmxPickListProposalTable.NAME, PmxPickListProposalTable.Columns.DocEntry.NAME );

            //    strBuilder.Append( " INNER JOIN " );
            //    strBuilder.Append( SqlCommandHelper.EscapeTableName( SboSalesOrderTable.NAME ) );
            //    strBuilder.Append( " ON " );
            //    SqlCommandHelper.AddTableColumnName( strBuilder, SboSalesOrderTable.NAME, SboSalesOrderTable.Columns.DocEntry.NAME );
            //    strBuilder.Append( " = " );
            //    SqlCommandHelper.AddTableColumnName( strBuilder, PmxPickListProposalLineTable.NAME, PmxPickListProposalLineTable.Columns.BaseEntry.NAME );
            //    strBuilder.Append( " AND " );
            //    SqlCommandHelper.AddTableColumnName( strBuilder, SboSalesOrderTable.NAME, SboSalesOrderTable.Columns.ObjType.NAME );
            //    strBuilder.Append( " = " );
            //    SqlCommandHelper.AddTableColumnName( strBuilder, PmxPickListProposalLineTable.NAME, PmxPickListProposalLineTable.Columns.BaseType.NAME );

            //    strBuilder.Append( " LEFT OUTER JOIN " );
            //    strBuilder.Append( SqlCommandHelper.EscapeTableName( PmxShippingTypeTable.NAME ) );
            //    strBuilder.Append( " ON " );
            //    SqlCommandHelper.AddTableColumnName( strBuilder, PmxShippingTypeTable.NAME, PmxShippingTypeTable.Columns.Code.NAME );
            //    strBuilder.Append( " = " );
            //    SqlCommandHelper.AddTableColumnName( strBuilder, SboSalesOrderTable.NAME, SboSalesOrderTable.Columns.TrnspCode.NAME );



            //    strBuilder.Append( " WHERE " );
            //    strBuilder.Append( SqlCommandHelper.EscapeColumnName( PmxPickListTable.NAME ) );
            //    strBuilder.Append( "." );
            //    strBuilder.Append( SqlCommandHelper.EscapeColumnName( PmxPickListTable.Columns.DocEntry.NAME ) );
            //    strBuilder.Append( " = " );
            //    strBuilder.Append( SqlCommandHelper.FormatAndEscapeColumnValue( pickListDocEntry, CultureInfo.InvariantCulture ) );


            //    return strBuilder.ToString();
            //}


            internal static string GetPickListMultiPickingInfo( string pickListBarcode, bool isProductionPicking, DbTool dbTool )
            {
                DbQueryBuilder strBuilder = dbTool.GetQueryBuilder();
                bool isHana = (dbTool.DbServerType == DbServerType.HANA);

                const string scannedPickList = "spl";

                strBuilder.AppendFormat("SELECT {0} AS spl_DocEntry",
                    SqlCommandHelper.EscapeTableColumnName(scannedPickList, PmxPickListTable.Columns.DocEntry.NAME));
                strBuilder.AppendFormat(", {0} AS spl_Status",
                    SqlCommandHelper.EscapeTableColumnName(scannedPickList, PmxPickListTable.Columns.PickListStatus.NAME));
                strBuilder.AppendFormat(", {0}",
                    SqlCommandHelper.EscapeTableColumnName(scannedPickList, PmxPickListTable.Columns.WaveKey.NAME));
                strBuilder.AppendFormat(", {0}",
                    SqlCommandHelper.EscapeTableColumnName(SboUserTable.NAME, SboUserTable.Columns.UserCode.NAME));
                strBuilder.AppendFormat(", {0}",
                    SqlCommandHelper.EscapeTableColumnName(PmxPickListTable.NAME, PmxPickListTable.Columns.DocEntry.NAME));
                strBuilder.AppendFormat(", {0}",
                    SqlCommandHelper.EscapeTableColumnName(PmxPickListTable.NAME, PmxPickListTable.Columns.PickListStatus.NAME));
                strBuilder.AppendFormat(", {0}",
                    SqlCommandHelper.EscapeTableColumnName(PmxPickListTable.NAME, PmxPickListTable.Columns.CardCode.NAME));
                strBuilder.AppendFormat(", {0}",
                    SqlCommandHelper.EscapeTableColumnName(PmxPickListTable.NAME, PmxPickListTable.Columns.CardName.NAME));
                strBuilder.AppendFormat(", {0}",
                     SqlCommandHelper.EscapeTableColumnName(PmxOseGeneralTable.NAME, PmxOseGeneralTable.Columns.PmxWarehouseCode.NAME));
                strBuilder.AppendFormat(", {0}",
                    SqlCommandHelper.EscapeTableColumnName(PmxPickListTable.NAME, PmxPickListTable.Columns.MultiPickingMovLoc.NAME));

                strBuilder.AppendFormat("\r\n FROM {0} {1} {2}",
                    SqlCommandHelper.EscapeTableName(PmxPickListTable.NAME), SqlCommandHelper.EscapeTableName(scannedPickList), strBuilder.TableUpdateLock);

                strBuilder.AppendFormat("\r\n LEFT JOIN ( {0} {1}",
                    SqlCommandHelper.EscapeTableName(PmxWaveTable.NAME), strBuilder.TableUpdateLock);
                {
                    strBuilder.AppendFormat("\r\n    INNER JOIN {0} {1}",
                        SqlCommandHelper.EscapeTableName(PmxPickListTable.NAME), strBuilder.TableUpdateLock);
                    strBuilder.AppendFormat( " ON {0} = {1}",
                        SqlCommandHelper.EscapeTableColumnName(PmxPickListTable.NAME, PmxPickListTable.Columns.WaveKey.NAME),
                        SqlCommandHelper.EscapeTableColumnName(PmxWaveTable.NAME, PmxWaveTable.Columns.InternalKey.NAME));

                    strBuilder.Append("\r\n    INNER JOIN ");
                    strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxOseGeneralTable.NAME));
                    strBuilder.Append(" ON ");
                    SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseGeneralTable.NAME, PmxOseGeneralTable.Columns.Code.NAME);
                    strBuilder.Append(" = ");
                    SqlCommandHelper.AddTableColumnName(strBuilder, PmxPickListTable.NAME, PmxPickListTable.Columns.DestStorLocCode.NAME);
                }
                strBuilder.AppendFormat("\r\n ) ON {0} = {1}",
                    SqlCommandHelper.EscapeTableColumnName(PmxWaveTable.NAME, PmxWaveTable.Columns.InternalKey.NAME),
                    SqlCommandHelper.EscapeTableColumnName(scannedPickList, PmxPickListTable.Columns.WaveKey.NAME));

                strBuilder.AppendFormat("\r\n LEFT JOIN {0}",
                    SqlCommandHelper.EscapeTableName(SboUserTable.NAME));
                strBuilder.AppendFormat(" ON {0} = {1}",
                    SqlCommandHelper.EscapeTableColumnName(SboUserTable.NAME, SboUserTable.Columns.InternalKey.NAME),
                    SqlCommandHelper.EscapeTableColumnName(PmxWaveTable.NAME, PmxWaveTable.Columns.LockedBy.NAME));

                strBuilder.AppendFormat("\r\n WHERE {0} = {1}",
                    SqlCommandHelper.EscapeTableColumnName(scannedPickList, PmxPickListTable.Columns.DocEntry.NAME),
                    SqlCommandHelper.FormatAndEscapeColumnValue(pickListBarcode, CultureInfo.InvariantCulture));

                List<string> allowedPickObjTypes = new List<string>();
                if ( isProductionPicking )
                {
                    allowedPickObjTypes.Add( PmxPickObjectTypeColumnValues.Production.VALUE );
                    allowedPickObjTypes.Add( PmxPickObjectTypeColumnValues.WhsTransferProd.VALUE );
                }
                else
                {
                    allowedPickObjTypes.Add( PmxPickObjectTypeColumnValues.Sales.VALUE );
                    allowedPickObjTypes.Add( PmxPickObjectTypeColumnValues.WhsTransfer.VALUE );
                }
                strBuilder.AppendFormat( "\r\n AND {0} IN ({1}) ",
                   SqlCommandHelper.EscapeTableColumnName( scannedPickList, PmxPickListTable.Columns.PickObjType.NAME ),
                   SqlCommandHelper.FormatAndEscapeColumnValue( allowedPickObjTypes, CultureInfo.InvariantCulture ) );

                return strBuilder.ToString();
            }

            internal static string UpdateMultiPickingMovLoc(SortedList<int, MultiPickingPickListInfo> multiPickingPickListInfos, bool doSelectOnly, PmxDbConnection dbConn)
            {
                StringBuilder strBuilder = new StringBuilder();

                if ( doSelectOnly )
                {
                    strBuilder.AppendFormat( "SELECT COUNT(1) FROM {0}",
                        SqlCommandHelper.EscapeTableName( PmxPickListTable.NAME ) );
                }
                else
                {
                    strBuilder.AppendFormat( "UPDATE {0}",
                        SqlCommandHelper.EscapeTableName( PmxPickListTable.NAME ) );

                    strBuilder.AppendFormat( " SET {0} = CASE {1}",
                        SqlCommandHelper.EscapeColumnName( PmxPickListTable.Columns.MultiPickingMovLoc.NAME ),
                        SqlCommandHelper.EscapeTableColumnName( PmxPickListTable.NAME, PmxPickListTable.Columns.DocEntry.NAME ) );
                    foreach ( MultiPickingPickListInfo info in multiPickingPickListInfos.Values )
                    {
                        strBuilder.AppendFormat( " WHEN {0} THEN {1}",
                            SqlCommandHelper.FormatAndEscapeColumnValue( info.DocEntry, dbConn.DbCultureInfo ),
                            SqlCommandHelper.FormatAndEscapeColumnValue( info.MovableLocation, dbConn.DbCultureInfo ) );
                    }
                    strBuilder.Append( " ELSE NULL END" );
                }

                string separator = "\r\n WHERE ";
                foreach (MultiPickingPickListInfo info in multiPickingPickListInfos.Values)
                {
                    strBuilder.Append(separator); separator = "\r\n OR ";

                    strBuilder.AppendFormat("( {0} = {1} AND ( {2} = {3} OR {2} IS NULL ) )",
                        SqlCommandHelper.EscapeTableColumnName(PmxPickListTable.NAME, PmxPickListTable.Columns.DocEntry.NAME),
                        SqlCommandHelper.FormatAndEscapeColumnValue(info.DocEntry, dbConn.DbCultureInfo),
                        SqlCommandHelper.EscapeTableColumnName(PmxPickListTable.NAME, PmxPickListTable.Columns.MultiPickingMovLoc.NAME),
                        SqlCommandHelper.FormatAndEscapeColumnValue(info.MovableLocation, dbConn.DbCultureInfo));
                }

                if ( dbConn.DbTool.DbServerType != DbServerType.HANA )
                {
                    strBuilder.Append("; SELECT @@ROWCOUNT");
                }

                return strBuilder.ToString();
            }

            internal static string GetPickListTypesForUser( string userCode, DbTool dbTool )
            {
                DbQueryBuilder qb = dbTool.GetQueryBuilder();

                qb.Append( "SELECT ",
                    qb.AsIdentifier( PmxUserPickListTypeTable.NAME, PmxUserPickListTypeTable.Columns.PickListType.NAME ) );

                qb.Append( "\r\n FROM ",
                    qb.AsIdentifier( PmxUserPickListTypeTable.NAME ) );

                qb.Append( "\r\n WHERE ", qb.AsIdentifier( PmxUserPickListTypeTable.NAME, PmxUserPickListTypeTable.Columns.UserCode.NAME ),
                                   " = ", qb.AsString( userCode ) );

                qb.Append( " AND ", qb.AsIdentifier( PmxUserPickListTypeTable.NAME, PmxUserPickListTypeTable.Columns.PickListType.NAME ),
                               " <> ''" );

                return qb.ToString();
            }

            /// <summary>
            /// Gets the item pick types for user.
            /// </summary>
            /// <param name="userCode">The user code.</param>
            /// <param name="dbCulture">The db culture.</param>
            /// <returns></returns>
            internal static string GetItemPickTypesForUser(string userCode, CultureInfo dbCulture)
            {
                StringBuilder strBuilder = new StringBuilder();

                strBuilder.AppendFormat("SELECT {0}",
                    SqlCommandHelper.EscapeTableColumnName(PmxUserItemPickTypeTable.NAME, PmxUserItemPickTypeTable.Columns.ItemPickType.NAME));

                strBuilder.AppendFormat(", {0}",
                    SqlCommandHelper.EscapeTableColumnName(PmxUserItemPickTypeTable.NAME, PmxUserItemPickTypeTable.Columns.ItemPickTypeProperty.NAME));

                strBuilder.AppendFormat("\r\n FROM {0}",
                    SqlCommandHelper.EscapeTableName(PmxUserItemPickTypeTable.NAME));

                strBuilder.AppendFormat("\r\n WHERE {0} = {1}",
                    SqlCommandHelper.EscapeTableColumnName(PmxUserItemPickTypeTable.NAME, PmxUserItemPickTypeTable.Columns.UserCode.NAME),
                    SqlCommandHelper.FormatAndEscapeColumnValue(userCode, dbCulture));

                return strBuilder.ToString();
            }



            internal static string SetLogUnitReadyForShipping(int luid, bool ready, DbTool dbTool)
            {
                return Produmex.Sbo.Logex.Data.SboCommands.SetLogUnitReadyForShipping( luid, ready, dbTool );

            }


            /// <summary>
            /// Sets the log unit ready for shipping.
            /// </summary>
            /// <param name="masterLuid">The master luid.</param>
            /// <param name="ready">if set to <c>true</c> [ready].</param>
            /// <param name="dbTool">The db tool.</param>
            /// <returns></returns>
            internal static string SetMasterLogUnitReadyForShipping(int masterLuid, bool ready, DbTool dbTool)
            {
                return Produmex.Sbo.Logex.Data.SboCommands.SetMasterLogUnitReadyForShipping(masterLuid, ready, dbTool);

            }


            internal static string GetSupplierPalletNumberForPickList(int pickListDocEntry, int pickListLineNum, DbTool dbTool)
            {
                return Produmex.Sbo.Logex.Data.SboCommands.GetSupplierPalletNumberForPickList( pickListDocEntry, pickListLineNum, dbTool );
            }

            internal static string GetAssignedWarehousesOnDeviceID(string deviceID)
            {
                StringBuilder strBuilder = new StringBuilder();

                strBuilder.AppendFormat("SELECT DISTINCT {0}",
                      SqlCommandHelper.EscapeTableColumnName(PmxThinClientWarehouseTable.NAME, PmxThinClientWarehouseTable.Columns.PmxWarehouseCode.NAME));

                strBuilder.AppendFormat(" FROM {0}",
                    SqlCommandHelper.EscapeTableName(PmxThinClientWarehouseTable.NAME));

                strBuilder.AppendFormat(" WHERE {0} = {1}",
                    SqlCommandHelper.EscapeTableColumnName(PmxThinClientWarehouseTable.NAME, PmxThinClientWarehouseTable.Columns.ParentCode.NAME),
                    SqlCommandHelper.FormatAndEscapeColumnValue(deviceID, CultureInfo.InvariantCulture));

                return strBuilder.ToString();
            }

            /// <summary>
            /// Gets the always set to status picked after picking.
            /// </summary>
            /// <param name="waveKey">The wave key.</param>
            /// <param name="cultureDatabase">The culture database.</param>
            /// <returns></returns>
            internal static string GetAlwaysSetToStatusPickedAfterPicking(int waveKey)
            {
                StringBuilder strBuilder = new StringBuilder();
                strBuilder.Append("SELECT ");

                SqlCommandHelper.AddTableColumnName(strBuilder, PmxPickListTable.NAME, PmxPickListTable.Columns.DocEntry.NAME);

                strBuilder.Append(" FROM ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxPickListTable.NAME));

                strBuilder.Append(" INNER JOIN ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxPickListTypeTable.NAME));
                strBuilder.Append(" ON ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxPickListTypeTable.NAME, PmxPickListTypeTable.Columns.Code.NAME);
                strBuilder.Append(" = ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxPickListTable.NAME, PmxPickListTable.Columns.PickListType.NAME);

                strBuilder.Append(" WHERE ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxPickListTable.NAME, PmxPickListTable.Columns.WaveKey.NAME);
                strBuilder.Append(" = ");
                strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(waveKey, CultureInfo.InvariantCulture));

                strBuilder.Append(" AND ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxPickListTypeTable.NAME, PmxPickListTypeTable.Columns.AlwaysSetToStatusPickedAfterPicking.NAME);
                strBuilder.Append(" = ");
                strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(SboBooleanColumnValues.Y.VALUE, CultureInfo.InvariantCulture));

                return strBuilder.ToString();
            }

        }
    }
}
